/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_apsgap.c 
 *
 * Purpose: MX motor driver for controlling an Advanced Photon Source
 *          undulator gap.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "mx_aout.h"
#include "d_apsgap.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_aps_gap_record_function_list = {
	mxd_aps_gap_initialize_type,
	NULL,
	NULL,
	mxd_aps_gap_create_record_structures,
	mxd_aps_gap_finish_record_initialization,
	mxd_aps_gap_delete_record,
	mxd_aps_gap_print_structure,
	NULL,
	mxd_aps_gap_read_parms_from_hardware,
	mxd_aps_gap_write_parms_to_hardware,
	mxd_aps_gap_open,
	mxd_aps_gap_close
};

MX_MOTOR_FUNCTION_LIST mxd_aps_gap_motor_function_list = {
	mxd_aps_gap_motor_is_busy,
	mxd_aps_gap_move_absolute,
	mxd_aps_gap_get_position,
	mxd_aps_gap_set_position,
	mxd_aps_gap_soft_abort,
	mxd_aps_gap_immediate_abort,
	mxd_aps_gap_positive_limit_hit,
	mxd_aps_gap_negative_limit_hit,
	mxd_aps_gap_find_home_position
};

/* APS gap motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_aps_gap_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_APS_GAP_STANDARD_FIELDS
};

long mxd_aps_gap_num_record_fields
		= sizeof( mxd_aps_gap_record_field_defaults )
			/ sizeof( mxd_aps_gap_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_aps_gap_record_field_def_ptr
			= &mxd_aps_gap_record_field_defaults[0];

/* Internal use only function. */

static mx_status_type
mxd_aps_gap_change_gap_setpoint( char *pvname, double value, int sector );

#define CHECK_STATUS(x) \
		do { \
			status = (x); \
			if ( status.code != MXE_SUCCESS ) \
				return status; \
		} while(0)

/* === */

MX_EXPORT mx_status_type
mxd_aps_gap_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_aps_gap_create_record_structures()";

	MX_MOTOR *motor;
	MX_APS_GAP *aps_gap;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	aps_gap = (MX_APS_GAP *) malloc( sizeof(MX_APS_GAP) );

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_APS_GAP structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = aps_gap;
	record->class_specific_function_list
				= &mxd_aps_gap_motor_function_list;

	motor->record = record;

	/* An APS gap motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_aps_gap_finish_record_initialization()";

	MX_APS_GAP *aps_gap;
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	aps_gap = (MX_APS_GAP *) (record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for record '%s' is NULL.",
			record->name );
	}

	if ( aps_gap->sector_number < 1 || aps_gap->sector_number > 35 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal APS sector number %d.  Allowed range is 1 to 35.",
			aps_gap->sector_number );
	}
	if ( aps_gap->motor_subtype < 1 || aps_gap->motor_subtype > 10 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal motor subtype number %d.  Allowed range is 1 to 10.",
			aps_gap->motor_subtype );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_aps_gap_print_structure()";

	MX_MOTOR *motor;
	MX_APS_GAP *aps_gap;
	char raw_units[4];
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	aps_gap = (MX_APS_GAP *) (record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type      = APS_GAP.\n\n");

	fprintf(file, "  name            = %s\n", record->name);
	fprintf(file, "  sector number   = %d\n", aps_gap->sector_number);
	fprintf(file, "  motor subtype   = %d\n", aps_gap->motor_subtype);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	switch( aps_gap->motor_subtype ) {
	case MXT_APS_GAP:
		strcpy( raw_units, "mm" );
		break;
	case MXT_APS_KEV:
		strcpy( raw_units, "keV" );
		break;
	default:
		strcpy( raw_units, "??" );
		break;
	}

	fprintf(file, "  position        = %g %s  (%g %s)\n",
		motor->position, motor->units,
		motor->raw_position.analog, raw_units );
	fprintf(file, "  scale           = %g %s per %s.\n",
		motor->scale, motor->units, raw_units );
	fprintf(file, "  offset          = %g %s.\n",
		motor->offset, motor->units);

	fprintf(file, "  backlash        = %g %s  (%g %s).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog, raw_units );
	fprintf(file, "  negative limit  = %g %s  (%g %s).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog, raw_units );
	fprintf(file, "  positive limit  = %g %s  (%g %s).\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.analog, raw_units );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband        = %g %s  (%g %s).\n\n",
		deadband, motor->units,
		motor->deadband.analog, raw_units );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_open( MX_RECORD *record )
{
	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_close( MX_RECORD *record )
{
	/* We need do nothing special here.  The connection to EPICS
	 * will go away automatically when the program exits.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_motor_is_busy()";

	MX_APS_GAP *aps_gap;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[300];
	short busy_record_value;
	int ezca_status;

	motor->busy = FALSE;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	aps_gap = (MX_APS_GAP *) (motor->record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf( pvname, "ID%02d:Busy.VAL", aps_gap->sector_number );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &busy_record_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read motion status of undulator gap for sector '%d' failed.  "
"PV name = '%s'.  EZCA status = %d (%s)",
			aps_gap->sector_number, pvname,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( busy_record_value == 0 ) {
		motor->busy = FALSE;
	} else {
		motor->busy = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

/* The following function is only used by mxd_aps_gap_move_absolute(). */

static mx_status_type
mxd_aps_gap_change_gap_setpoint( char *pvname, double value, int sector )
{
	const char fname[] = "mxd_aps_gap_change_gap_setpoint()";

	int ezca_status;
	char *ezca_error_message;
	char mx_error_message[300];

	/* Send the new setpoint value. */

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to change APS sector %d undulator gap setpoint failed.  "
	"PV name = '%s'.  EZCA status = %d (%s)",
			sector, pvname, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_move_absolute()";

	MX_APS_GAP *aps_gap;
	double new_position;
	char pvname[60];
	double open_gap_setpoint;
	char *ezca_error_message;
	char mx_error_message[300];
	int ezca_status;
	int sector;
	long access_mode, move_command;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed is NULL.");
	}

	aps_gap = (MX_APS_GAP *) (motor->record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for record '%s' is NULL.",
			motor->record->name );
	}

	sector = aps_gap->sector_number;

	new_position = motor->raw_setpoint.analog;

	/* Do we have permission to change the undulator gap right now? */

	sprintf( pvname, "ID%02d:AccessSecurity.VAL", sector );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &access_mode );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to check the access mode of the APS sector %d undulator gap failed.  "
"PV name = '%s'.  EZCA status = %d (%s)",
			sector, pvname, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( access_mode != 0 ) {
		return mx_error( MXE_PERMISSION_DENIED, fname,
"The sector %d insertion device is not in user control mode right now.",
			sector );
	}

	/* Construct the names of the EPICS process variables that 
	 * correspond to the motor setpoints.
	 */

	switch ( aps_gap->motor_subtype ) {
	case MXT_APS_GAP:
		sprintf( pvname, "ID%02d:GapSet.VAL", sector );
		break;
	case MXT_APS_KEV:
		sprintf( pvname, "ID%02d:EnergySet.VAL", sector );
		break;
	default:
		return mx_error(MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal APS gap motor subtype %d.", aps_gap->motor_subtype);
		break;
	}

	/* Change the setpoints as necessary to ensure correct operation. */

	switch( aps_gap->motor_subtype ) {

	/* For MXT_APS_GAP and MXT_APS_KEV, we need to change the setpoint
	 * to full open and then reset it to the position we really want.
	 * This is to overcome a problem that can occur when we use these
	 * two motor types to perform a move when the undulator is currently
	 * tapered.
	 *
	 * The nature of the problem is that the last setpoint set using
	 * the taper control variables is copied over to the setpoint
	 * for the untapered gap for some unknown reason.  If we then
	 * command a move using the the untapered variables to a position
	 * that is equal to the setpoint, EPICS solemnly informs us that
	 * we are already at the requested gap even though we really aren't.
	 * Thus, the procedure we must follow is:
	 *
	 * 1.  Set the setpoint to full open (200 mm or 14.107 keV).
	 * 2.  Then, set the setpoint to the real requested value.
	 */
	case MXT_APS_GAP:
		open_gap_setpoint = 200.0;

#if 0
		CHECK_STATUS( mxd_aps_gap_change_gap_setpoint(
			pvname, open_gap_setpoint, sector ));
#endif
		CHECK_STATUS( mxd_aps_gap_change_gap_setpoint(
			pvname, new_position, sector ));
		break;
	case MXT_APS_KEV:
		open_gap_setpoint = 14.107;

#if 0
		CHECK_STATUS( mxd_aps_gap_change_gap_setpoint(
			pvname, open_gap_setpoint, sector ));
#endif
		CHECK_STATUS( mxd_aps_gap_change_gap_setpoint(
			pvname, new_position, sector ));
		break;
	}

	/* Finally, tell the undulator to start moving. */

	sprintf( pvname, "ID%02d:Start.VAL", sector );

	move_command = 1;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &move_command );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to start move of APS sector %d undulator gap failed.  "
	"PV name = '%s'.  EZCA status = %d (%s)",
			sector, pvname, ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_get_position()";

	MX_APS_GAP *aps_gap;
	double present_value;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[300];
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	aps_gap = (MX_APS_GAP *) (motor->record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for record '%s' is NULL.",
			motor->record->name );
	}

	switch( aps_gap->motor_subtype ) {
	case MXT_APS_GAP:
		sprintf( pvname, "ID%02d:Gap.VAL",
				aps_gap->sector_number );
		break;
	case MXT_APS_KEV:
		sprintf( pvname, "ID%02d:Energy.VAL",
				aps_gap->sector_number );
		break;
	default:
		return mx_error(MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal APS gap motor subtype %d.", aps_gap->motor_subtype);
		break;
	}

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &present_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to get position of APS sector %d undulator gap failed.  "
	"PV name = '%s'.  EZCA status = %d (%s)",
			aps_gap->sector_number, pvname,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	motor->raw_position.analog = present_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not supported for an APS undulator gap." );
}

MX_EXPORT mx_status_type
mxd_aps_gap_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_soft_abort()";

	MX_APS_GAP *aps_gap;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[300];
	short stop_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	aps_gap = (MX_APS_GAP *) (motor->record->record_type_struct);

	if ( aps_gap == (MX_APS_GAP *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_APS_GAP pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf( pvname, "ID%02d:Stop.VAL", aps_gap->sector_number );

	/* Stop the gap change. */

	stop_field = 1;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &stop_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to abort APS sector %d undulator gap change failed.  "
	"PV name = '%s'.  EZCA status = %d (%s)",
			aps_gap->sector_number, pvname,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_immediate_abort( MX_MOTOR *motor )
{
	return mxd_aps_gap_soft_abort( motor );
}

MX_EXPORT mx_status_type
mxd_aps_gap_positive_limit_hit( MX_MOTOR *motor )
{
	/* There is no way provided by the APS to determine this. */

	motor->positive_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_negative_limit_hit( MX_MOTOR *motor )
{
	/* There is no way provided by the APS to determine this. */

	motor->negative_limit_hit = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_aps_gap_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_aps_gap_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'find home position' is not supported for an APS undulator gap." );
}

#endif /* HAVE_EPICS */

