/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_cmpmtr.c
 *
 * Purpose: MX driver for the Compumotor 6000 series of motor controllers.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "d_cmpmtr.h"

/* ============ Motor channels ============ */

MX_RECORD_FUNCTION_LIST mxd_compumotor_record_function_list = {
	mxd_compumotor_initialize_type,
	NULL,
	NULL,
	mxd_compumotor_create_record_structures,
	mxd_compumotor_finish_record_initialization,
	mxd_compumotor_delete_record,
	mxd_compumotor_print_structure,
	NULL,
	mxd_compumotor_read_parms_from_hardware,
	mxd_compumotor_write_parms_to_hardware,
	mxd_compumotor_open,
	mxd_compumotor_close,
	NULL,
	mxd_compumotor_resynchronize
};

MX_MOTOR_FUNCTION_LIST mxd_compumotor_motor_function_list = {
	mxd_compumotor_motor_is_busy,
	mxd_compumotor_move_absolute,
	mxd_compumotor_get_position,
	mxd_compumotor_set_position,
	mxd_compumotor_soft_abort,
	mxd_compumotor_immediate_abort,
	mxd_compumotor_positive_limit_hit,
	mxd_compumotor_negative_limit_hit,
	mxd_compumotor_find_home_position
};

/* Compumotor motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_compumotor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_STEPPER_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_COMPUMOTOR_STANDARD_FIELDS
};

long mxd_compumotor_num_record_fields
		= sizeof( mxd_compumotor_record_field_defaults )
			/ sizeof( mxd_compumotor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_compumotor_rfield_def_ptr
			= &mxd_compumotor_record_field_defaults[0];

#define COMPUMOTOR_DEBUG	FALSE

MX_EXPORT mx_status_type
mxd_compumotor_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_create_record_structures()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	compumotor = (MX_COMPUMOTOR *) malloc( sizeof(MX_COMPUMOTOR) );

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_COMPUMOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = compumotor;
	record->class_specific_function_list
				= &mxd_compumotor_motor_function_list;

	motor->record = record;

	/* An elapsed time motor is treated as an stepper motor. */

	motor->subclass = MXC_MTR_STEPPER;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_print_structure()";

	MX_MOTOR *motor;
	MX_GENERIC *generic;
	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR *compumotor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	compumotor = (MX_COMPUMOTOR *) (record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_COMPUMOTOR pointer for record '%s' is NULL.", record->name );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			record->name );
	}

	generic = (MX_GENERIC *)
			compumotor_interface_record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_GENERIC pointer for record '%s' is NULL.", record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);

	fprintf(file, "  Motor type        = COMPUMOTOR motor.\n\n");

	fprintf(file, "  name              = %s\n", record->name);
	fprintf(file, "  port name         = %s\n", generic->record->name);
	fprintf(file, "  controller number = %d\n",
					compumotor->controller_number);
	fprintf(file, "  axis number       = %d\n", compumotor->axis_number);
	fprintf(file, "  use encoder       = %d\n", compumotor->use_encoder);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
		}
	
	fprintf(file, "  position          = %g %s  (%ld steps)\n",
			motor->position, motor->units,
			motor->raw_position.stepper );
	fprintf(file, "  scale             = %g %s per step.\n",
			motor->scale, motor->units);
	fprintf(file, "  offset            = %g %s.\n",
			motor->offset, motor->units);
	
	fprintf(file, "  backlash          = %g %s  (%ld steps)\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.stepper );
	
	fprintf(file, "  negative limit    = %g %s  (%ld steps)\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.stepper );

	fprintf(file, "  positive limit    = %g %s  (%ld steps)\n",
		motor->positive_limit, motor->units,
		motor->raw_positive_limit.stepper );

	deadband = motor->scale * (double) motor->deadband.stepper;

	fprintf(file, "  deadband          = %g %s  (%ld steps)\n\n",
		deadband, motor->units,
		motor->deadband.stepper );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_read_parms_from_hardware()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;
	MX_RECORD *compumotor_interface_record;
	MX_GENERIC *generic;
	double position;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	compumotor = (MX_COMPUMOTOR *) (record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			record->name );
	}

	generic = (MX_GENERIC *)
			compumotor_interface_record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for Compumotor interface '%s' is NULL.",
			compumotor_interface_record->name );
	}

#if 0
	/* Flush out any unwritten and unread characters before sending
	 * commands to the Compumotor controller.
	 */

	status = mx_generic_discard_unwritten_chars( generic );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

	status = mx_generic_discard_unread_chars( generic );

	if ( status.code != MXE_SUCCESS )
		return status;
#endif

	/* Get the current position.  This has the side effect of
	 * updating the position value in the MX_MOTOR structure.
	 */

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we read out.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_write_parms_to_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_write_parms_to_hardware()";

	MX_MOTOR *motor;
	MX_COMPUMOTOR *compumotor;
	MX_RECORD *compumotor_interface_record;
	MX_GENERIC *generic;
#if 0
	mx_status_type status;
#endif

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	compumotor = (MX_COMPUMOTOR *) (record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE,
			"MX_COMPUMOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			record->name );
	}

	generic = (MX_GENERIC *)
			compumotor_interface_record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for Compumotor interface '%s' is NULL.",
			compumotor_interface_record->name );
	}

#if 0
	/* Flush out any written and unread characters before sending
	 * commands to the Compumotor controller.
	 */

	status = mx_generic_discard_unwritten_chars( generic );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

#endif

	/* === Now we can set the motor position. === */

#if 0	/* For now, this is disabled. */

	status = mxd_compumotor_set_position_steps(
			motor, motor->raw_motor_steps );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* At present, the position is the only parameter we set here.
	 * We rely on the non-volatile memory of the controller
	 * for the rest of the controller parameters.
	 */
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_open( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_open()";

	MX_MOTOR *motor;
	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	char response[80];
	int length;
	mx_status_type status;

	MX_DEBUG(2,("%s invoked.", fname));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	compumotor = (MX_COMPUMOTOR *) (record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			record->name );
	}

	MX_DEBUG(2,("%s: About to send first command.", fname));

	/* Verify that the controller is present by asking for 
	 * its revision number.
	 */

	sprintf( command, "%d_!TREV", compumotor->controller_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	MX_DEBUG(2,("%s: Returned from first command.  status = %d",
		fname, status.code));

	switch( status.code ) {
	case MXE_SUCCESS:
		break;
	case MXE_INTERFACE_IO_ERROR:
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Cannot communicate with Compumotor controller %d on RS-232 port '%s'.  "
"Is it turned on?",
			compumotor->controller_number,
			compumotor_interface->rs232_record->name );
		break;
	default:
		return status;
		break;
	}

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to controller revision level command.");
	}

	MX_DEBUG(2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_close( MX_RECORD *record )
{
	/* Nothing to do. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxd_compumotor_resynchronize()";

	MX_COMPUMOTOR *compumotor;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) record->record_type_struct;

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	status = mxi_compumotor_resynchronize(
			compumotor->compumotor_interface_record );

	return status;
}

/* ============ Motor specific functions ============ */

MX_EXPORT mx_status_type
mxd_compumotor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_motor_is_busy()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	char response[80];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	if ( response[0] == '1' ) {
		motor->busy = TRUE;
	} else {
		motor->busy = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_move_absolute()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	long motor_steps;
	int i, length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	/* Set the position to move to. */

	motor_steps = motor->raw_setpoint.stepper;

	sprintf( command, "%d_!%dD%ld", compumotor->controller_number,
			compumotor->axis_number, motor_steps );

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Construct the move command. */

	sprintf( command, "%d_!GO", compumotor->controller_number );

	length = strlen( command );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		if ( (i+1) == compumotor->axis_number ) {
			command[ length + i ] = '1';
		} else {
			command[ length + i ] = 'X';
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Command the move to start. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_get_position()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	char response[50];
	long motor_steps;
	int num_tokens;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	if ( compumotor->use_encoder ) {
		sprintf( command, "%d_!%dTPE",
			compumotor->controller_number,
			compumotor->axis_number );
	} else {
		sprintf( command, "%d_!%dTPM",
			compumotor->controller_number,
			compumotor->axis_number );
	}

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_tokens = sscanf( response, "%ld", &motor_steps );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"No position value seen in response to '%s' command.  "
		"Response seen = '%s'", command, response );
	}

	motor->raw_position.stepper = motor_steps;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_set_position()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	MX_RECORD **motor_array;
	long motor_steps, other_motor_steps;
	char command[100];
	char buffer[50];
	int i, length, command_buffer_left;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	i = compumotor->controller_number - 1;

	motor_array = compumotor_interface->motor_array[i];

	motor_steps = motor->raw_setpoint.stepper;

	/* Construct a PSET command. */

	sprintf( command, "%d_!PSET", compumotor->controller_number );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {
		if ( i != 0 ) {
			length = strlen( command );
			command_buffer_left = sizeof(buffer) - length - 1;
			strncat( command, ",", command_buffer_left );
		}
		if ( i+1 == compumotor->axis_number ) {
			sprintf( buffer, "%ld", motor_steps );
		} else {
			if ( motor_array[i] == (MX_RECORD *) NULL ) {
				strcpy( buffer, "0" );
			} else {
				status = mx_motor_get_position_steps(
					motor_array[i], &other_motor_steps );

				if ( status.code != MXE_SUCCESS )
					return status;

				sprintf( buffer, "%ld", other_motor_steps );
			}
		}
		length = strlen( command );
		command_buffer_left = sizeof(buffer) - length - 1;

		strncat( command, buffer, command_buffer_left );
	}

	/* Send the PSET command. */

	status = mxi_compumotor_command( compumotor_interface, command,
			NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_soft_abort()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	int i, length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	/* Construct a stop command that only stops this axis. */

	sprintf( command, "%d_!S", compumotor->controller_number );

	length = strlen( command );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		if ( (i+1) == compumotor->axis_number ) {
			command[ length + i ] = '1';
		} else {
			command[ length + i ] = '0';
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Send the stop command. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_immediate_abort()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	/* Send a kill motion command that halts all axes attached to
	 * this controller.
	 */

	sprintf( command, "%d_!K", compumotor->controller_number );

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

MX_EXPORT mx_status_type
mxd_compumotor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_positive_limit_hit()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	char response[80];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );

	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	/* When constructing bit indices for the output of the TAS
	 * command, remember that the output of TAS also has
	 * embedded underscores in it that you must skip over too.
	 */

	if ( ( response[17] == '1' ) || ( response[20] == '1' ) ) {
		motor->positive_limit_hit = TRUE;
	} else {
		motor->positive_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_negative_limit_hit()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	char response[80];
	int length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf( command, "%d_!%dTAS", compumotor->controller_number,
					compumotor->axis_number );


	status = mxi_compumotor_command( compumotor_interface, command,
			response, sizeof response, COMPUMOTOR_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	length = strlen( response );

	if ( length <= 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
		"Zero length response to motor status command.");
	}

	/* When constructing bit indices for the output of the TAS
	 * command, remember that the output of TAS also has
	 * embedded underscores in it that you must skip over too.
	 */

	if ( ( response[18] == '1' ) || ( response[21] == '1' ) ) {
		motor->negative_limit_hit = TRUE;
	} else {
		motor->negative_limit_hit = FALSE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_compumotor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_compumotor_find_home_position()";

	MX_RECORD *compumotor_interface_record;
	MX_COMPUMOTOR_INTERFACE *compumotor_interface;
	MX_COMPUMOTOR *compumotor;
	char command[20];
	int i, length;
	mx_status_type status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_MOTOR pointer passed is NULL." );
	}

	compumotor = (MX_COMPUMOTOR *) (motor->record->record_type_struct);

	if ( compumotor == (MX_COMPUMOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_COMPUMOTOR pointer for motor is NULL." );
	}

	compumotor_interface_record = (MX_RECORD *)
			(compumotor->compumotor_interface_record);

	if ( compumotor_interface_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Interface record pointer for Compumotor motor '%s' is NULL.",
			motor->record->name );
	}

	compumotor_interface = (MX_COMPUMOTOR_INTERFACE *)
			(compumotor_interface_record->record_type_struct);

	if ( compumotor_interface == (MX_COMPUMOTOR_INTERFACE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_COMPUMOTOR_INTERFACE pointer for Compumotor motor record '%s' is NULL.",
			motor->record->name );
	}

	/* Construct the home search command. */

	sprintf( command, "%d_!HOM", compumotor->controller_number );

	length = strlen( command );

	for ( i = 0; i < MX_MAX_COMPUMOTOR_AXES; i++ ) {

		if ( (i+1) == compumotor->axis_number ) {
			switch( motor->home_search ) {
			case 1:
				command[ length + i ] = '0';
				break;
			case (-1):
				command[ length + i ] = '1';
				break;
			default:
				command[ length + i ] = 'X';
			}
		} else {
			command[ length + i ] = 'X';
		}
	}

	command[ length + MX_MAX_COMPUMOTOR_AXES ] = '\0';

	/* Command the home search to start. */

	status = mxi_compumotor_command( compumotor_interface, command,
					NULL, 0, COMPUMOTOR_DEBUG );

	return status;
}

