/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_epmtr.c 
 *
 * Purpose: MX motor driver for the EPICS APS-XFD developed motor record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "mx_aout.h"
#include "d_epmtr.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_motor_record_function_list = {
	mxd_epics_motor_initialize_type,
	NULL,
	NULL,
	mxd_epics_motor_create_record_structures,
	mxd_epics_motor_finish_record_initialization,
	mxd_epics_motor_delete_record,
	mxd_epics_motor_print_structure,
	NULL,
	mxd_epics_motor_read_parms_from_hardware,
	mxd_epics_motor_write_parms_to_hardware,
	mxd_epics_motor_open,
	mxd_epics_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_epics_motor_motor_function_list = {
	mxd_epics_motor_motor_is_busy,
	mxd_epics_motor_move_absolute,
	mxd_epics_motor_get_position,
	mxd_epics_motor_set_position,
	mxd_epics_motor_soft_abort,
	mxd_epics_motor_immediate_abort,
	mxd_epics_motor_positive_limit_hit,
	mxd_epics_motor_negative_limit_hit,
	mxd_epics_motor_find_home_position
};

/* EPICS motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_EPICS_MOTOR_STANDARD_FIELDS
};

long mxd_epics_motor_num_record_fields
		= sizeof( mxd_epics_motor_record_field_defaults )
			/ sizeof( mxd_epics_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_motor_rfield_def_ptr
			= &mxd_epics_motor_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxd_epics_motor_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	epics_motor = (MX_EPICS_MOTOR *) malloc( sizeof(MX_EPICS_MOTOR) );

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = epics_motor;
	record->class_specific_function_list
				= &mxd_epics_motor_motor_function_list;

	motor->record = record;

	/* An EPICS motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_epics_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_print_structure()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	epics_motor = (MX_EPICS_MOTOR *) (record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type      = EPICS_MOTOR.\n\n");

	fprintf(file, "  name            = %s\n", record->name);
	fprintf(file, "  EPICS record    = %s\n",
					epics_motor->epics_record_name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf(file, "  position        = %g %s  (%g).\n",
		motor->position, motor->units, 
		motor->raw_position.analog );
	fprintf(file, "  scale           = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf(file, "  offset          = %g %s.\n",
		motor->offset, motor->units );
	fprintf(file, "  backlash        = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf(file, "  negative limit  = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf(file, "  positive limit  = %g %s  (%g).\n",
	    motor->positive_limit, motor->units,
	    motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf(file, "  deadband        = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_motor_open()";

	MX_MOTOR *motor;
	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	char epics_device_type[80];
	int ezca_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer for record '%s' is NULL.",
			record->name);
	}

	epics_motor = (MX_EPICS_MOTOR *) (record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			record->name );
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for its EPICS device type.
	 */

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DTYP" );

	ezca_status = ezcaGet( pvname, ezcaString, 1, epics_device_type );

	MX_DEBUG( 2, ("(%s) epics device type = '%s'",
		epics_motor->epics_record_name, epics_device_type));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Error getting EPICS device type for MX record '%s'. EZCA status = %d (%s)",
			record->name, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_close( MX_RECORD *record )
{
	/* We need do nothing special here.  The connection to EPICS
	 * will go away automatically when the program exits.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_motor_is_busy()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	short done_moving_field;
	int ezca_status;

	motor->busy = FALSE;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DMOV" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &done_moving_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read motion status of motor '%s' failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}


	switch ( done_moving_field ) {
	case 0:
		motor->busy = TRUE;
		break;
	case 1:
		motor->busy = FALSE;
		break;
	default:
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Done moving field (.DMOV) for motor %s had an unexpected value of %d",
			epics_motor->epics_record_name, done_moving_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_move_absolute()";

	MX_EPICS_MOTOR *epics_motor;
	double new_position;
	char pvname[60];
	chid *motor_channel_id;
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status, ca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	new_position = motor->raw_setpoint.analog;

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".VAL" );

#if 0		/* use ezcaPut() */

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &new_position );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to move EPICS motor '%s' failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

#else		/* don't use ezcaPut() */

	/* A ezcaPut() call here would attempt to synchronously wait for
	 * the end of the motor move, which is not what we want at all.
	 * We must dive down into raw Channel Access to be able to start
	 * a motor without waiting for the move to complete.
	 */

	ezca_status = ezcaPvToChid( pvname, &motor_channel_id );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get EPICS channel id for '%s' failed. EZCA status = %d (%s)",
			pvname, ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Send the move command to Channel Access. */

	ca_status = ca_put( DBR_DOUBLE, *motor_channel_id, &new_position );

	if ( ca_status != ECA_NORMAL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Attempt to move motor '%s' via EPICS CA failed. CA status = %d",
			epics_motor->epics_record_name, ca_status );
	}

	/* Flush out all pending IO requests in Channel Access to ensure
	 * that the motor starts moving as soon as possible.
	 */

	ca_status = ca_flush_io();

	if ( ca_status != ECA_NORMAL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Attempt to send buffered Channel Access commands for motor '%s' failed. "
"CA status = %d",
			epics_motor->epics_record_name, ca_status );
	}

#endif		/* ezcaPut() */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_get_position()";

	MX_EPICS_MOTOR *epics_motor;
	double present_value;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".RBV" );

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &present_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to get position of EPICS motor %s failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	motor->raw_position.analog = present_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_set_position( MX_MOTOR *motor )
{
	/* Setting the motor position is a three step process:
	 * 
	 * 1.  .SET <-- 1
	 * 2.  .VAL <-- new value
	 * 3.  .SET <-- 0
	 */

	const char fname[] = "mxd_epics_motor_set_position()";

	MX_EPICS_MOTOR *epics_motor;
	double user_position;
	short set_flag;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	user_position = motor->raw_setpoint.analog;

	/*** Change the motor record from 'Use' to 'Set' mode. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SET" );

	set_flag = 1;

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &set_flag );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to change EPICS motor '%s' to Set mode failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/*** Change the user position in the EPICS database. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-6 );
	strcat( pvname, ".VAL" );

	ezca_status = ezcaPut(pvname, ezcaDouble, 1, &user_position);

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set user position for EPICS motor '%s' failed.  "
"EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/*** Put the motor record back into 'Use' mode. ***/

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SET" );

	set_flag = 0;

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &set_flag );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to change EPICS motor '%s' to Use mode failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

/* mxd_epics_motor_soft_abort() stops the motor in a way that
 * allows the motor to immediately be moved again without taking
 * any special effort.
 */

MX_EXPORT mx_status_type
mxd_epics_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_soft_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short stop_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".STOP" );

	/* Stop the motor. */

	stop_field = 1;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &stop_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to soft abort EPICS motor '%s' failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* The .STOP field is automatically reset to 0 by the IOC database,
	 * so the motor is immediately ready to move again.
	 */

	return MX_SUCCESSFUL_RESULT;
}

/* mxd_epics_motor_immediate_abort() stops the motor in a way that
 * requires manual intervention (for example, via an MEDM screen)
 * to reset the .SMPG field back to "Go".
 */

MX_EXPORT mx_status_type
mxd_epics_motor_immediate_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_immediate_abort()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short smpg_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".SPMG" );

	/* Stop the motor. */

	smpg_field = 0;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &smpg_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to immediately abort EPICS motor '%s' failed. EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_positive_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_positive_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short hls_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".HLS" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &hls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check positive limit for EPICS motor '%s' failed. "
	"EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( hls_field == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_negative_limit_hit( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_negative_limit_hit()";

	MX_EPICS_MOTOR *epics_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short lls_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	epics_motor = (MX_EPICS_MOTOR *) (motor->record->record_type_struct);

	if ( epics_motor == (MX_EPICS_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strncpy( pvname, epics_motor->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".LLS" );

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &lls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check negative limit for EPICS motor '%s' failed. "
	"EZCA status = %d (%s)",
			epics_motor->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( lls_field == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_epics_motor_find_home_position()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
	"'find home position' is not yet implemented for an EPICS motor." );
}

#endif /* HAVE_EPICS */

