/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_epscal.c
 *
 * Purpose: MX scaler driver for the EPICS APS-XFD developed scaler record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_scalr.h"
#include "d_epscal.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the scaler driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_scaler_record_function_list = {
	mxd_epics_scaler_initialize_type,
	NULL,
	NULL,
	mxd_epics_scaler_create_record_structures,
	mxd_epics_scaler_finish_record_initialization,
	mxd_epics_scaler_delete_record,
	mxd_epics_scaler_print_structure,
	NULL,
	mxd_epics_scaler_read_parms_from_hardware,
	mxd_epics_scaler_write_parms_to_hardware,
	mxd_epics_scaler_open,
	mxd_epics_scaler_close
};

MX_SCALER_FUNCTION_LIST mxd_epics_scaler_scaler_function_list = {
	mxd_epics_scaler_clear,
	mxd_epics_scaler_overflow_set,
	mxd_epics_scaler_read,
	mxd_epics_scaler_read_and_clear,
	mxd_epics_scaler_write
};

/* EPICS scaler data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_scaler_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_SCALER_STANDARD_FIELDS,
	MXD_EPICS_SCALER_STANDARD_FIELDS
};

long mxd_epics_scaler_num_record_fields
		= sizeof( mxd_epics_scaler_record_field_defaults )
		  / sizeof( mxd_epics_scaler_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_scaler_rfield_def_ptr
			= &mxd_epics_scaler_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_epics_scaler_initialize_type( long type )
{
	/* Nothing here yet. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_create_record_structures()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;

	/* Allocate memory for the necessary structures. */

	scaler = (MX_SCALER *) malloc( sizeof(MX_SCALER) );

	if ( scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCALER structure." );
	}

	epics_scaler = (MX_EPICS_SCALER *)
				malloc( sizeof(MX_EPICS_SCALER) );

	if ( epics_scaler == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_SCALER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = scaler;
	record->record_type_struct = epics_scaler;
	record->class_specific_function_list
			= &mxd_epics_scaler_scaler_function_list;

	scaler->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_finish_record_initialization( MX_RECORD *record )
{
	const char fname[]
		= "mxd_epics_scaler_finish_record_initialization()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	int scaler_number;

	scaler = (MX_SCALER *)( record->record_class_struct );

	epics_scaler = (MX_EPICS_SCALER *)( record->record_type_struct );

	/* Check that the scaler number is valid. */

	scaler_number = epics_scaler->scaler_number;

	if ( scaler_number < 1 || scaler_number > 16 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"EPICS scaler channel number %d is out of allowed range 2-16",
			scaler_number );
	}

	if ( scaler_number == 1 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"EPICS scaler channel 1 is reserved for timer use "
		"and cannot be used as a normal scaler.  Use channels 2-16" );
	}

	/* EPICS scalers don't seem to have a way of testing whether
	 * their value has overflowed.
	 */

	scaler->automatic_overflow_check = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_print_structure()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	long current_value;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SCALER pointer for record '%s' is NULL.", record->name);
	}

	epics_scaler = (MX_EPICS_SCALER *) (record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_EPICS_SCALER pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "SCALER parameters for scaler '%s':\n", record->name);

	fprintf(file, "  Scaler type           = EPICS_SCALER.\n\n");
	fprintf(file, "  EPICS record          = %s\n",
					epics_scaler->epics_record_name);
	fprintf(file, "  scaler number         = %d\n",
					epics_scaler->scaler_number);

	status = mx_scaler_read( record, &current_value );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read value of scaler '%s'",
			record->name );
	}

	fprintf(file, "  present value         = %ld\n", current_value);
	fprintf(file, "  dark current          = %g counts per second.\n",
						scaler->dark_current);
	fprintf(file, "  subtract_dark_current = %d\n",
						scaler->subtract_dark_current);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_scaler_open()";

	MX_SCALER *scaler;
	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	char epics_device_type[80];
	int ezca_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	scaler = (MX_SCALER *) (record->record_class_struct);

	if ( scaler == (MX_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCALER pointer for record '%s' is NULL.",
			record->name);
	}

	epics_scaler = (MX_EPICS_SCALER *) (record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_EPICS_SCALER pointer for scaler record '%s' is NULL.",
			record->name );
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for its EPICS device type.
	 */

	strncpy( pvname, epics_scaler->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DTYP" );

	ezca_status = ezcaGet( pvname, ezcaString, 1, epics_device_type );

	MX_DEBUG( 2, ("(%s) epics device type = '%s'",
		epics_scaler->epics_record_name, epics_device_type));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		    "Error getting EPICS device type. EZCA status = %d (%s)",
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_close( MX_RECORD *record )
{
	/* Nothing to do here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_clear( MX_SCALER *scaler )
{
#if 1
	/* It appears that it is not possible to directly clear
	 * an EPICS scaler channel.  So all we can do is return
	 * a success code.
	 */

	scaler->value = 0L;

	return MX_SUCCESSFUL_RESULT;
#else
	const char fname[] = "mxd_epics_scaler_clear()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long scaler_value;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *) (scaler->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_SCALER pointer is NULL.");
	}

	sprintf( pvname, "%s.S%d", epics_scaler->epics_record_name,
				epics_scaler->scaler_number );

	scaler_value = 0L;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &scaler_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to clear EPICS scaler '%s' channel %d failed. EZCA status = %d (%s)",
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	scaler->value = 0L;

	return MX_SUCCESSFUL_RESULT;
#endif
}

MX_EXPORT mx_status_type
mxd_epics_scaler_overflow_set( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_overflow_set()";

	scaler->overflow_set = 0;

	return mx_error( MXE_UNSUPPORTED, fname,
		"This type of scaler cannot check for overflow set." );
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_read()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long value;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *)(scaler->record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_SCALER pointer is NULL.");
	}

	sprintf( pvname, "%s.S%d", epics_scaler->epics_record_name,
				epics_scaler->scaler_number );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read from EPICS scaler '%s' channel %d failed. "
"EZCA status = %d (%s)",
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		scaler->value = 0L;

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	scaler->value = value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_read_and_clear( MX_SCALER *scaler )
{
	mx_status_type status;

	status = mxd_epics_scaler_read( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxd_epics_scaler_clear( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_scaler_write( MX_SCALER *scaler )
{
	const char fname[] = "mxd_epics_scaler_write()";

	MX_EPICS_SCALER *epics_scaler;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long value;
	int ezca_status;

	epics_scaler = (MX_EPICS_SCALER *)(scaler->record->record_type_struct);

	if ( epics_scaler == (MX_EPICS_SCALER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"EPICS_SCALER pointer is NULL.");
	}

	value = scaler->value;

	sprintf( pvname, "%s.S%d", epics_scaler->epics_record_name,
				epics_scaler->scaler_number );

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to write to EPICS scaler '%s' channel %d failed. "
"EZCA status = %d (%s)",
			epics_scaler->epics_record_name,
			epics_scaler->scaler_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */

