/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_eptime.c
 *
 * Purpose: MX timer driver for the EPICS APS-XFD developed scaler record.
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_timer.h"
#include "d_eptime.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the timer driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_epics_timer_record_function_list = {
	mxd_epics_timer_initialize_type,
	NULL,
	NULL,
	mxd_epics_timer_create_record_structures,
	mxd_epics_timer_finish_record_initialization,
	mxd_epics_timer_delete_record,
	NULL,
	NULL,
	mxd_epics_timer_read_parms_from_hardware,
	mxd_epics_timer_write_parms_to_hardware,
	mxd_epics_timer_open,
	mxd_epics_timer_close
};

MX_TIMER_FUNCTION_LIST mxd_epics_timer_timer_function_list = {
	mxd_epics_timer_is_busy,
	mxd_epics_timer_start,
	mxd_epics_timer_stop,
	mxd_epics_timer_get_current_time
};

/* EPICS timer data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_epics_timer_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_TIMER_STANDARD_FIELDS,
	MXD_EPICS_TIMER_STANDARD_FIELDS
};

long mxd_epics_timer_num_record_fields
		= sizeof( mxd_epics_timer_record_field_defaults )
		  / sizeof( mxd_epics_timer_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_epics_timer_rfield_def_ptr
			= &mxd_epics_timer_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_epics_timer_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_timer_create_record_structures()";

	MX_TIMER *timer;
	MX_EPICS_TIMER *epics_timer;

	/* Allocate memory for the necessary structures. */

	timer = (MX_TIMER *) malloc( sizeof(MX_TIMER) );

	if ( timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TIMER structure." );
	}

	epics_timer = (MX_EPICS_TIMER *)
				malloc( sizeof(MX_EPICS_TIMER) );

	if ( epics_timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_EPICS_TIMER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = timer;
	record->record_type_struct = epics_timer;
	record->class_specific_function_list
			= &mxd_epics_timer_timer_function_list;

	timer->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_finish_record_initialization( MX_RECORD *record )
{
	return mx_timer_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxd_epics_timer_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_read_parms_from_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_epics_timer_read_parms_from_hardware() called."));

	/* No module state to save, so there is nothing to read. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_write_parms_to_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_epics_timer_write_parms_to_hardware() called."));

	/* No module state to restore. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_open( MX_RECORD *record )
{
	const char fname[] = "mxd_epics_timer_open()";

	MX_TIMER *timer;
	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[150];
	char epics_device_type[80];
	int ezca_status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	timer = (MX_TIMER *) (record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TIMER pointer for record '%s' is NULL.",
			record->name);
	}

	epics_timer = (MX_EPICS_TIMER *) (record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_EPICS_TIMER pointer for timer record '%s' is NULL.",
			record->name );
	}

	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	/* Here we find out if this EPICS record really exists in an
	 * IOC's database by asking for its EPICS device type.
	 */

	strncpy( pvname, epics_timer->epics_record_name, sizeof(pvname)-5 );
	strcat( pvname, ".DTYP" );

	ezca_status = ezcaGet( pvname, ezcaString, 1, epics_device_type );

	MX_DEBUG( 2, ("(%s) epics device type = '%s'",
		epics_timer->epics_record_name, epics_device_type));

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
		    "Error getting EPICS device type. EZCA status = %d (%s)",
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_close( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_epics_timer_close() invoked."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_is_busy( MX_TIMER *timer, int *busy )
{
	const char fname[] = "mxd_epics_timer_is_busy()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	long count_field;
	int ezca_status;

	epics_timer = (MX_EPICS_TIMER *) (timer->record->record_type_struct);

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	ezca_status = ezcaGet( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to read busy status of EPICS timer '%s' failed. "
"EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	switch ( count_field ) {
	case 0:
		*busy = FALSE;
		break;
	case 1:
		*busy = TRUE;
		break;
	default:
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Count field (.CNT) for scaler '%s' had an unexpected value of %d",
			epics_timer->epics_record_name, count_field );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_start( MX_TIMER *timer, double seconds )
{
	const char fname[] = "mxd_epics_timer_start()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long count_field;
#if 0
	long counter_mode;
#endif

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

#if 0
	/* Make sure the counter is in one-shot mode. */

	sprintf( pvname, "%s_mode.VAL", epics_timer->epics_record_name );

	counter_mode = 0;	/* One-shot mode. */

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &seconds );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set one-shot mode for EPICS counter '%s' channel %d failed. "
"EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			epics_timer->timer_number,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}
#endif

	/* Set the number of seconds to count for.  This assumes that
	 * the .FREQ field has already been set up correctly.
	 */

	sprintf( pvname, "%s.TP", epics_timer->epics_record_name );

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &seconds );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to set counting time for EPICS timer '%s' failed. "
"EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	/* Start the timer counting. */

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	count_field = 1;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to start EPICS timer '%s' failed. EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_stop( MX_TIMER *timer, double *seconds_left )
{
	const char fname[] = "mxd_epics_timer_stop()";

	MX_EPICS_TIMER *epics_timer;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	int ezca_status;
	long count_field;

	epics_timer = (MX_EPICS_TIMER *) ( timer->record->record_type_struct );

	if ( epics_timer == (MX_EPICS_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_EPICS_TIMER structure pointer is NULL.");
	}

	/* Stop the timer from counting. */

	sprintf( pvname, "%s.CNT", epics_timer->epics_record_name );

	count_field = 0;

	ezca_status = ezcaPut( pvname, ezcaLong, 1, &count_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
"Attempt to stop EPICS timer '%s' failed. EZCA status = %d (%s)",
			epics_timer->epics_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_epics_timer_get_current_time( MX_TIMER *timer, double *seconds )
{
	const char fname[] = "mxd_epics_timer_get_current_time()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"The current elapsed time cannot be read from this timer." );
}

#endif /* HAVE_EPICS */

