/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_pdi45.c
 *
 * Purpose: MX drivers for the PDI Model 45 DAQ interface.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "pdi45.h"
#include "d_pdi45.h"

/* ============ Generic functions for all device types ============ */

MX_RECORD_FUNCTION_LIST mxd_pdi45_record_function_list = {
	mxd_pdi45_initialize_type,
	mxd_pdi45_read_record_description,
	mxd_pdi45_write_record_description,
	NULL,
	NULL,
	NULL,
	mxd_pdi45_print_structure,
	mxd_pdi45_print_summary,
	mxd_pdi45_read_parms_from_hardware,
	mxd_pdi45_write_parms_to_hardware,
	mxd_pdi45_open,
	mxd_pdi45_close
};

MX_EXPORT mx_status_type
mxd_pdi45_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pdi45_read_record_description( char *buffer,
			MX_RECORD *record, MX_RECORD *record_list_head )
{
	const char fname[] = "mxd_pdi45_read_record_description()";

	MX_PDI45 *pdi45;
	MX_ANALOG_INPUT *adc;
	MX_ANALOG_OUTPUT *dac;
	MX_DIGITAL_INPUT *din;
	MX_DIGITAL_OUTPUT *dout;
	MX_SCALER *scaler;
	MX_TIMER *timer;

	MX_PDI45_ADC *pdi45_adc;
	MX_PDI45_DIN *pdi45_dac;
	MX_PDI45_DIN *pdi45_din;
	MX_PDI45_DOUT *pdi45_dout;
	MX_PDI45_SCALER *pdi45_scaler;
	MX_PDI45_TIMER *pdi45_timer;

	MX_RECORD *pdi45_record;
	int num_tokens;
	char format_string[50];
	char pdi45_name[MXU_RECORD_NAME_LENGTH + 1];
	int channel;

	MX_DEBUG(-2, ("mxd_pdi45_read_record_description() invoked."));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"Device record pointer is NULL." );
	}

	pdi45 = NULL;

	switch( record->type ) {
	case MXC_ANALOG_INPUT:
		adc = (MX_ANALOG_INPUT *) (record->record_class_struct);
		if ( adc == (MX_ANALOG_INPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_ANALOG_INPUT structure in the record passed was not initialized.");
		}
		if ( adc->record->type != MXT_AIN_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
		"The MX_ANALOG_INPUT structure is not for a PDI45 ADC.");
		}
		break;
	case MXC_ANALOG_OUTPUT:
		dac = (MX_ANALOG_OUTPUT *) (record->record_class_struct);
		if ( dac == (MX_ANALOG_OUTPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_ANALOG_OUTPUT structure in the record passed was not initialized.");
		}
		if ( dac->record->type != MXT_AOU_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
		"The MX_ANALOG_OUTPUT structure is not for a PDI45 DAC.");
		}
		break;
	case MXC_DIGITAL_INPUT:
		din = (MX_DIGITAL_INPUT *) (record->record_class_struct);
		if ( din == (MX_DIGITAL_INPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_DIGITAL_INPUT structure in the record passed was not initialized.");
		}
		if ( din->record->type != MXT_DIN_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
	"The MX_DIGITAL_INPUT structure is not for a PDI45 digital input.");
		}
		break;
	case MXC_DIGITAL_OUTPUT:
		dout = (MX_DIGITAL_OUTPUT *) (record->record_class_struct);
		if ( dout == (MX_DIGITAL_OUTPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_DIGITAL_OUTPUT structure in the record passed was not initialized.");
		}
		if ( dout->record->type != MXT_DOU_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
	"The MX_DIGITAL_OUTPUT structure is not for a PDI45 digital output.");
		}
		break;
	case MXC_SCALER:
		scaler = (MX_SCALER *) (record->record_class_struct);
		if ( scaler == (MX_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_SCALER structure in the record passed was not initialized.");
		}
		if ( scaler->record->type != MXT_SCL_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
	"The MX_SCALER structure is not for a PDI45 scaler.");
		}
		break;
	case MXC_TIMER:
		timer = (MX_TIMER *) (record->record_class_struct);
		if ( timer == (MX_TIMER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"The MX_TIMER structure in the record passed was not initialized.");
		}
		if ( timer->record->type != MXT_TIM_PDI45 ) {
			return mx_error( MXE_TYPE_MISMATCH, fname,
	"The MX_TIMER structure is not for a PDI45 timer.");
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The device '%s' had an unexpected record type = %d",
			record->name, record->type );
		break;
	}

	/* Scan the rest of the savefile string. */

	sprintf(format_string,
		"%%*s %%*s %%*s %%*s %%%ds %%d",
		MXU_RECORD_NAME_LENGTH);

	num_tokens = sscanf( buffer, format_string,
			&pdi45_name[0], &channel );

	if ( num_tokens != 2 ) {
		return mx_error( MXE_UNPARSEABLE_STRING, fname,
		"Unparseable savefile record.  Line contents = '%s'",
			buffer );
	}

	MX_DEBUG(-2, ("Record name = '%s', PDI45 name = '%s', channel = %d",
		record->name, pdi45_name, channel) );

	/* Find the first record with the same name as the listed
	 * PDI45 interface name.  Currently forward references are not
	 * allowed.
	 */

	pdi45_record = mx_get_record( record_list_head, pdi45_name );

	if ( pdi45_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_UNDEFINED_RECORD, fname,
		"The PDI45 interface '%s' has not yet been declared.",
			pdi45_name );
	}

	/* --- Verify that this record is a PDI Model 45 interface. --- */

	if ( pdi45_record->superclass != MXR_INTERFACE ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"'%s' is not an interface record.", pdi45_name );
	}

#if 0
	if ( pdi45_record->class != MXI_PDI45 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"'%s' is not a PDI Model 45.", pdi45_name );
	}
#endif

	/* Now fill in the structures for the individual devices types. */

	switch( record->type ) {
	case MXC_ANALOG_INPUT:
		pdi45_adc = (MX_PDI45_ADC *) malloc( sizeof(MX_PDI45_ADC) );
		if ( pdi45_adc == (MX_PDI45_ADC *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_ADC structure.");
		}

		record->record_type_struct = pdi45_adc;
		pdi45_adc->pdi45_record = pdi45_record;
		pdi45_adc->pdi45 = pdi45;
		pdi45_adc->channel = channel;

		break;
		
	case MXC_ANALOG_OUTPUT:
		pdi45_dac = (MX_PDI45_DIN *) malloc( sizeof(MX_PDI45_DIN) );
		if ( pdi45_dac == (MX_PDI45_DIN *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_DIN structure.");
		}

		record->record_type_struct = pdi45_dac;
		pdi45_dac->pdi45_record = pdi45_record;
		pdi45_dac->pdi45 = pdi45;
		pdi45_dac->channel = channel;

		break;
		
	case MXC_DIGITAL_INPUT:
		pdi45_din = (MX_PDI45_DIN *) malloc( sizeof(MX_PDI45_DIN) );
		if ( pdi45_din == (MX_PDI45_DIN *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_DIN structure.");
		}

		record->record_type_struct = pdi45_din;
		pdi45_din->pdi45_record = pdi45_record;
		pdi45_din->pdi45 = pdi45;
		pdi45_din->channel = channel;

		break;
		
	case MXC_DIGITAL_OUTPUT:
		pdi45_dout = (MX_PDI45_DOUT *) malloc( sizeof(MX_PDI45_DOUT) );
		if ( pdi45_dout == (MX_PDI45_DOUT *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_DOUT structure.");
		}

		record->record_type_struct = pdi45_dout;
		pdi45_dout->pdi45_record = pdi45_record;
		pdi45_dout->pdi45 = pdi45;
		pdi45_dout->channel = channel;

		break;
		
	case MXC_SCALER:
		pdi45_scaler = (MX_PDI45_SCALER *)
					malloc( sizeof(MX_PDI45_SCALER) );
		if ( pdi45_scaler == (MX_PDI45_SCALER *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_SCALER structure.");
		}

		record->record_type_struct = pdi45_scaler;
		pdi45_scaler->pdi45_record = pdi45_record;
		pdi45_scaler->pdi45 = pdi45;
		pdi45_scaler->channel = channel;

		break;
		
	case MXC_TIMER:
		pdi45_timer = (MX_PDI45_TIMER *)
					malloc( sizeof(MX_PDI45_TIMER) );
		if ( pdi45_timer == (MX_PDI45_TIMER *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory trying to allocate an MX_PDI45_TIMER structure.");
		}

		record->record_type_struct = pdi45_timer;
		pdi45_timer->pdi45_record = pdi45_record;
		pdi45_timer->pdi45 = pdi45;
		pdi45_timer->channel = channel;

		break;
	}	
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pdi45_write_record_description( MX_RECORD *record, char *buffer )
{
	const char fname[] = "mxd_pdi45_write_record_description()";

	MX_ANALOG_INPUT *adc;
	MX_ANALOG_OUTPUT *dac;
	MX_DIGITAL_INPUT *din;
	MX_DIGITAL_OUTPUT *dout;
	MX_SCALER *scaler;
	MX_TIMER *timer;

	MX_PDI45_ADC *pdi45_adc;
	MX_PDI45_DIN *pdi45_dac;
	MX_PDI45_DIN *pdi45_din;
	MX_PDI45_DOUT *pdi45_dout;
	MX_PDI45_SCALER *pdi45_scaler;
	MX_PDI45_TIMER *pdi45_timer;

	char format_string[50];

	MX_DEBUG(-2, ("mxd_pdi45_write_record_description() invoked."));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"Device record pointer is NULL." );
	}

	switch( record->type ) {
	case MXC_ANALOG_INPUT:
		sprintf( format_string,
		"%%-%ds device analog_input pdi45_adc %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		adc = (MX_ANALOG_INPUT *) (record->record_class_struct);
		if ( adc == (MX_ANALOG_INPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_ANALOG_INPUT pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_adc = (MX_PDI45_ADC *) (record->record_type_struct);
		if ( pdi45_adc == (MX_PDI45_ADC *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_ADC pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_adc->pdi45_record->name,
			pdi45_adc->channel );
		break;

	case MXC_ANALOG_OUTPUT:
		sprintf( format_string,
		"%%-%ds device analog_output pdi45_dac %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		dac = (MX_ANALOG_OUTPUT *) (record->record_class_struct);
		if ( dac == (MX_ANALOG_OUTPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_ANALOG_OUTPUT pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_dac = (MX_PDI45_DIN *) (record->record_type_struct);
		if ( pdi45_dac == (MX_PDI45_DIN *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_DIN pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_dac->pdi45_record->name,
			pdi45_dac->channel );
		break;

	case MXC_DIGITAL_INPUT:
		sprintf( format_string,
		"%%-%ds device digital_input pdi45_din %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		din = (MX_DIGITAL_INPUT *) (record->record_class_struct);
		if ( din == (MX_DIGITAL_INPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_DIGITAL_INPUT pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_din = (MX_PDI45_DIN *) (record->record_type_struct);
		if ( pdi45_din == (MX_PDI45_DIN *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_DIN pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_din->pdi45_record->name,
			pdi45_din->channel );
		break;

	case MXC_DIGITAL_OUTPUT:
		sprintf( format_string,
		"%%-%ds device digital_output pdi45_dout %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		dout = (MX_DIGITAL_OUTPUT *) (record->record_class_struct);
		if ( dout == (MX_DIGITAL_OUTPUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_DIGITAL_OUTPUT pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_dout = (MX_PDI45_DOUT *) (record->record_type_struct);
		if ( pdi45_dout == (MX_PDI45_DOUT *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_DOUT pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_dout->pdi45_record->name,
			pdi45_dout->channel );
		break;

	case MXC_SCALER:
		sprintf( format_string,
		"%%-%ds device scaler pdi45_scaler %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		scaler = (MX_SCALER *) (record->record_class_struct);
		if ( scaler == (MX_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCALER pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_scaler = (MX_PDI45_SCALER *) (record->record_type_struct);
		if ( pdi45_scaler == (MX_PDI45_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_SCALER pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_scaler->pdi45_record->name,
			pdi45_scaler->channel );
		break;

	case MXC_TIMER:
		sprintf( format_string,
		"%%-%ds device timer pdi45_timer %%-%ds %%d",
		MXU_RECORD_NAME_LENGTH, MXU_RECORD_NAME_LENGTH );

		timer = (MX_TIMER *) (record->record_class_struct);
		if ( timer == (MX_TIMER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_TIMER pointer for record '%s' is NULL",
			record->name);
		}

		pdi45_timer = (MX_PDI45_TIMER *) (record->record_type_struct);
		if ( pdi45_timer == (MX_PDI45_TIMER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PDI45_TIMER pointer for record '%s' is NULL",
			record->name);
		}
		sprintf( buffer, format_string,
			record->name, pdi45_timer->pdi45_record->name,
			pdi45_timer->channel );
		break;

	default:
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Unexpected record type %d found for record '%s'",
			record->type, record->name );
		break;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pdi45_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_print_structure()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45_print_summary( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_print_summary()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45_read_parms_from_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_read_parms_from_hardware()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45_write_parms_to_hardware( MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_write_parms_to_hardware()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45_open( MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_open()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45_close( MX_RECORD *record )
{
	const char fname[] = "mxd_pdi45_close()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Analog input channels ============ */

MX_ANALOG_INPUT_FUNCTION_LIST mxd_pdi45adc_analog_input_function_list = {
	mxd_pdi45adc_read
};

MX_EXPORT mx_status_type
mxd_pdi45adc_read( MX_ANALOG_INPUT *adc )
{
	const char fname[] = "mxd_pdi45adc_read()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Analog output channels ============ */

MX_ANALOG_OUTPUT_FUNCTION_LIST mxd_pdi45dac_analog_output_function_list = {
	mxd_pdi45dac_read,
	mxd_pdi45dac_write
};

MX_EXPORT mx_status_type
mxd_pdi45dac_read( MX_ANALOG_OUTPUT *dac )
{
	const char fname[] = "mxd_pdi45dac_read()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45dac_write( MX_ANALOG_OUTPUT *dac )
{
	const char fname[] = "mxd_pdi45dac_write()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Digital input channels ============ */

MX_DIGITAL_INPUT_FUNCTION_LIST mxd_pdi45din_digital_input_function_list = {
	mxd_pdi45din_read
};

MX_EXPORT mx_status_type
mxd_pdi45din_read( MX_DIGITAL_INPUT *dinput )
{
	const char fname[] = "mxd_pdi45din_read()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Digital output channels ============ */

MX_DIGITAL_OUTPUT_FUNCTION_LIST mxd_pdi45dout_digital_output_function_list = {
	mxd_pdi45dout_read,
	mxd_pdi45dout_write
};

MX_EXPORT mx_status_type
mxd_pdi45dout_read( MX_DIGITAL_OUTPUT *doutput )
{
	const char fname[] = "mxd_pdi45dout_read()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45dout_write( MX_DIGITAL_OUTPUT *doutput )
{
	const char fname[] = "mxd_pdi45dout_write()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Scaler channels ============ */

MX_SCALER_FUNCTION_LIST mxd_pdi45scaler_scaler_function_list = {
	mxd_pdi45scaler_clear,
	mxd_pdi45scaler_overflow_set,
	mxd_pdi45scaler_read,
	mxd_pdi45scaler_read_and_clear,
	mxd_pdi45scaler_write
};

MX_EXPORT mx_status_type
mxd_pdi45scaler_clear( MX_SCALER *scaler )
{
	const char fname[] = "mxd_pdi45scaler_clear()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45scaler_overflow_set( MX_SCALER *scaler )
{
	const char fname[] = "mxd_pdi45scaler_overflow_set()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45scaler_read( MX_SCALER *scaler )
{
	const char fname[] = "mxd_pdi45scaler_read()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45scaler_read_and_clear( MX_SCALER *scaler )
{
	const char fname[] = "mxd_pdi45scaler_read_and_clear()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45scaler_write( MX_SCALER *scaler )
{
	const char fname[] = "mxd_pdi45scaler_write()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

/* ============ Timer channels ============ */

MX_TIMER_FUNCTION_LIST mxd_pdi45timer_timer_function_list = {
	mxd_pdi45timer_is_busy,
	mxd_pdi45timer_start,
	mxd_pdi45timer_stop,
	mxd_pdi45timer_get_current_time
};

MX_EXPORT mx_status_type
mxd_pdi45timer_is_busy( MX_TIMER *timer, int *busy )
{
	const char fname[] = "mxd_pdi45timer_is_busy()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45timer_start( MX_TIMER *timer, double seconds )
{
	const char fname[] = "mxd_pdi45timer_start()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45timer_stop( MX_TIMER *timer, double *seconds_left)
{
	const char fname[] = "mxd_pdi45timer_stop()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_pdi45timer_get_current_time( MX_TIMER *timer, double *seconds )
{
	const char fname[] = "mxd_pdi45timer_get_current_time()";
	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname, "Not yet implemented." );
}

