/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_pmacdp.c 
 *
 * Purpose: MX motor driver for controlling a Delta Tau PMAC motor
 *          through Tom Coleman's EPICS PMAC dual-ported RAM interface.
 *
 *          Note that this driver uses the EZCA channel access library
 *          which is packaged as an extension to EPICS and is not in
 *          the base release.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_motor.h"
#include "mx_aout.h"
#include "d_pmacdp.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "ezca.h"

/* Initialize the motor driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_pmac_dpr_motor_record_function_list = {
	mxd_pmac_dpr_motor_initialize_type,
	NULL,
	NULL,
	mxd_pmac_dpr_motor_create_record_structures,
	mxd_pmac_dpr_motor_finish_record_initialization,
	mxd_pmac_dpr_motor_delete_record,
	mxd_pmac_dpr_motor_print_structure,
	NULL,
	mxd_pmac_dpr_motor_read_parms_from_hardware,
	mxd_pmac_dpr_motor_write_parms_to_hardware,
	mxd_pmac_dpr_motor_open,
	mxd_pmac_dpr_motor_close
};

MX_MOTOR_FUNCTION_LIST mxd_pmac_dpr_motor_motor_function_list = {
	mxd_pmac_dpr_motor_motor_is_busy,
	mxd_pmac_dpr_motor_move_absolute,
	mxd_pmac_dpr_motor_get_position,
	mxd_pmac_dpr_motor_set_position,
	mxd_pmac_dpr_motor_soft_abort,
	mxd_pmac_dpr_motor_immediate_abort,
	mxd_pmac_dpr_motor_positive_limit_hit,
	mxd_pmac_dpr_motor_negative_limit_hit,
	mxd_pmac_dpr_motor_find_home_position
};

/* PMAC dual-ported RAM motor data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_pmac_dpr_motor_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_ANALOG_MOTOR_STANDARD_FIELDS,
	MX_MOTOR_STANDARD_FIELDS,
	MXD_PMAC_DPR_MOTOR_STANDARD_FIELDS
};

long mxd_pmac_dpr_motor_num_record_fields
		= sizeof( mxd_pmac_dpr_motor_record_field_defaults )
			/ sizeof( mxd_pmac_dpr_motor_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_pmac_dpr_motor_rfield_def_ptr
			= &mxd_pmac_dpr_motor_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_initialize_type( long type )
{
		/* Nothing needed here. */

		return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_dpr_motor_create_record_structures()";

	MX_MOTOR *motor;
	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;

	/* Allocate memory for the necessary structures. */

	motor = (MX_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MOTOR structure." );
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *) malloc( sizeof(MX_MOTOR) );

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_PMAC_DPR_MOTOR structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = motor;
	record->record_type_struct = pmac_dpr_motor;
	record->class_specific_function_list
				= &mxd_pmac_dpr_motor_motor_function_list;

	motor->record = record;

	/* A DAC motor is treated as an analog motor. */

	motor->subclass = MXC_MTR_ANALOG;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_finish_record_initialization( MX_RECORD *record )
{
	mx_status_type status;

	status = mx_motor_finish_record_initialization( record );

	return status;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_pmac_dpr_motor_print_structure()";

	MX_MOTOR *motor;
	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	double position, deadband;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	motor = (MX_MOTOR *) (record->record_class_struct);

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer for record '%s' is NULL.", record->name );
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *) (record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_DPR_MOTOR pointer for record '%s' is NULL.",
			record->name );
	}

	fprintf(file, "MOTOR parameters for motor '%s':\n", record->name);
	fprintf(file, "  Motor type                = PMAC_DPR_MOTOR.\n\n");

	fprintf(file, "  name                      = %s\n", record->name);
	fprintf(file, "  actual position record    = %s\n",
			pmac_dpr_motor->actual_position_record_name);
	fprintf(file, "  requested position record = %s\n",
			pmac_dpr_motor->requested_position_record_name);
	fprintf(file, "  in position record        = %s\n",
			pmac_dpr_motor->in_position_record_name);
	fprintf(file, "  abort record              = %s\n",
			pmac_dpr_motor->abort_record_name);

	status = mx_motor_get_position( record, &position );

	if ( status.code != MXE_SUCCESS ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
			"Unable to read position of motor '%s'",
			record->name );
	}

	fprintf( file, "  position                  = %g %s  (%g).\n",
		motor->position, motor->units,
		motor->raw_position.analog );
	fprintf( file, "  scale                     = %g %s per EPICS EGU.\n",
		motor->scale, motor->units );
	fprintf( file, "  offset                    = %g %s.\n",
		motor->offset, motor->units);
	fprintf( file, "  backlash                  = %g %s  (%g).\n",
		motor->backlash_correction, motor->units,
		motor->raw_backlash_correction.analog );
	fprintf( file, "  negative limit            = %g %s  (%g).\n",
		motor->negative_limit, motor->units,
		motor->raw_negative_limit.analog );
	fprintf( file, "  positive limit            = %g %s  (%g).\n",
	        motor->positive_limit, motor->units,
	        motor->raw_positive_limit.analog );

	deadband = motor->scale * motor->deadband.analog;

	fprintf( file, "  deadband                  = %g %s  (%g).\n\n",
		deadband, motor->units,
		motor->deadband.analog );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_read_parms_from_hardware( MX_RECORD *record )
{
	/* All saving of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_write_parms_to_hardware( MX_RECORD *record )
{
	/* All setting of parameters is handled by EPICS, so we need not
	 * do anything here.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_open( MX_RECORD *record )
{
	/* Turn off the automatic printing of EZCA error messages
	 * to stderr, since we want to handle that ourselves anyway.
	 */

	ezcaAutoErrorMessageOff();

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_close( MX_RECORD *record )
{
	/* We need do nothing special here.  The connection to EPICS
	 * will go away automatically when the program exits.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_motor_is_busy( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_motor_is_busy()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[200];
	short in_position_value;
	int i, max_retries;
	int ezca_status;

	motor->busy = FALSE;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *) (motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC_DPR_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strcpy( pvname, pmac_dpr_motor->in_position_record_name );

	max_retries = 9;

	for ( i = 0; i < max_retries; i++ ) {
		ezca_status = ezcaGet( pvname,
					ezcaShort, 1, &in_position_value );

		if ( ezca_status != EZCA_OK ) {
			ezcaGetErrorString( NULL, &ezca_error_message );
			sprintf( mx_error_message,
"Attempt to read motion status of PMAC DPR motor '%s' via record '%s' failed. "
" EZCA status = %d (%s)",
				motor->record->name,
				pvname, ezca_status, ezca_error_message );

			ezcaFree( (void *) ezca_error_message );

			return mx_error(MXE_DEVICE_IO_ERROR,
				fname, mx_error_message);
		}

		switch ( in_position_value ) {
		case 0:
			motor->busy = TRUE;

			return MX_SUCCESSFUL_RESULT;

			break;
		case 1:
			/* Go back and check again. */

			break;
		default:
			return mx_error( MXE_DEVICE_IO_ERROR, fname,
			"In position record '%s' for PMAC DPR motor '%s' "
			"had an unexpected value of %d",
			pvname, motor->record->name, in_position_value );

			break;
		}
	}

	motor->busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_move_absolute( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_move_absolute()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	double new_position;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed is NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *)
					(motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_DPR_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	new_position = motor->raw_setpoint.analog;

	/* Send the move command. */

	strcpy(pvname, pmac_dpr_motor->requested_position_record_name);

	ezca_status = ezcaPut( pvname, ezcaDouble, 1, &new_position );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
			"Attempt to move PMAC DPR motor '%s' via '%s' failed.  "
			"EZCA status = %d (%s)",
			motor->record->name,
			pmac_dpr_motor->requested_position_record_name,
			ezca_status, ezca_error_message );

		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_get_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_get_position()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	double scale, offset, present_value;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MOTOR pointer passed NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *)
					(motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_PMAC_DPR_MOTOR pointer for record '%s' is NULL.",
			motor->record->name );
	}

	scale = motor->scale;
	offset = motor->offset;

	strcpy(pvname, pmac_dpr_motor->actual_position_record_name);

	ezca_status = ezcaGet( pvname, ezcaDouble, 1, &present_value );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to get position of PMAC DPR motor '%s' via '%s' failed.  "
	"EZCA status = %d (%s)",
			motor->record->name,
			pmac_dpr_motor->actual_position_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	motor->raw_position.analog = present_value;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_set_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_set_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
	"'set position' is not supported for an PMAC DPR motor." );
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_soft_abort( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_soft_abort()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short stop_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *)
					(motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC_DPR_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	strcpy( pvname, pmac_dpr_motor->abort_record_name );

	/* Stop the motor move. */

	stop_field = 1;		/* Stop */

	ezca_status = ezcaPut( pvname, ezcaShort, 1, &stop_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to abort move for PMAC DPR motor '%s' via '%s' failed.  "
	"EZCA status = %d (%s)",
			motor->record->name,
			pmac_dpr_motor->abort_record_name,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_immediate_abort( MX_MOTOR *motor )
{
	return mxd_pmac_dpr_motor_soft_abort( motor );
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_positive_limit_hit( MX_MOTOR *motor )
{
#if 1
	motor->positive_limit_hit = FALSE;
#else
	const char fname[] = "mxd_pmac_dpr_motor_positive_limit_hit()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short hls_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *)
				(motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC_DPR_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf(pvname, "%s:PsLimSet", pmac_dpr_motor->epics_record_namestem);

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &hls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check positive limit for PMAC DPR motor '%s' failed. "
	"EZCA status = %d (%s)",
			pmac_dpr_motor->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( hls_field == 0 ) {
		motor->positive_limit_hit = FALSE;
	} else {
		motor->positive_limit_hit = TRUE;
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_negative_limit_hit( MX_MOTOR *motor )
{
#if 1
	motor->negative_limit_hit = FALSE;
#else
	const char fname[] = "mxd_pmac_dpr_motor_negative_limit_hit()";

	MX_PMAC_DPR_MOTOR *pmac_dpr_motor;
	char pvname[60];
	char *ezca_error_message;
	char mx_error_message[160];
	short lls_field;
	int ezca_status;

	if ( motor == (MX_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MOTOR pointer passed is NULL.");
	}

	pmac_dpr_motor = (MX_PMAC_DPR_MOTOR *)
				(motor->record->record_type_struct);

	if ( pmac_dpr_motor == (MX_PMAC_DPR_MOTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PMAC_DPR_MOTOR pointer for motor record '%s' is NULL.",
			motor->record->name );
	}

	sprintf(pvname, "%s:NgLimSet", pmac_dpr_motor->epics_record_namestem);

	ezca_status = ezcaGet( pvname, ezcaShort, 1, &lls_field );

	if ( ezca_status != EZCA_OK ) {
		ezcaGetErrorString( NULL, &ezca_error_message );
		sprintf( mx_error_message,
	"Attempt to check negative limit for PMAC DPR motor '%s' failed. "
	"EZCA status = %d (%s)",
			pmac_dpr_motor->epics_record_namestem,
			ezca_status, ezca_error_message );
		ezcaFree( (void *) ezca_error_message );

		return mx_error(MXE_DEVICE_IO_ERROR, fname, mx_error_message);
	}

	if ( lls_field == 0 ) {
		motor->negative_limit_hit = FALSE;
	} else {
		motor->negative_limit_hit = TRUE;
	}
#endif

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_pmac_dpr_motor_find_home_position( MX_MOTOR *motor )
{
	const char fname[] = "mxd_pmac_dpr_motor_find_home_position()";

	return mx_error( MXE_UNSUPPORTED, fname,
"'find home position' is not supported for an PMAC DPR motor." );
}

#endif /* HAVE_EPICS */

