/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_stimer.c
 *
 * Purpose: MX timer driver to control simple software timers.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_timer.h"
#include "d_stimer.h"

/* Initialize the timer driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_soft_timer_record_function_list = {
	mxd_soft_timer_initialize_type,
	NULL,
	NULL,
	mxd_soft_timer_create_record_structures,
	mxd_soft_timer_finish_record_initialization,
	mxd_soft_timer_delete_record,
	NULL,
	NULL,
	mxd_soft_timer_read_parms_from_hardware,
	mxd_soft_timer_write_parms_to_hardware,
	mxd_soft_timer_open,
	mxd_soft_timer_close
};

MX_TIMER_FUNCTION_LIST mxd_soft_timer_timer_function_list = {
	mxd_soft_timer_timer_is_busy,
	mxd_soft_timer_timer_start,
	mxd_soft_timer_timer_stop,
	mxd_soft_timer_timer_get_current_time
};

/* Soft timer data structures. */

MX_RECORD_FIELD_DEFAULTS mxd_soft_timer_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_TIMER_STANDARD_FIELDS
};

long mxd_soft_timer_num_record_fields
		= sizeof( mxd_soft_timer_record_field_defaults )
		  / sizeof( mxd_soft_timer_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_soft_timer_rfield_def_ptr
			= &mxd_soft_timer_record_field_defaults[0];

MX_EXPORT mx_status_type
mxd_soft_timer_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_soft_timer_create_record_structures()";

	MX_TIMER *timer;
	MX_SOFT_TIMER *soft_timer;

	/* Allocate memory for the necessary structures. */

	timer = (MX_TIMER *) malloc( sizeof(MX_TIMER) );

	if ( timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TIMER structure." );
	}

	soft_timer = (MX_SOFT_TIMER *)
				malloc( sizeof(MX_SOFT_TIMER) );

	if ( soft_timer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SOFT_TIMER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = timer;
	record->record_type_struct = soft_timer;
	record->class_specific_function_list
			= &mxd_soft_timer_timer_function_list;

	timer->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_finish_record_initialization( MX_RECORD *record )
{
	return mx_timer_finish_record_initialization( record );
}

MX_EXPORT mx_status_type
mxd_soft_timer_delete_record( MX_RECORD *record )
{
	MX_SOFT_TIMER *soft_timer;

	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	soft_timer = (MX_SOFT_TIMER *) record->record_type_struct;

	if ( soft_timer != NULL ) {

		free( soft_timer );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_read_parms_from_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_soft_timer_read_parms_from_hardware() called."));

	/* No module state to save, so there is nothing to read. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_write_parms_to_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_soft_timer_write_parms_to_hardware() called."));

	/* No module state to restore. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_open( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_soft_timer_open() invoked."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_close( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxd_soft_timer_close() invoked."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_timer_is_busy( MX_TIMER *timer, int *busy )
{
	const char fname[] = "mxd_soft_timer_timer_is_busy()";

	if ( busy == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"Pointer to busy variable is NULL.");
	}

	*busy = FALSE;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_timer_start( MX_TIMER *timer, double seconds )
{
	unsigned long sleep_seconds;
	unsigned long sleep_milliseconds;

	if ( seconds <= 0.0 )
		return MX_SUCCESSFUL_RESULT;

	sleep_seconds = (unsigned long) ( seconds );

	sleep_milliseconds = (unsigned long)
			( 1000.0 * ( seconds - (double) sleep_seconds ) );

	mx_sleep( sleep_seconds );

	mx_msleep( sleep_milliseconds );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_timer_stop( MX_TIMER *timer, double *seconds_left)
{
	*seconds_left = 0.0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_soft_timer_timer_get_current_time( MX_TIMER *timer, double *seconds )
{
	const char fname[] = "mxd_soft_timer_timer_get_current_time() called.";

	return mx_error( MXE_UNSUPPORTED, fname,
	"This timer does not support reading the current elapsed time." );
}

