/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    d_trump.c
 *
 * Purpose: MX multichannel analyzer driver for the EG&G Ortec Trump MCA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"

#if HAVE_ORTEC_UMCBI

#ifndef OS_WIN32
#error "This driver is only supported under Win32 (Windows NT/95)."
#endif

#include <windows.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_mca.h"
#include "i_umcbi.h"
#include "d_trump.h"

/* The following header file is supplied by Ortec as part of the UMCBI kit. */

#include "mcbcio32.h"

/* Initialize the MCA driver jump table. */

MX_RECORD_FUNCTION_LIST mxd_trump_record_function_list = {
	mxd_trump_initialize_type,
	NULL,
	NULL,
	mxd_trump_create_record_structures,
	mxd_trump_finish_record_initialization,
	mxd_trump_delete_record,
	mxd_trump_print_structure,
	NULL,
	mxd_trump_read_parms_from_hardware,
	mxd_trump_write_parms_to_hardware,
	mxd_trump_open,
	mxd_trump_close
};

MX_MCA_FUNCTION_LIST mxd_trump_mca_function_list = {
	mxd_trump_start,
	mxd_trump_stop,
	mxd_trump_read,
	mxd_trump_clear,
	mxd_trump_busy,
	mxd_trump_start_for_preset_time,
	mxd_trump_get_roi,
	mxd_trump_set_roi,
	mxd_trump_get_roi_integral,
	mxd_trump_get_num_channels,
	mxd_trump_set_num_channels,
	mxd_trump_get_real_time,
	mxd_trump_get_live_time,
};

MX_RECORD_FIELD_DEFAULTS mxd_trump_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_MCA_STANDARD_FIELDS,
	MXD_TRUMP_STANDARD_FIELDS
};

long mxd_trump_num_record_fields
		= sizeof( mxd_trump_record_field_defaults )
		  / sizeof( mxd_trump_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_trump_rfield_def_ptr
			= &mxd_trump_record_field_defaults[0];

/* A private function for the use of the driver. */

static mx_status_type
mxd_trump_get_pointers( MX_MCA *mca,
			MX_TRUMP_MCA **trump_mca,
			MX_UMCBI_DETECTOR **umcbi_detector,
			const char *calling_fname )
{
	const char fname[] = "mxd_trump_get_pointers()";

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The MX_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( trump_mca == (MX_TRUMP_MCA **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_TRUMP_MCA pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( umcbi_detector == (MX_UMCBI_DETECTOR **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_UMCBI_DETECTOR pointer passed by '%s' was NULL.",
			calling_fname );
	}

	*trump_mca = (MX_TRUMP_MCA *) (mca->record->record_type_struct);

	if ( *trump_mca == (MX_TRUMP_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_TRUMP_MCA pointer for record '%s' is NULL.",
			mca->record->name );
	}

	*umcbi_detector = (*trump_mca)->detector;

	if ( *umcbi_detector == (MX_UMCBI_DETECTOR *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_UMCBI_DETECTOR pointer for record '%s' is NULL.",
			mca->record->name );
	}
	return MX_SUCCESSFUL_RESULT;
}

/* === */

MX_EXPORT mx_status_type
mxd_trump_initialize_type( long type )
{
	/* Nothing here yet. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxd_trump_create_record_structures()";

	MX_MCA *mca;
	MX_TRUMP_MCA *trump_mca;

	/* Allocate memory for the necessary structures. */

	mca = (MX_MCA *) malloc( sizeof(MX_MCA) );

	if ( mca == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_MCA structure." );
	}

	trump_mca = (MX_TRUMP_MCA *) malloc( sizeof(MX_TRUMP_MCA) );

	if ( trump_mca == (MX_TRUMP_MCA *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_TRUMP_MCA structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = mca;
	record->record_type_struct = trump_mca;
	record->class_specific_function_list = &mxd_trump_mca_function_list;

	mca->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxd_trump_finish_record_initialization()";

	MX_MCA *mca;
	MX_TRUMP_MCA *trump_mca;

#if 1
	mca = (MX_MCA *) record->record_class_struct;

	mca->channel_array = ( unsigned long * )
		malloc( mca->max_num_channels * sizeof( unsigned long ) );

	if ( mca->channel_array == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory allocating an %lu channel data array.",
			mca->max_num_channels );
	}
#endif

	trump_mca = (MX_TRUMP_MCA *) record->record_type_struct;

	trump_mca->detector = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_delete_record( MX_RECORD *record )
{
	if ( record == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}
	if ( record->record_type_struct != NULL ) {
		free( record->record_type_struct );

		record->record_type_struct = NULL;
	}
	if ( record->record_class_struct != NULL ) {
		free( record->record_class_struct );

		record->record_class_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_print_structure( FILE *file, MX_RECORD *record )
{
	const char fname[] = "mxd_trump_print_structure()";

	MX_MCA *mca;
	MX_TRUMP_MCA *trump_mca;
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed is NULL.");
	}

	mca = (MX_MCA *) (record->record_class_struct);

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_MCA pointer for record '%s' is NULL.", record->name);
	}

	trump_mca = (MX_TRUMP_MCA *) (record->record_type_struct);

	if ( trump_mca == (MX_TRUMP_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_TRUMP_MCA pointer for record '%s' is NULL.", record->name);
	}

	fprintf(file, "MCA parameters for record '%s':\n", record->name);

	fprintf(file, "  MCA type              = TRUMP_MCA.\n\n");
	fprintf(file, "  UMCBI record          = '%s'\n",
					trump_mca->umcbi_record->name);
	fprintf(file, "  detector number       = %d\n",
					trump_mca->detector_number);
	fprintf(file, "  maximum # of channels = %lu\n",
					mca->max_num_channels);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_open( MX_RECORD *record )
{
	const char fname[] = "mxd_trump_open()";

	MX_MCA *mca;
	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	void *detector_handle;
	char response[80];
	mx_status_type status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	mca = (MX_MCA *) (record->record_class_struct);

	if ( mca == (MX_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MCA pointer for record '%s' is NULL.",
			record->name);
	}

	trump_mca = (MX_TRUMP_MCA *) (record->record_type_struct);

	if ( trump_mca == (MX_TRUMP_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRUMP_MCA pointer for scaler record '%s' is NULL.",
			record->name );
	}

	status = mxi_umcbi_get_detector_struct( trump_mca->umcbi_record,
			trump_mca->detector_number, &detector );

	if ( status.code != MXE_SUCCESS )
		return status;

	trump_mca->detector = detector;

	detector->record = record;

	detector_handle = MIOOpenDetector(trump_mca->detector_number, "", "");

	if ( detector_handle == NULL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Attempt to open connection to MCA '%s' via MIOOpenDetector() failed.  "
"Reason = '%s'.", record->name, mxi_umcbi_strerror() );
	}

	detector->detector_handle = detector_handle;

	/* Verify that we are talking to a Trump MCA card. */

	status = mxi_umcbi_command( detector, "SHOW_VERSION",
					response, sizeof response, TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( strncmp( response, "$FTRMP", 6 ) != 0 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
		"Detector '%s' is not a Trump multichannel buffer card.  "
		"Response to 'SHOW_VERSION' was '%s'.",
			record->name, response );
	}

	MX_DEBUG(-2,("%s: Connection to MCA '%s' was successfully opened.",
		fname, record->name));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_close( MX_RECORD *record )
{
	const char fname[] = "mxd_trump_close()";

	MX_TRUMP_MCA *trump_mca;
	BOOL status;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}

	trump_mca = (MX_TRUMP_MCA *) (record->record_type_struct);

	if ( trump_mca == (MX_TRUMP_MCA *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TRUMP_MCA pointer for scaler record '%s' is NULL.",
			record->name );
	}

	status = MIOCloseDetector( trump_mca->detector->detector_handle );

	if ( status == FALSE ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Attempt to close connection to MCA '%s' via MIOCloseDetector() failed.  "
"Reason = '%s'.", record->name, mxi_umcbi_strerror() );
	}

	MX_DEBUG(-2,("%s: Connection to MCA '%s' was successfully closed.",
		fname, record->name ));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_start( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_start()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	mx_status_type status;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "START", NULL, 0, TRUE );

	return status;
}

MX_EXPORT mx_status_type
mxd_trump_stop( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_stop()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	mx_status_type status;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "STOP", NULL, 0, TRUE );

	return status;
}

MX_EXPORT mx_status_type
mxd_trump_read( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_read()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	mx_status_type status;

	DWORD *data_ptr;
	DWORD data_mask, roi_mask;
	WORD returned_channels;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG(-2,("%s: Before mxd_trump_get_num_channels()", fname));

	status = mxd_trump_get_num_channels( mca );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG(-2,("%s: mca = %p", fname, mca));
	MX_DEBUG(-2,("%s: detector = %p", fname, detector));
	MX_DEBUG(-2,("%s: detector->detector_handle = %p",
			fname, detector->detector_handle));
	MX_DEBUG(-2,("%s: mca->current_num_channels = %lu",
			fname, mca->current_num_channels));
	MX_DEBUG(-2,("%s: mca->channel_array = %p",
			fname, mca->channel_array));

	MX_DEBUG(-2,("%s: Before MIOGetData()", fname));

	data_ptr = MIOGetData( detector->detector_handle,
				0, ( WORD ) mca->current_num_channels,
				( LPDWORD ) mca->channel_array,
				&returned_channels, &data_mask,
				&roi_mask, "" );

	MX_DEBUG(-2,("%s: After MIOGetData()", fname));
	MX_DEBUG(-2,("%s: returned channels = %hd", fname, returned_channels));
	MX_DEBUG(-2,("%s: data_mask = %lu 0x%x", fname, data_mask, data_mask));
	MX_DEBUG(-2,("%s: roi_mask = %lu 0x%x", fname, roi_mask, roi_mask));

	if ( data_ptr == NULL ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
	"Transfer of %lu channels from MCA '%s' failed.  Reason = '%s'",
			mca->current_num_channels,
			mca->record->name,
			mxi_umcbi_strerror() );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_clear( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_clear()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	mx_status_type status;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "CLEAR", NULL, 0, TRUE );

	return status;
}

MX_EXPORT mx_status_type
mxd_trump_busy( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_busy()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	char response[20];
	mx_status_type status;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "SHOW_ACTIVE",
				response, sizeof response, TRUE );

	if ( strcmp( response, "$C00000087" ) == 0 ) {
		mca->busy = FALSE;
	} else {
		mca->busy = TRUE;
	}

	return status;
}

MX_EXPORT mx_status_type
mxd_trump_start_for_preset_time( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_start_for_preset_time()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	char command[40];
	unsigned long tick_count;
	mx_status_type status;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "CLEAR", NULL, 0, TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	tick_count = ( mca->preset_time ) * MXD_TRUMP_TICKS_PER_SECOND;

	sprintf( command, "SET_LIVE_PRESET %lu", tick_count );

	status = mxi_umcbi_command( detector, command, NULL, 0, TRUE );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mxi_umcbi_command( detector, "START", NULL, 0, TRUE );

	return status;
}

MX_EXPORT mx_status_type
mxd_trump_get_roi( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_get_roi()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_trump_set_roi( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_set_roi()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_trump_get_roi_integral( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_get_roi_integral()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_trump_get_num_channels( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_get_num_channels()";

	MX_TRUMP_MCA *trump_mca;
	MX_UMCBI_DETECTOR *detector;
	mx_status_type status;

	WORD num_channels;

	status = mxd_trump_get_pointers( mca, &trump_mca, &detector, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	num_channels = MIOGetDetLength( detector->detector_handle );

	mca->current_num_channels = (unsigned long) num_channels;

	if ( mca->current_num_channels > mca->max_num_channels ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
"The detector '%s' is reported by UMCBI to have %lu channels, "
"but the record '%s' is only configured to support up to %lu channels.",
			detector->name, mca->current_num_channels,
			mca->record->name, mca->max_num_channels );
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_trump_set_num_channels( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_set_num_channels()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_trump_get_real_time( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_get_real_time()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

MX_EXPORT mx_status_type
mxd_trump_get_live_time( MX_MCA *mca )
{
	const char fname[] = "mxd_trump_get_live_time()";

	return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
		"This function is not yet implemented." );
}

#endif /* HAVE_ORTEC_UMCBI */

