/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    i_mmx000.c
 *
 * Purpose: MX driver for Newport MM3000 and MM4000 4-axis motor controllers.
 *
 * Note:    For the MM4000, this driver _assumes_ that the value of the
 *          field "Terminator" under "General Setup" for the controller
 *          is set to CR/LF.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_gener.h"
#include "mx_rs232.h"
#include "i_mmx000.h"

MX_RECORD_FUNCTION_LIST mxi_mmx000_record_function_list = {
	mxi_mmx000_initialize_type,
	NULL,
	NULL,
	mxi_mmx000_create_record_structures,
	mxi_mmx000_finish_record_initialization,
	mxi_mmx000_delete_record,
	NULL,
	NULL,
	mxi_mmx000_read_parms_from_hardware,
	mxi_mmx000_write_parms_to_hardware,
	mxi_mmx000_open,
	mxi_mmx000_close,
	NULL,
	mxi_mmx000_resynchronize
};

MX_GENERIC_FUNCTION_LIST mxi_mmx000_generic_function_list = {
	mxi_mmx000_getchar,
	mxi_mmx000_putchar,
	mxi_mmx000_read,
	mxi_mmx000_write,
	mxi_mmx000_input_is_available,
	mxi_mmx000_discard_unread_input,
	mxi_mmx000_discard_unwritten_output
};

MX_RECORD_FIELD_DEFAULTS mxi_mmx000_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_MMX000_STANDARD_FIELDS
};

long mxi_mmx000_num_record_fields
		= sizeof( mxi_mmx000_record_field_defaults )
			/ sizeof( mxi_mmx000_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_mmx000_rfield_def_ptr
			= &mxi_mmx000_record_field_defaults[0];

#define MXI_MMX000_DEBUG		FALSE

/*==========================*/

/* Declare some private functions for the driver here. */

static mx_status_type mx_construct_mm3000_error_response(char *, const char[]);
static mx_status_type mx_construct_mm4000_error_response(char *, const char[]);

/*==========================*/

MX_EXPORT mx_status_type
mxi_mmx000_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_mmx000_create_record_structures()";

	MX_GENERIC *generic;
	MX_MMX000 *mmx000;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	mmx000 = (MX_MMX000 *) malloc( sizeof(MX_MMX000) );

	if ( mmx000 == (MX_MMX000 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
			"Can't allocate memory for MX_MMX000 structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = mmx000;
	record->class_specific_function_list
				= &mxi_mmx000_generic_function_list;

	generic->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_mmx000_finish_record_initialization()";

	MX_MMX000 *mmx000;
	int i;

	mmx000 = (MX_MMX000 *) record->record_type_struct;

	mmx000->record = record;

	/* Only RS-232 records are supported. */

	if (mmx000->rs232_record->superclass != MXR_INTERFACE ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"'%s' is not an interface record.",
			mmx000->rs232_record->name );
	}

	if ( mmx000->rs232_record->class != MXI_RS232 ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"'%s' is not an RS-232 interface.",
			mmx000->rs232_record->name );
	}

	/* Mark the MMX000 device as being closed. */

	record->flags |= ( ~MXF_REC_OPEN );

	/* Null out the list of attached motors. */

	for ( i = 0; i < MX_NUM_NEWPORT_MMX000_MOTORS; i++ ) {
		mmx000->motor_record[i] = NULL;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_open( MX_RECORD *record )
{
	const char fname[] = "mxi_mmx000_open()";

	MX_MMX000 *mmx000;
	MX_GENERIC *generic;
	char response[80];
	char version_string[80];
	mx_status_type status;

	MX_DEBUG(2, ("mxi_mmx000_open() invoked."));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *) (record->record_type_struct);

	if ( mmx000 == (MX_MMX000 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for record '%s' is NULL.", record->name);
	}

	generic = (MX_GENERIC *) record->record_class_struct;

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for record '%s' is NULL.", record->name);
	}

	MX_DEBUG(2,("%s: Before first putline.", fname));

	/* Verify that the Newport controller is active by querying
	 * it for its firmware version number.
	 */

	status = mxi_mmx000_putline( mmx000, "VE", FALSE );

	if ( status.code != MXE_SUCCESS )
		return status;

	/******* Read the controller version string. *******/

	status = mxi_mmx000_getline(mmx000, response, sizeof response, FALSE);

	/* This is the first time that we have sent a command to the
	 * Newport controller.  We check the return status to see if
	 * we timed out.
	 */

	switch( status.code ) {
	case MXE_SUCCESS:
		break;
	case MXE_NOT_READY:
		return mx_error( MXE_NOT_READY, fname,
"No response from the Newport controller '%s'.  Is it turned on?",
			record->name );
		break;
	default:
		return status;
		break;
	}

	switch( record->type ) {
	case MXI_GEN_MM3000:
		strcpy( version_string, "Newport Corp. MM3000 Version" );
		break;
	case MXI_GEN_MM4000:
		strcpy( version_string, "VE MM400" );
		break;
	default:
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"This record is not a Newport MM3000 or MM4000 record.  "
	"The record is of type = %d.", record->type );
		break;
	}

	if (strncmp( response, version_string, strlen(version_string)) != 0) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"The device for interface '%s' is not an MM3000 or MM4000 controller.  "
"Response to 'VE' command was '%s'", record->name, response );
	}

	/* ******* Do controller type specific initialization. *******/

	if ( record->type == MXI_GEN_MM3000 ) {
		/* Set error reporting to be compatible with the way
		 * the MM4000 is handled.  This means the following
		 * bits in the FO word are always set to particular
		 * values:
		 *
		 * B0 (set to 0) - Enable appended text and long error codes.
		 * B1 (set to 1) - Disable automatic RS232 error message.
		 */

		status = mxi_mmx000_putline( mmx000,
			"FO|02;FO&FE", FALSE );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Disable the long BEEEEEEEPs after any error, since the
		 * controller doesn't send any characters out to the
		 * communications port until after the beep is over.
		 *
		 * This is done by setting bit B6 in the FS word.
		 */

		status = mxi_mmx000_putline( mmx000, "FS|40", FALSE );

		if ( status.code != MXE_SUCCESS )
			return status;
	} else
	if ( record->type == MXI_GEN_MM4000 ) {
		/* As far as I know, the MM4000 doesn't need any special
		 * customizations.
		 */
	}

	MX_DEBUG(2,("%s complete.", fname));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_close( MX_RECORD *record )
{
	/* Nothing to do here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_resynchronize( MX_RECORD *record )
{
	const char fname[] = "mxi_mmx000_resynchronize()";

	MX_MMX000 *mmx000;
	char response[80], banner[40];
	size_t length;
	mx_status_type status;

	mmx000 = (MX_MMX000 *) record->record_type_struct;

	if ( mmx000 == (MX_MMX000 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_MMX000 pointer for record '%s' is NULL.",
			record->name );
	}

	/* Discard any unread RS-232 input or unwritten RS-232 output. */

	status = mx_rs232_discard_unwritten_output( mmx000->rs232_record,
							MXI_MMX000_DEBUG );

	if ( status.code != MXE_SUCCESS && status.code != MXE_UNSUPPORTED )
		return status;

	status = mx_rs232_discard_unread_input( mmx000->rs232_record,
							MXI_MMX000_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Send an identify command to verify that the controller is there. */

	status = mxi_mmx000_command( mmx000, "VE",
			response, sizeof response, NULL, 0, MXI_MMX000_DEBUG );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Verify that we got the identification message that we expected. */

	switch( record->type ) {
	case MXI_GEN_MM3000:
		strcpy( banner, "Newport Corp. MM3000 Version" );
		break;
	case MXI_GEN_MM4000:
		strcpy( banner, " MM400" );
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Attempted to resynchronize a record that is not a Newport motor controller.  "
"Instead, it is of type %ld.", record->type );
		break;
	}

	length = strlen( banner );

	if ( strncmp( response, banner, length ) != 0 ) {
		return mx_error( MXE_DEVICE_IO_ERROR, fname,
"Did not get Newport identification banner in response to a 'VE' command.  "
"Instead got the response '%s'", response );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_getchar( MX_GENERIC *generic, char *c, int flags )
{
	const char fname[] = "mxi_mmx000_getchar()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_getchar( mmx000->rs232_record, c, flags );

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_putchar( MX_GENERIC *generic, char c, int flags )
{
	const char fname[] = "mxi_mmx000_putchar()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_putchar( mmx000->rs232_record, c, flags );

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_read( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_mmx000_read()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_read(mmx000->rs232_record, buffer, count, NULL, 0);

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_write( MX_GENERIC *generic, void *buffer, size_t count )
{
	const char fname[] = "mxi_mmx000_write()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_write(mmx000->rs232_record, buffer, count, NULL, 0);

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_input_is_available( MX_GENERIC *generic, int *input_available )
{
	const char fname[] = "mxi_mmx000_input_is_available()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_input_is_available(
				mmx000->rs232_record, input_available );

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_discard_unread_input( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_mmx000_discard_unread_input()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_discard_unread_input( mmx000->rs232_record,
						debug_flag );

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_discard_unwritten_output( MX_GENERIC *generic, int debug_flag )
{
	const char fname[] = "mxi_mmx000_discard_unwritten_output()";

	MX_MMX000 *mmx000;
	mx_status_type status;

	if ( generic == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_GENERIC pointer passed is NULL.");
	}

	mmx000 = (MX_MMX000 *)(generic->record->record_type_struct);

	if ( mmx000 == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_MMX000 pointer for MX_GENERIC structure is NULL." );
	}

	status = mx_rs232_discard_unwritten_output( mmx000->rs232_record,
							debug_flag );

	return status;
}

/* === Functions specific to this driver. === */

MX_EXPORT mx_status_type
mxi_mmx000_command( MX_MMX000 *mmx000, char *command,
		char *response, int response_buffer_length,
		char *error_message, int error_message_buffer_length,
		int debug_flag )
{
	const char fname[] = "mxi_mmx000_command()";

	MX_GENERIC *generic;
	unsigned long sleep_ms;
	int i, j, max_attempts;
	size_t prefix_length, response_length;
	char alt_error_message_buffer[80];
	char *local_error_buffer;
	int local_error_buffer_length;
	mx_status_type status, command_status;

	if ( mmx000 == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_MMX000 pointer passed was NULL." );
	}

	if ( command == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"'command' buffer pointer passed was NULL.  No command sent.");
	}

	if ( error_message == NULL ) {
		local_error_buffer = &alt_error_message_buffer[0];
		local_error_buffer_length = sizeof(alt_error_message_buffer)-1;
	} else {
		local_error_buffer = error_message;
		local_error_buffer_length = error_message_buffer_length;
	}

	if ( local_error_buffer_length < 4 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Error message buffer length of %d is too short.  Must be at least 4.",
			local_error_buffer_length );
	}

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	if ( generic == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_GENERIC pointer for MMX000 record '%s' is NULL.",
			mmx000->record->name );
	}

	/* Send the command string. */

	status = mxi_mmx000_putline( mmx000, command, debug_flag );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Get the response, if one is expected. */

	i = 0;

	if ( mmx000->record->type == MXI_GEN_MM3000 ) {
		max_attempts = 2;
		sleep_ms = 1000;
	} else {
		max_attempts = 1;
		sleep_ms = 0;
	}

	if ( response != NULL ) {
		for ( i=0; i < max_attempts; i++ ) {
			if ( i > 0 ) {
				MX_DEBUG(-2,
			("%s: No response yet to '%s' command.  Retrying...",
					mmx000->record->name, command ));
			}

			status = mxi_mmx000_getline( mmx000, response,
					response_buffer_length, debug_flag );

			if ( status.code == MXE_SUCCESS ) {
				break;		/* Exit the for() loop. */

			} else if ( status.code != MXE_NOT_READY ) {
				MX_DEBUG(-2,
				("*** Exiting with status = %ld",status.code));
				return status;
			}
			mx_msleep(sleep_ms);
		}

		if ( i >= max_attempts ) {
			status = mxi_mmx000_discard_unread_input(
						generic, debug_flag );

			if ( status.code != MXE_SUCCESS ) {
				mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Failed at attempt to discard unread characters in buffer for record '%s'",
					mmx000->record->name );
			}

			mx_error( MXE_INTERFACE_IO_ERROR, fname,
				"No response seen to '%s' command", command );
		}

		if ( mmx000->record->type == MXI_GEN_MM4000 ) {

			/* Most (but _not_ _all_) response strings from
			 * an MM4000 includes the command prefix (if any)
			 * followed by the two letter command (converted
			 * to uppercase).  The data for the response
			 * follows after.
			 * 
			 * For compatibility with the MM3000, we trim
			 * the echoed command and prefix off of the
			 * front of the response string, checking that
			 * we got a valid looking response first.
			 * Note that this code assumes that there are
			 * no unnecessary spaces in the command sent
			 * to the Newport controller.
			 */

			/* First, how long is the command prefix if present? 
			 * The prefix is assumed to be an integer.
			 */

			prefix_length = strspn( command, "0123456789" );

			for ( i = 0; i < (prefix_length + 2); i++ ) {
				if ( response[i] != toupper( command[i] ) ) {

					return mx_error(
						MXE_INTERFACE_IO_ERROR, fname,
"Command and prefix in response '%s' does not match original command '%s'",
						response, command );
				}
			}

			response_length = strlen( response );

			for ( i = 0, j = prefix_length + 2; 
			   i <= response_length; i++, j++ ) {
				response[i] = response[j];
			}
		}
	}

	/* Check to see if there was an error in the command we just did. */

	status = mxi_mmx000_putline( mmx000, "TB", debug_flag );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Read back the error string. */

	if ( i > 0 ) {
		mx_msleep( (i+1) * sleep_ms);
	}

	status = mxi_mmx000_getline( mmx000, local_error_buffer,
			local_error_buffer_length, debug_flag );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* MM3000's and MM4000's have differently formatted responses
	 * to the TB command.
	 */

	command_status = MX_SUCCESSFUL_RESULT;

	switch ( mmx000->record->type ) {
	case MXI_GEN_MM3000:
		if ( local_error_buffer[0] != 'E' ) {
			mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Response to TB command was _not_ an error message string.  Response = '%s'",
				local_error_buffer );
		}

		if (strncmp( local_error_buffer, "E00", 3 ) != 0) {

			command_status = mx_construct_mm3000_error_response(
				local_error_buffer, fname );
		}
		break;

	case MXI_GEN_MM4000:
		if ( strcmp( local_error_buffer, "TB@ No error" ) != 0 ) {

			command_status = mx_construct_mm4000_error_response(
				local_error_buffer, fname );
		}
		break;

	default:
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"%s invoked for device which is not a Newport MM3000 or MM4000.  "
	"Device type = %ld.", mmx000->record->type );
		break;
	}

	if ( command_status.code != MXE_SUCCESS ) {
		status = mxi_mmx000_discard_unread_input(generic, debug_flag);

		if ( status.code != MXE_SUCCESS ) {
			mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Failed at attempt to discard unread characters in buffer for record '%s'",
				mmx000->record->name );
		}
	}

	return command_status;
}

static mx_status_type
mx_construct_mm3000_error_response( char *error_message, const char fname[] )
{
	/* See Appendix A of the MM3000 User's Manual for what the
	 * original MM3000 error codes mean.
	 */

	static long translated_error_code[] = {
		MXE_SUCCESS,			/*  0 */
		MXE_INTERFACE_IO_ERROR,		/*  1 */
		MXE_DEVICE_IO_ERROR,		/*  2 */
		MXE_INTERFACE_IO_ERROR,		/*  3 */
		MXE_DEVICE_IO_ERROR,		/*  4 */
		MXE_DEVICE_IO_ERROR,		/*  5 */
		MXE_NOT_READY,			/*  6 */
		MXE_DEVICE_IO_ERROR,		/*  7 */
		MXE_DEVICE_ACTION_FAILED,	/*  8 */
		MXE_DEVICE_ACTION_FAILED,	/*  9 */
		MXE_DEVICE_ACTION_FAILED,	/* 10 */
		MXE_DEVICE_ACTION_FAILED,	/* 11 */
		MXE_INTERFACE_ACTION_FAILED,	/* 12 */
		MXE_INTERRUPTED,		/* 13 */
		MXE_OUT_OF_MEMORY,		/* 14 */
		MXE_ILLEGAL_ARGUMENT,		/* 15 */
		MXE_ILLEGAL_ARGUMENT,		/* 16 */
		MXE_INTERFACE_ACTION_FAILED,	/* 17 */
		MXE_ILLEGAL_ARGUMENT,		/* 18 */
		MXE_INTERFACE_ACTION_FAILED,	/* 19 */
		MXE_INTERFACE_ACTION_FAILED,	/* 20 */
		MXE_INTERFACE_ACTION_FAILED,	/* 21 */
		MXE_INTERFACE_ACTION_FAILED,	/* 22 */
		MXE_INTERFACE_IO_ERROR,		/* 23 */
		MXE_ILLEGAL_ARGUMENT,		/* 24 */
		MXE_CONTROLLER_INTERNAL_ERROR,	/* 25 */
		MXE_INTERFACE_IO_ERROR,		/* 26 */
		MXE_INTERFACE_IO_ERROR,		/* 27 */
		MXE_CONTROLLER_INTERNAL_ERROR,	/* 28 */
		MXE_NOT_READY,			/* 29 */
		MXE_INTERFACE_ACTION_FAILED,	/* 30 */
		MXE_INTERFACE_ACTION_FAILED,	/* 31 */
		MXE_INTERFACE_ACTION_FAILED,	/* 32 */
		MXE_DEVICE_ACTION_FAILED,	/* 33 */
		MXE_CONTROLLER_INTERNAL_ERROR,	/* 34 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 35 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 36 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 37 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 38 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 39 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 40 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 41 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 42 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 43 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 44 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 45 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 46 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 47 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 48 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 49 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 50 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 51 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 52 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 53 */
		MXE_LIMIT_WAS_EXCEEDED,		/* 54 */
		MXE_DEVICE_ACTION_FAILED,	/* 55 */
		MXE_DEVICE_ACTION_FAILED,	/* 56 */
		MXE_DEVICE_ACTION_FAILED,	/* 57 */
	};

	static int num_translated_codes = sizeof( translated_error_code )
				/ sizeof( translated_error_code[0] );

	long mx_error_code;
	int mm3000_error_code, num_tokens;

	num_tokens = sscanf( error_message, "E%d", &mm3000_error_code );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"No error code seen in response to TB command.  Response = '%s'",
			error_message );
	}

	if ( mm3000_error_code < 0
	  || mm3000_error_code >= num_translated_codes ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Illegal error code %d seen in response to TB command.  Response = '%s'",
			mm3000_error_code, error_message );
	}

	mx_error_code = translated_error_code[ mm3000_error_code ];

	return mx_error( mx_error_code, fname,
		"MM3000 error message = '%s'", error_message );
}
	

static mx_status_type
mx_construct_mm4000_error_response( char *error_message, const char fname[] )
{
	/* See Appendix A of the MM4000 User's Manual and the Appendix of
	 * the MM4000 User's Manual Addendum for what the original MM4000
	 * error codes mean.
	 */

	static struct {
		char mm4000_error_code;
		long mx_error_code;
	} error_code_table[] = {
		{'A',	MXE_CONTROLLER_INTERNAL_ERROR},
		{'B',	MXE_ILLEGAL_ARGUMENT},
		{'C',	MXE_ILLEGAL_ARGUMENT},
		{'D',	MXE_DEVICE_ACTION_FAILED},
		{'E',	MXE_ILLEGAL_ARGUMENT},
		{'F',	MXE_ILLEGAL_ARGUMENT},
		{'G',	MXE_ILLEGAL_ARGUMENT},
		{'H',	MXE_INTERFACE_ACTION_FAILED},
		{'I',	MXE_ILLEGAL_ARGUMENT},
		{'J',	MXE_ILLEGAL_ARGUMENT},
		{'K',	MXE_ILLEGAL_ARGUMENT},
		{'L',	MXE_INTERFACE_IO_ERROR},
		{'M',	MXE_OUT_OF_MEMORY},
		{'N',	MXE_ILLEGAL_ARGUMENT},
		{'O',	MXE_ILLEGAL_ARGUMENT},
		{'P',	MXE_ILLEGAL_ARGUMENT},
		{'Q',	MXE_ILLEGAL_ARGUMENT},
		{'R',	MXE_ILLEGAL_ARGUMENT},
		{'S',	MXE_INTERFACE_IO_ERROR},
		{'T',	MXE_DEVICE_ACTION_FAILED},
		{'U',	MXE_CONTROLLER_INTERNAL_ERROR},
		{'V',	MXE_ILLEGAL_ARGUMENT},
		{'W',	MXE_ILLEGAL_ARGUMENT},
		{'X',	MXE_ILLEGAL_ARGUMENT},
		{'Y',	MXE_ILLEGAL_ARGUMENT},
		{'Z',	MXE_ILLEGAL_ARGUMENT},
		{'[',	MXE_ILLEGAL_ARGUMENT},
		{'\\',	MXE_ILLEGAL_ARGUMENT},
		{']',	MXE_ILLEGAL_ARGUMENT},
		{'^',	MXE_ILLEGAL_ARGUMENT},
		{'_',	MXE_ILLEGAL_ARGUMENT},
		{'`',	MXE_ILLEGAL_ARGUMENT},
		{'a',	MXE_ILLEGAL_ARGUMENT},
		{'b',	MXE_ILLEGAL_ARGUMENT},
		{'c',	MXE_ILLEGAL_ARGUMENT},
		{'d',	MXE_ILLEGAL_ARGUMENT},
		{'e',	MXE_ILLEGAL_ARGUMENT},
		{'f',	MXE_ILLEGAL_ARGUMENT},
		{'g',	MXE_ILLEGAL_ARGUMENT},
		{'h',	MXE_WOULD_EXCEED_LIMIT}
	};

	static int num_error_codes = sizeof( error_code_table )
				/ sizeof( error_code_table[0] );

	long mx_error_code;
	int i, num_tokens;
	char mm4000_error_code;
	mx_status_type mx_status;

	num_tokens = sscanf( error_message, "TB%c", &mm4000_error_code );

	if ( num_tokens != 1 ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
	"No error code seen in response to TB command.  Response = '%s'",
			error_message );
	}

	for ( i = 0; num_error_codes; i++ ) {
		if (error_code_table[i].mm4000_error_code == mm4000_error_code)
			break;		/* exit the for() loop */
	}

	if ( i >= num_error_codes ) {
		return mx_error( MXE_INTERFACE_IO_ERROR, fname,
"Illegal error code '%d' seen in response to TB command.  Response = '%s'",
			mm4000_error_code, error_message );
	}

	mx_error_code = error_code_table[i].mx_error_code;

	if ( mm4000_error_code == 'D' ) {
		mx_status = mx_error( mx_error_code, fname,
	"The motor driver power is probably turned off.  "
	"Try pressing the 'Motor on' button on the MM4000 controller.  "
	"MM4000 error message = '%s'",
				error_message );

	} else if ( mm4000_error_code == 'Q' ) {
		mx_status = mx_error( mx_error_code, fname,
	"The MM4000 front panel menus are probably not at the top level.  "
	"Try pressing QUIT and/or NEXT menu items until you get to a menu "
	"titled 'Select Action'.  "
	"MM4000 error message = '%s'",
				error_message );
	} else {
		mx_status = mx_error( mx_error_code, fname,
			"MM4000 error message = '%s'", error_message );
	}
	return mx_status;
}

MX_EXPORT mx_status_type
mxi_mmx000_getline( MX_MMX000 *mmx000,
		char *buffer, long buffer_length, int debug_flag )
{
	const char fname[] = "mxi_mmx000_getline()";

	MX_GENERIC *generic;
	char c;
	int i;
	mx_status_type status;

	if ( debug_flag ) {
		MX_DEBUG(-2, ("mxi_mmx000_getline() invoked."));
	}

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	for ( i = 0; i < (buffer_length - 1) ; i++ ) {
		status = mx_generic_getchar( generic, &c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS ) {
			/* Make the buffer contents a valid C string
			 * before returning, so that we can at least
			 * see what appeared before the error.
			 */

			buffer[i] = '\0';

			if ( debug_flag ) {
				MX_DEBUG(-2,
				("mxi_mmx000_getline() failed.\nbuffer = '%s'",
					buffer));

				MX_DEBUG(-2,
				("Failed with status = %d, c = 0x%x '%c'",
				status, c, c));
			}

			return status;
		}

		if ( debug_flag ) {
			MX_DEBUG(-2,
			("mxi_mmx000_getline: received c = 0x%x '%c'", c, c));
		}

		buffer[i] = c;

		/* Responses from the MM3000 should be terminated
		 * by the end of line sequence CR, LF and the code
		 * assumes that an MM4000 has been set up compatibly
		 * with CR/LF as the terminator.
		 */

		if ( c == MX_CR ) {
			i++;

			/* There is also a line feed character to throw away.*/

			status = mx_generic_getchar( generic,
						&c, MXF_GENERIC_WAIT );

			if ( status.code != MXE_SUCCESS ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: failed to see any character after a CR character.", fname));
				}

				return status;
			}

			if ( c != MX_LF ) {
				if ( debug_flag ) {
					MX_DEBUG(-2,
("%s: The character after a CR was not an LF.  Instead saw 0%o", fname, c));
				}
			}

			if ( debug_flag ) {
				MX_DEBUG(-2,
			("mxi_mmx000_getline: received c = 0x%x '%c'", c, c));
			}

			break;		/* exit the for() loop */
		}
	}

	/* Check to see if the last character was an end of string
	 * and if it is then overwrite it with a NULL.
	 */

	if ( buffer[i-1] != MX_CR ) {
		status = mx_error( MXE_INTERFACE_IO_ERROR, fname,
			"Warning: Input buffer overrun." );

		buffer[i] = '\0';
	} else {
		status = MX_SUCCESSFUL_RESULT;

		buffer[i-1] = '\0';
	}

#if MXI_MMX000_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_mmx000_getline: buffer = '%s'", buffer) );
	}

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_putline( MX_MMX000 *mmx000, char *buffer, int debug_flag )
{
	MX_GENERIC *generic;
	char *ptr;
	char c;
	mx_status_type status;

#if MXI_MMX000_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_mmx000_putline: buffer = '%s'",buffer));
	}

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	ptr = &buffer[0];

	while ( *ptr != '\0' ) {
		c = *ptr;

		status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( debug_flag ) {
			MX_DEBUG(-2,
			("mxi_mmx000_putline: sent 0x%x '%c'", c, c));
		}

		ptr++;
	}

	/* Send a CR after the end of the command. */

	c = MX_CR;

	status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( debug_flag ) {
		MX_DEBUG(-2,("mxi_mmx000_putline(eos): sent 0x%x '%c'", c, c));
	}

	/* If this is an MM4000, the code expects that the line terminator
	 * is set to be CR/LF, so we must send an LF as well.
	 */

	if ( mmx000->record->type == MXI_GEN_MM4000 ) {
		c = MX_LF;

		status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( debug_flag ) {
			MX_DEBUG(-2,
			("mxi_mmx000_putline(eos): sent 0x%x '%c'", c, c));
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_mmx000_getc( MX_MMX000 *mmx000, char *c, int debug_flag )
{
	MX_GENERIC *generic;
	mx_status_type status;

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	status = mx_generic_getchar( generic, c, MXF_GENERIC_WAIT );

#if MXI_MMX000_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_mmx000_getc: received 0x%x '%c'", *c, *c));
	}

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_putc( MX_MMX000 *mmx000, char c, int debug_flag )
{
	MX_GENERIC *generic;
	mx_status_type status;

	generic = (MX_GENERIC *) mmx000->record->record_class_struct;

	status = mx_generic_putchar( generic, c, MXF_GENERIC_WAIT );

#if MXI_MMX000_DEBUG
	if ( 1 ) {
#else
	if ( debug_flag ) {
#endif
		MX_DEBUG(-2, ("mxi_mmx000_putc: sent 0x%x '%c'", c, c));
	}

	return status;
}

MX_EXPORT mx_status_type
mxi_mmx000_turn_motor_drivers_on( MX_MMX000 *mmx000, int debug_flag )
{
	mx_status_type status;

	status = mxi_mmx000_command(mmx000, "MO", NULL, 0, NULL, 0, debug_flag);

	return status;
}

