/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:
 *
 * Purpose: MX driver for Prairie Digital, Inc. Model 45 data acquisition
 *          and control module.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_rs232.h"
#include "pdi45.h"
#include "i_pdi45.h"

MX_RECORD_FUNCTION_LIST mxi_pdi45_record_function_list = {
	mxi_pdi45_initialize_type,
	NULL,
	NULL,
	mxi_pdi45_create_record_structures,
	mxi_pdi45_finish_record_initialization,
	mxi_pdi45_delete_record,
	NULL,
	NULL,
	mxi_pdi45_read_parms_from_hardware,
	mxi_pdi45_write_parms_to_hardware,
	mxi_pdi45_open,
	mxi_pdi45_close
};

MX_GENERIC_FUNCTION_LIST mxi_pdi45_generic_function_list = {
	NULL
#if 0
	mxi_pdi45_getchar,
	mxi_pdi45_putchar,
	mxi_pdi45_read,
	mxi_pdi45_write,
	mxi_pdi45_input_is_available,
	mxi_pdi45_discard_unread_input,
	mxi_pdi45_discard_unwritten_output
#endif
};

MX_RECORD_FIELD_DEFAULTS mxi_pdi45_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXI_PDI45_STANDARD_FIELDS
};

long mxi_pdi45_num_record_fields
		= sizeof( mxi_pdi45_record_field_defaults )
			/ sizeof( mxi_pdi45_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_pdi45_rfield_def_ptr
			= &mxi_pdi45_record_field_defaults[0];

MX_EXPORT mx_status_type
mxi_pdi45_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_pdi45_create_record_structures()";

	MX_GENERIC *generic;
	MX_PDI45 *pdi45;

	/* Allocate memory for the necessary structures. */

	generic = (MX_GENERIC *) malloc( sizeof(MX_GENERIC) );

	if ( generic == (MX_GENERIC *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_GENERIC structure." );
	}

	pdi45 = (MX_PDI45 *) malloc( sizeof(MX_PDI45) );

	if ( pdi45 == (MX_PDI45 *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_PDI45 structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = generic;
	record->record_type_struct = pdi45;
	record->class_specific_function_list
				= &mxi_pdi45_generic_function_list;

	generic->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_finish_record_initialization( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_read_parms_from_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_write_parms_to_hardware( MX_RECORD *record )
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_pdi45_open( MX_RECORD *record )
{
	const char fname[] = "mxi_pdi45_open()";

	MX_PDI45 *pdi45;
	mx_status_type status;

	MX_DEBUG(2, ("mxi_pdi45_open() invoked."));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	pdi45 = (MX_PDI45 *) (record->record_class_struct);

	if ( pdi45 == (MX_PDI45 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PDI45 pointer for record '%s' is NULL.", record->name);
	}

	status = pdi45_open( pdi45 );

	return status;
}

MX_EXPORT mx_status_type
mxi_pdi45_close( MX_RECORD *record )
{
	const char fname[] = "mxi_pdi45_close()";

	MX_PDI45 *pdi45;
	mx_status_type status;

	MX_DEBUG(2, ("mxi_pdi45_close() invoked."));

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	pdi45 = (MX_PDI45 *) (record->record_class_struct);

	if ( pdi45 == (MX_PDI45 *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_PDI45 pointer for record '%s' is NULL.", record->name);
	}

	status = pdi45_close( pdi45 );

	return status;
}

