/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name: i_scamac.c
 *
 * Purpose: An MX CAMAC driver for a software emulated interface.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_camac.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "i_scamac.h"

MX_RECORD_FUNCTION_LIST mxi_scamac_record_function_list = {
	mxi_scamac_initialize_type,
	NULL,
	NULL,
	mxi_scamac_create_record_structures,
	mxi_scamac_finish_record_initialization,
	mxi_scamac_delete_record,
	NULL,
	NULL,
	mxi_scamac_read_parms_from_hardware,
	mxi_scamac_write_parms_to_hardware,
	mxi_scamac_open,
	mxi_scamac_close
};

MX_CAMAC_FUNCTION_LIST mxi_scamac_camac_function_list = {
	mxi_scamac_get_lam_status,
	mxi_scamac_send_control_word,
	mxi_scamac_camac
};

#define TEST_X_DIMENSION	2
#define TEST_Y_DIMENSION	3
#define TEST_Z_DIMENSION	4

MX_RECORD_FIELD_DEFAULTS mxi_scamac_record_field_defaults[] = {
  MX_RECORD_STANDARD_FIELDS,
  MX_CAMAC_STANDARD_FIELDS,
  MXI_SCAMAC_STANDARD_FIELDS
};

long mxi_scamac_num_record_fields = sizeof( mxi_scamac_record_field_defaults )
				/ sizeof( mxi_scamac_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxi_scamac_record_field_def_ptr
			= &mxi_scamac_record_field_defaults[0];

#define LOGFILE_NAME_LENGTH 255

MX_EXPORT mx_status_type
mxi_scamac_initialize_type( long type )
{
	/* Nothing needed here. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxi_scamac_create_record_structures()";

	MX_CAMAC *crate;
	MX_SCAMAC *scamac;

	/* Allocate memory for the necessary structures. */

	crate = (MX_CAMAC *) malloc( sizeof(MX_CAMAC) );

	if ( crate == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_CAMAC structure." );
	}

	scamac = (MX_SCAMAC *) malloc( sizeof(MX_SCAMAC) );

	if ( scamac == NULL ) {
		free(crate);

		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAMAC structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = crate;
	record->record_type_struct = scamac;
	record->class_specific_function_list = &mxi_scamac_camac_function_list;

	crate->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxi_scamac_finish_record_initialization()";

	MX_SCAMAC *scamac;

	/* Get pointer to the array through the MX_SCAMAC structure. */

	scamac = (MX_SCAMAC *)(record->record_type_struct);

	if ( scamac == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCAMAC structure pointer is NULL." );
	}

	scamac->logfile = fopen( scamac->logfile_name, "w" );

	if ( scamac->logfile == NULL ) {
		return mx_error( MXE_FILE_IO_ERROR, fname,
			"Can't open CAMAC log file '%s'", scamac->logfile );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        if ( record->record_class_struct != NULL ) {
                free( record->record_class_struct );

                record->record_class_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_read_parms_from_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxi_scamac_read_parms_from_hardware() called."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_write_parms_to_hardware( MX_RECORD *record )
{
	MX_DEBUG(2, ("mxi_scamac_write_parms_to_hardware() called."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_open( MX_RECORD *record )
{
	MX_DEBUG( 2, ("mxi_scamac_open() called."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_close( MX_RECORD *record )
{
	MX_DEBUG( 2, ("mxi_scamac_close() called."));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_get_lam_status( MX_CAMAC *crate, int *lam_n )
{
	*lam_n = 0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_send_control_word( MX_CAMAC *crate, mx_uint8_type control_word)
{
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxi_scamac_camac( MX_CAMAC *crate, int slot, int subaddress,
		int function_code, mx_sint32_type *data, int *Q, int *X)
{
	const char fname[] = "mxi_scamac_camac()";

	MX_SCAMAC *scamac;

#ifdef DEBUG
	if ( function_code >= 16 && function_code <= 23 ) {
		MX_DEBUG(2, ("mxi_scamac_camac(%s, %d, %d, %d, %d) invoked.",
			crate->record->name, slot, subaddress,
			function_code, *data));
	} else {
		MX_DEBUG(2, ("mxi_scamac_camac(%s, %d, %d, %d) invoked.",
			crate->record->name, slot, subaddress, function_code));
	}
#endif

	scamac = (MX_SCAMAC *) (crate->record->record_type_struct);

	if ( scamac == (MX_SCAMAC *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SCAMAC structure pointer is NULL.");
	}

	if ( (scamac->logfile) != NULL ) {
		if ( function_code >= 16 && function_code <= 23 ) {
			fprintf((scamac->logfile),
				"camac: %s, %d, %d, %d, %ld (0x%lx)\n",
				crate->record->name, slot, subaddress,
				function_code, *data, *data);

			fflush(scamac->logfile);
		} else {
			fprintf((scamac->logfile),
				"camac: %s, %d, %d, %d\n",
				crate->record->name, slot, subaddress,
				function_code);

			fflush(scamac->logfile);
		}
	}

	if ( slot < 1 || slot > 24 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal CAMAC slot number = %d", slot);
	}

	if ( subaddress < 0 || subaddress > 15 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal CAMAC subaddress number = %d", subaddress);
	}

	if ( function_code < 0 || function_code > 31 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Illegal CAMAC function code = %d", function_code);
	}

	/* For CAMAC read operations, always return 1. */

	if ( function_code >= 0 && function_code <= 7 ) {

		*data = 1;

	}

	*Q = 1;
	*X = 1;

	return MX_SUCCESSFUL_RESULT;
}
