/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    m_epmtrd.c
 *
 * Purpose: This code is used to implement an EPICS server-side interface
 *          to a generic MX motor by using soft records in an EPICS IOC
 *          to communicate with EPICS clients by.
 *
 *          Note that this is _not_ the record you use to communicate with
 *          an OMS VME58 from an MX client program.  Look in the file
 *          "libmx/d_epmtr.c" for that driver.
 */

#include <stdio.h>

#include "mxconfig.h"

#if HAVE_EPICS

#include <stdlib.h>

#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "m_epmtrd.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"

MX_RECORD_FUNCTION_LIST mxm_epics_motor_db_record_function_list = {
	mxm_epics_motor_db_initialize_type,
	NULL,
	NULL,
	mxm_epics_motor_db_create_record_structures,
	mxm_epics_motor_db_finish_record_initialization,
	mxm_epics_motor_db_delete_record,
	NULL,
	NULL,
	mxm_epics_motor_db_read_parms_from_hardware,
	mxm_epics_motor_db_write_parms_to_hardware,
	mxm_epics_motor_db_open,
	mxm_epics_motor_db_close
};

MX_RECORD_FIELD_DEFAULTS mxm_epics_motor_db_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXM_EPICS_MOTOR_DB_STANDARD_FIELDS
};

long mxm_epics_motor_db_num_record_fields
                = sizeof( mxm_epics_motor_db_record_field_defaults )
                        / sizeof( mxm_epics_motor_db_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxm_epics_motor_db_rfield_def_ptr
                        = &mxm_epics_motor_db_record_field_defaults[0];

/* === */

MX_EXPORT mx_status_type
mxm_epics_motor_db_initialize_type( long type )
{
                /* Nothing needed here. */

                return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_create_record_structures( MX_RECORD *record )
{
        const char fname[] = "mxm_epics_motor_db_create_record_structures()";

        MX_EPICS_MAILBOX_MOTOR_DB *epics_mailbox_motor_db;
	MX_RECORD *list_head_record;
	MX_LIST_HEAD *list_head;
	long server_protocols_active;

	/* Check to see if we are being invoked from a client program. */

	list_head_record = record->list_head;

	list_head = (MX_LIST_HEAD *)
			(list_head_record->record_superclass_struct);

	if ( list_head->is_server == FALSE ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"The MX record '%s' cannot be used here.  "
	"The MX record for an EPICS mailbox motor database can be used in "
	"an MX server program, but _not_ in an MX client program.  "
	"Perhaps you want the 'epics_motor' record type instead?",
			record->name );
	}

	server_protocols_active = list_head->server_protocols_active;

	if ( ( server_protocols_active & MXPROT_EPICS_MAILBOX ) == 0 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The EPICS mailbox record '%s' may not be used since EPICS mailbox support "
"was not enabled for this server with the '-E' option.",
			record->name );
	}

        /* Allocate memory for the necessary structures. */

        epics_mailbox_motor_db = (MX_EPICS_MAILBOX_MOTOR_DB *)
				malloc( sizeof(MX_EPICS_MAILBOX_MOTOR_DB) );

        if ( epics_mailbox_motor_db == (MX_EPICS_MAILBOX_MOTOR_DB *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Can't allocate memory for MX_EPICS_MAILBOX_MOTOR_DB structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = NULL;
        record->record_type_struct = epics_mailbox_motor_db;
        record->class_specific_function_list = NULL;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_finish_record_initialization( MX_RECORD *record )
{
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_delete_record( MX_RECORD *record )
{
        if ( record == NULL ) {
                return MX_SUCCESSFUL_RESULT;
        }
        if ( record->record_type_struct != NULL ) {
                free( record->record_type_struct );

                record->record_type_struct = NULL;
        }
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_read_parms_from_hardware( MX_RECORD *record )
{
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_write_parms_to_hardware( MX_RECORD *record )
{
        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_open( MX_RECORD *record )
{
        const char fname[] = "mxm_epics_motor_db_open()";

	MX_DEBUG(-2,("%s invoked for record '%s'.", fname, record->name ));

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxm_epics_motor_db_close( MX_RECORD *record )
{
        /* We need do nothing special here.  The connection to EPICS
         * will go away automatically when the program exits.
         */

        return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
