/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_gpib.c
 *
 * Purpose: MX function library of generic GPIB operations.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_util.h"
#include "mx_gpib.h"
#include "mx_drivr.h"

MX_EXPORT mx_status_type
mx_gpib_finish_record_initialization( MX_RECORD *gpib_record )
{
	const char fname[] = "mx_gpib_finish_record_initialization()";

	MX_GPIB *gpib;
	int i;

	if ( gpib_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL." );
	}
	gpib = (MX_GPIB *) (gpib_record->record_class_struct);

	if ( gpib == (MX_GPIB *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_GPIB pointer for record '%s' is NULL.",
			gpib_record->name );
	}

	/* Initialize the field values for all of the GPIB addresses. */


	for ( i = 0; i < MX_NUM_GPIB_ADDRESSES; i++ ) {

		gpib->open_semaphore[i] = 0;
		gpib->secondary_address[i] = 0;

		gpib->io_timeout_value[i] = gpib->default_io_timeout_value;
		gpib->eoi_mode[i] = gpib->default_eoi_mode;
		gpib->read_terminator[i] = gpib->default_read_terminator;
		gpib->write_terminator[i] = gpib->default_write_terminator;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_gpib_get_pointers( MX_RECORD *gpib_record,
			MX_GPIB **gpib,
			MX_GPIB_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_gpib_get_pointers()";

	if ( gpib_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The gpib_record pointer passed by '%s' was NULL.",
			calling_fname );
	}

	if ( gpib_record->class != MXI_GPIB ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
	"The record '%s' passed by '%s' is not a GPIB interface.  "
	"(superclass = %ld, class = %ld, type = %ld)",
			gpib_record->name, calling_fname,
			gpib_record->superclass,
			gpib_record->class,
			gpib_record->type );
	}

	if ( gpib != (MX_GPIB **) NULL ) {
		*gpib = (MX_GPIB *) (gpib_record->record_class_struct);

		if ( *gpib == (MX_GPIB *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_GPIB pointer for record '%s' passed by '%s' is NULL.",
				gpib_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_GPIB_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_GPIB_FUNCTION_LIST *)
				(gpib_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_GPIB_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_GPIB_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				gpib_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_gpib_open_device( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_open_device()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( gpib->open_semaphore[ address ] > 0 ) {

		( gpib->open_semaphore[ address ] )++;
	} else {

		fptr = fl_ptr->open_device;

		if ( fptr == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"open_device function pointer is NULL.");
		}

		status = (*fptr)( gpib, address );

		if ( status.code != MXE_SUCCESS ) {
			gpib->open_semaphore[ address ] = 0;

			return status;
		} else {
			gpib->open_semaphore[ address ] = 1;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_gpib_close_device( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_close_device()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( gpib->open_semaphore[ address ] > 1 ) {

		( gpib->open_semaphore[ address ] )--;

	} else if ( gpib->open_semaphore[ address ] <= 0 ) {

		return mx_error( MXE_FUNCTION_FAILED, fname,
		"Attempted to close GPIB address %d on interface '%s' "
		"when it was already closed.",
			address, gpib_record->name );
	} else {

		fptr = fl_ptr->close_device;

		if ( fptr == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"close_device function pointer is NULL.");
		}

		status = (*fptr)( gpib, address );

		gpib->open_semaphore[ address ] = 0;

		if ( status.code != MXE_SUCCESS )
			return status;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_gpib_read_parms_from_device( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_read_parms_from_device()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->read_parms_from_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"read_parms_from_device function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_write_parms_to_device( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_write_parms_to_device()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->write_parms_to_device;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"write_parms_to_device function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_read( MX_RECORD *gpib_record,
		int address,
		char *buffer,
		size_t max_bytes_to_read,
		size_t *bytes_read,
		int flags)
{
	const char fname[] = "mx_gpib_read()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)(MX_GPIB *, int, char *, size_t, size_t *, int);
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->read;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"read function pointer is NULL.");
	}

	status = (*fptr)( gpib, address,
			buffer, max_bytes_to_read, bytes_read, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_write( MX_RECORD *gpib_record,
		int address,
		char *buffer,
		size_t bytes_to_write,
		size_t *bytes_written,
		int flags)
{
	const char fname[] = "mx_gpib_write()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)(MX_GPIB *, int, char *, size_t, size_t *, int);
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->write;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"write function pointer is NULL.");
	}

	status = (*fptr)( gpib, address,
			buffer, bytes_to_write, bytes_written, flags );

	return status;
}

/*
 * mx_gpib_getline() and mx_gpib_putline() are variants of mx_gpib_read()
 * and mx_gpib_write() that are intended for use in sending and receiving
 * ASCII character strings rather than binary data.
 */

MX_EXPORT mx_status_type
mx_gpib_getline( MX_RECORD *gpib_record,
		int address,
		char *buffer,
		size_t max_bytes_to_read,
		size_t *bytes_read,
		int flags)
{
	const char fname[] = "mx_gpib_getline()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	size_t real_max_bytes_to_read, real_bytes_read;
	mx_status_type (*fptr)(MX_GPIB *, int, char *, size_t, size_t *, int);
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->read;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"read function pointer is NULL.");
	}

	/* Make sure to leave an extra byte at the end for the '\0'
	 * character to be placed at the end of the string.
	 */

	real_max_bytes_to_read = max_bytes_to_read - 1;

	status = (*fptr)( gpib, address,
		buffer, real_max_bytes_to_read, &real_bytes_read, flags );

	buffer[real_bytes_read] = '\0';

	if ( bytes_read != NULL ) {
		*bytes_read = real_bytes_read + 1;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_putline( MX_RECORD *gpib_record,
		int address,
		char *buffer,
		size_t *bytes_written,
		int flags)
{
	const char fname[] = "mx_gpib_putline()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	size_t bytes_to_write;
	mx_status_type (*fptr)(MX_GPIB *, int, char *, size_t, size_t *, int);
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->write;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"write function pointer is NULL.");
	}

	bytes_to_write = strlen( buffer );

	status = (*fptr)( gpib, address,
			buffer, bytes_to_write, bytes_written, flags );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_interface_clear( MX_RECORD *gpib_record )
{
	const char fname[] = "mx_gpib_interface_clear()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB * );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->interface_clear;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"interface_clear function pointer is NULL.");
	}

	status = (*fptr)( gpib );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_device_clear( MX_RECORD *gpib_record )
{
	const char fname[] = "mx_gpib_device_clear()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB * );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->device_clear;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"device_clear function pointer is NULL.");
	}

	status = (*fptr)( gpib );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_selective_device_clear( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_selective_device_clear()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->selective_device_clear;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"selective_device_clear function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_local_lockout( MX_RECORD *gpib_record )
{
	const char fname[] = "mx_gpib_local_lockout()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB * );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->local_lockout;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"local_lockout function pointer is NULL.");
	}

	status = (*fptr)( gpib );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_remote_enable( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_remote_enable()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->remote_enable;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"remote_enable function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_go_to_local( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_go_to_local()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->go_to_local;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"go_to_local function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_trigger( MX_RECORD *gpib_record, int address )
{
	const char fname[] = "mx_gpib_trigger()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->trigger;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"trigger function pointer is NULL.");
	}

	status = (*fptr)( gpib, address );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_wait_for_service_request( MX_RECORD *gpib_record, double timeout )
{
	const char fname[] = "mx_gpib_wait_for_service_request()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, double );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->wait_for_service_request;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"wait_for_service_request function pointer is NULL.");
	}

	status = (*fptr)( gpib, timeout );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_serial_poll( MX_RECORD *gpib_record,
			int address, unsigned char *serial_poll_byte )
{
	const char fname[] = "mx_gpib_serial_poll()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB *, int, unsigned char * );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->serial_poll;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"serial_poll function pointer is NULL.");
	}

	status = (*fptr)( gpib, address, serial_poll_byte );

	return status;
}

MX_EXPORT mx_status_type
mx_gpib_serial_poll_disable( MX_RECORD *gpib_record )
{
	const char fname[] = "mx_gpib_serial_poll_disable()";

	MX_GPIB *gpib;
	MX_GPIB_FUNCTION_LIST *fl_ptr;
	mx_status_type (*fptr)( MX_GPIB * );
	mx_status_type status;

	status = mx_gpib_get_pointers( gpib_record,
					&gpib, &fl_ptr, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	fptr = fl_ptr->serial_poll_disable;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"serial_poll_disable function pointer is NULL.");
	}

	status = (*fptr)( gpib );

	return status;
}

