/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_mca.c
 *
 * Purpose: MX function library of generic multichannel analyzer operations.
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_mca.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_mca_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_mca_get_pointers( MX_RECORD *mca_record,
			MX_MCA **mca,
			MX_MCA_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_mca_get_pointers()";

	if ( mca_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The mca_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( mca_record->class != MXC_MULTICHANNEL_ANALYZER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not an MCA.",
			mca_record->name, calling_fname );
	}

	if ( mca != (MX_MCA **) NULL ) {
		*mca = (MX_MCA *) (mca_record->record_class_struct);

		if ( *mca == (MX_MCA *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_MCA pointer for record '%s' passed by '%s' is NULL",
				mca_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_MCA_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_MCA_FUNCTION_LIST *)
				(mca_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_MCA_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_MCA_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				mca_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_mca_start( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_start()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_fn = function_list->start;

	if ( start_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"start function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*start_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_stop( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_stop()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *stop_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	stop_fn = function_list->stop;

	if ( stop_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"stop function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*stop_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_read( MX_RECORD *mca_record, unsigned long **channel_array )
{
	const char fname[] = "mx_mca_read()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( channel_array == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"channel_array pointer passed is NULL." );
	}

	*channel_array = NULL;

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*read_fn)( mca );

	if ( status.code == MXE_SUCCESS ) {
		*channel_array = mca->channel_array;
	}
	return status;
}

MX_EXPORT mx_status_type
mx_mca_clear( MX_RECORD *mca_record )
{
	const char fname[] = "mx_mca_clear()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *clear_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	clear_fn = function_list->clear;

	if ( clear_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*clear_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_is_busy( MX_RECORD *mca_record, int *busy )
{
	const char fname[] = "mx_mca_is_busy()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *busy_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	busy_fn = function_list->busy;

	if ( busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"busy function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*busy_fn)( mca );

	*busy = mca->busy;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_start_for_preset_time( MX_RECORD *mca_record, double preset_seconds )
{
	const char fname[] = "mx_mca_start_for_preset_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *start_for_preset_time_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	start_for_preset_time_fn = function_list->start_for_preset_time;

	if ( start_for_preset_time_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"start_for_preset_time function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	mca->preset_time = preset_seconds;

	status = (*start_for_preset_time_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_roi( MX_RECORD *mca_record,
			unsigned long *roi_start,
			unsigned long *roi_end )
{
	const char fname[] = "mx_mca_get_roi()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_roi_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( roi_start == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"roi_start pointer passed is NULL." );
	}
	if ( roi_end == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"roi_end pointer passed is NULL." );
	}
	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_roi_fn = function_list->get_roi;

	if ( get_roi_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_roi function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*get_roi_fn)( mca );

	*roi_start = mca->roi_start;
	*roi_end = mca->roi_end;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_set_roi( MX_RECORD *mca_record,
			unsigned long roi_start,
			unsigned long roi_end )
{
	const char fname[] = "mx_mca_set_roi()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *set_roi_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_roi_fn = function_list->set_roi;

	if ( set_roi_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_roi function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	mca->roi_start = roi_start;
	mca->roi_end = roi_end;

	status = (*set_roi_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_roi_integral( MX_RECORD *mca_record,
			unsigned long *roi_integral )
{
	const char fname[] = "mx_mca_get_roi_integral()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_roi_integral_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( roi_integral == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"roi_integral pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_roi_integral_fn = function_list->get_roi_integral;

	if ( get_roi_integral_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_roi_integral function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*get_roi_integral_fn)( mca );

	*roi_integral = mca->roi_integral;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_num_channels( MX_RECORD *mca_record,
			unsigned long *num_channels )
{
	const char fname[] = "mx_mca_get_num_channels()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_num_channels_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( num_channels == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"num_channels pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_num_channels_fn = function_list->get_num_channels;

	if ( get_num_channels_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_num_channels function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*get_num_channels_fn)( mca );

	*num_channels = mca->current_num_channels;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_set_num_channels( MX_RECORD *mca_record,
			unsigned long num_channels )
{
	const char fname[] = "mx_mca_set_num_channels()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *set_num_channels_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	set_num_channels_fn = function_list->set_num_channels;

	if ( set_num_channels_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"set_num_channels function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	mca->current_num_channels = num_channels;

	status = (*set_num_channels_fn)( mca );

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_real_time( MX_RECORD *mca_record, double *real_time )
{
	const char fname[] = "mx_mca_get_real_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_real_time_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( real_time == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"real_time pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_real_time_fn = function_list->get_real_time;

	if ( get_real_time_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_real_time function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*get_real_time_fn)( mca );

	*real_time = mca->real_time;

	return status;
}

MX_EXPORT mx_status_type
mx_mca_get_live_time( MX_RECORD *mca_record, double *live_time )
{
	const char fname[] = "mx_mca_get_live_time()";

	MX_MCA *mca;
	MX_MCA_FUNCTION_LIST *function_list;
	mx_status_type ( *get_live_time_fn ) ( MX_MCA * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for MCA '%s'", mca_record->name));

	if ( live_time == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"live_time pointer passed is NULL." );
	}

	status = mx_mca_get_pointers( mca_record,
					&mca, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	get_live_time_fn = function_list->get_live_time;

	if ( get_live_time_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"get_live_time function ptr for MX_MCA ptr 0x%p is NULL.",
			mca);
	}

	/* All is well, so invoke the function. */

	status = (*get_live_time_fn)( mca );

	*live_time = mca->live_time;

	return status;
}

