/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:     mx_motor.h
 *
 * Purpose:  Main header file for the MX motor class.
 */

#ifndef __MX_MOTOR_H__
#define __MX_MOTOR_H__

/* Include hardware and operating system dependent definitions and symbols. */

#include "mx_rec.h"

/* ------------------------------------------------------------------ */

#define MX_MOTOR_STEP_FUZZ    (1.0e-9)
#define MX_MOTOR_ANALOG_FUZZ  (1.0e-9)

#define MXF_MTR_NONE               	0
#define MXF_MTR_SHOW_MOVE          	1
#define MXF_MTR_NOWAIT             	2
#define MXF_MTR_IGNORE_KEYBOARD    	4
#define MXF_MTR_IGNORE_BACKLASH    	8
#define MXF_MTR_GO_TO_BACKLASH_POSITION	16
#define MXF_MTR_ONLY_CHECK_LIMITS	32

/* MX_MOTOR contains the data that is common to all motor types. */

typedef struct {
	MX_RECORD *record; /* Pointer to the MX_RECORD structure that points
			    * to this motor.
			    */

	union { long stepper;
		double analog;
	} raw_setpoint;

	union { long stepper;
		double analog;
	} raw_position;

	union { long stepper;
		double analog;
	} raw_backlash_correction;

	union { long stepper;
		double analog;
	} raw_negative_limit;

	union { long stepper;
		double analog;
	} raw_positive_limit;

	union { long stepper;
		double analog;
	} deadband;

	int subclass;

	double scale;
	double offset;
	char   units[MXU_UNITS_NAME_LENGTH];

	double setpoint;
	double position;
	double backlash_correction;
	double negative_limit;
	double positive_limit;

	int busy;
	int soft_abort;
	int immediate_abort;
	int negative_limit_hit;
	int positive_limit_hit;
	int home_search;
	int set_mode;

	double dial_scale;
	double dial_offset;
	char   dial_units[MXU_UNITS_NAME_LENGTH];
	double dial_setpoint;
	double dial_position;
	double dial_backlash_correction;
	double dial_negative_limit;
	double dial_positive_limit;
} MX_MOTOR;

#define MXLV_MTR_BUSY			1001
#define MXLV_MTR_SETPOINT		1002
#define MXLV_MTR_POSITION		1003
#define MXLV_MTR_SOFT_ABORT		1004
#define MXLV_MTR_POSITIVE_LIMIT_HIT	1005
#define MXLV_MTR_NEGATIVE_LIMIT_HIT	1006

#define MX_STEPPER_MOTOR_STANDARD_FIELDS \
  {-1, -1, "raw_setpoint", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_setpoint.stepper), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_position", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_position.stepper), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_backlash_correction", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, \
					raw_backlash_correction.stepper), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_negative_limit", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_negative_limit.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_positive_limit", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_positive_limit.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "deadband", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, deadband.stepper),\
	{0}, NULL, MXFF_IN_DESCRIPTION}


#define MX_ANALOG_MOTOR_STANDARD_FIELDS \
  {-1, -1, "raw_setpoint", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_setpoint.analog), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "raw_position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_position.analog), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_backlash_correction", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, \
					raw_backlash_correction.analog), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_negative_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_negative_limit.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "raw_positive_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, raw_positive_limit.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "deadband", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, deadband.analog),\
	{0}, NULL, MXFF_IN_DESCRIPTION}


#define MX_MOTOR_STANDARD_FIELDS \
  {-1, -1, "subclass", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, subclass), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "scale", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, scale), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "offset", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, offset), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "units", MXFT_STRING, NULL, 1, {MXU_UNITS_NAME_LENGTH}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, units), \
	{sizeof(char)}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {MXLV_MTR_SETPOINT, -1, "setpoint", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, setpoint), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_POSITION, -1, "position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, position), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "backlash_correction", MXFT_DOUBLE, NULL, 0, {0},\
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, backlash_correction), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "negative_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, negative_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "positive_limit", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, positive_limit),\
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_BUSY, -1, "busy", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, busy), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_SOFT_ABORT, -1, "soft_abort", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, soft_abort), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "immediate_abort", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, immediate_abort), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_POSITIVE_LIMIT_HIT, -1, "negative_limit_hit", MXFT_INT, \
	NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, negative_limit_hit), \
	{0}, NULL, 0}, \
  \
  {MXLV_MTR_NEGATIVE_LIMIT_HIT, -1, "positive_limit_hit", MXFT_INT, \
	NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, positive_limit_hit), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "home_search", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, home_search), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "set_mode", MXFT_INT, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, set_mode), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "dial_scale", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_scale), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "dial_offset", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_offset), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "dial_units", MXFT_STRING, NULL, 1, {MXU_UNITS_NAME_LENGTH}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_units), \
	{sizeof(char)}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "dial_setpoint", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_setpoint), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "dial_position", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_position), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "dial_backlash_correction", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_backlash_correction), \
	{0}, NULL, 0}, \
  \
  {-1, -1, "dial_negative_limit", MXFT_DOUBLE, NULL, 0, {0},\
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_negative_limit),\
	{0}, NULL, 0}, \
  \
  {-1, -1, "dial_positive_limit", MXFT_DOUBLE, NULL, 0, {0},\
	MXF_REC_CLASS_STRUCT, offsetof(MX_MOTOR, dial_positive_limit),\
	{0}, NULL, 0}

/*
 * The structure type MX_MOTOR_FUNCTION_LIST contains a list of
 * pointers to functions that vary from motor type to motor type.
 * By abstracting away the individual motor operations in this
 * fashion, it is possible to minimize device dependencies in
 * the main motor source code.
 */

typedef struct {
	mx_status_type ( *motor_is_busy )( MX_MOTOR *motor );
	mx_status_type ( *move_absolute )( MX_MOTOR *motor );
	mx_status_type ( *get_position )( MX_MOTOR *motor );
	mx_status_type ( *set_position )( MX_MOTOR *motor );
	mx_status_type ( *soft_abort )( MX_MOTOR *motor );
	mx_status_type ( *immediate_abort )( MX_MOTOR *motor );
	mx_status_type ( *positive_limit_hit )( MX_MOTOR *motor );
	mx_status_type ( *negative_limit_hit )( MX_MOTOR *motor );
	mx_status_type ( *find_home_position )( MX_MOTOR *motor );
} MX_MOTOR_FUNCTION_LIST;

typedef mx_status_type (*MX_DISPLAY_MOTOR_MOVE_FPTR)(int, int, MX_RECORD **);

/* ============== Public functions and symbols . ================= */

MX_API mx_status_type mx_motor_finish_record_initialization(
						MX_RECORD *motor_record );

MX_API mx_status_type mx_motor_get_pointers( MX_RECORD *motor_record,
		MX_MOTOR **motor, MX_MOTOR_FUNCTION_LIST **function_list_ptr,
		const char *calling_fname );

MX_API mx_status_type mx_motor_is_busy( MX_RECORD *motor_record, int *busy );

MX_API mx_status_type mx_motor_soft_abort( MX_RECORD *motor_record );

MX_API mx_status_type mx_motor_immediate_abort( MX_RECORD *motor_record );

/* === Move by engineering units functions. === */

MX_API mx_status_type mx_motor_move_relative(MX_RECORD *motor_record,
			double relative_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags);

MX_API mx_status_type mx_motor_move_absolute(MX_RECORD *motor_record,
			double position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags);

MX_API mx_status_type mx_motor_array_move_absolute(
			int num_motor_records,
			MX_RECORD **motor_record_array,
			double *position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags );

MX_API mx_status_type mx_motor_array_move_segment(
			int num_motor_records,
			MX_RECORD **motor_record_array,
			double *position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags );

MX_API mx_status_type mx_motor_get_position( MX_RECORD *motor_record,
							double *position );
MX_API mx_status_type mx_motor_set_position( MX_RECORD *motor_record, 
							double position );

MX_API mx_status_type mx_motor_positive_limit_hit(
				MX_RECORD *motor_record, int *limit_hit );

MX_API mx_status_type mx_motor_negative_limit_hit(
				MX_RECORD *motor_record, int *limit_hit );

MX_API mx_status_type mx_motor_find_home_position( MX_RECORD *motor_record,
					int direction );

MX_API mx_status_type mx_motor_zero_position_value( MX_RECORD *motor_record );

MX_API mx_status_type mx_is_motor_position_between_software_limits(
			MX_RECORD *motor_record, double position,
			int *result_flag, int generate_error_message );

/* === Move by steps functions. (MXC_MTR_STEPPER) === */

MX_API mx_status_type mx_motor_move_relative_steps(MX_RECORD *motor_record,
			long motor_steps,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags);

MX_API mx_status_type mx_motor_move_absolute_steps(MX_RECORD *motor_record,
			long motor_steps,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags);

MX_API mx_status_type mx_motor_get_position_steps(
				MX_RECORD *motor_record, long *motor_steps );
MX_API mx_status_type mx_motor_set_position_steps(
				MX_RECORD *motor_record, long motor_steps );

MX_API mx_status_type mx_motor_steps_to_go(
				MX_RECORD *motor_record, long *steps_to_go );

MX_API mx_status_type
		mx_is_stepper_motor_position_between_software_limits(
			MX_RECORD *motor_record, long step_position,
			int *result_flag, int generate_error_message );

/* === Move analog functions. (MXC_MTR_ANALOG) === */

MX_API mx_status_type mx_motor_move_absolute_analog(MX_RECORD *motor_record,
			double motor_position,
			MX_DISPLAY_MOTOR_MOVE_FPTR display_motor_move_fn,
			int flags);

MX_API mx_status_type mx_motor_get_position_analog(
				MX_RECORD *motor_record, double *position );
MX_API mx_status_type mx_motor_set_position_analog(
				MX_RECORD *motor_record, double position );

MX_API mx_status_type
		mx_is_analog_motor_position_between_software_limits(
			MX_RECORD *motor_record, double analog_position,
			int *result_flag, int generate_error_message );

/* === */

extern MX_RECORD_FUNCTION_LIST mx_motor_record_function_list;

#endif /* __MX_MOTOR_H__ */
