/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_scalr.c
 *
 * Purpose: MX function library of generic scaler operations.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_scalr.h"

/*=======================================================================*/

/* This function is a utility function to consolidate all of the pointer
 * mangling that often has to happen at the beginning of an mx_scaler_...
 * function.  The parameter 'calling_fname' is passed so that error messages
 * will appear with the name of the calling function.
 */

MX_EXPORT mx_status_type
mx_scaler_get_pointers( MX_RECORD *scaler_record,
			MX_SCALER **scaler,
			MX_SCALER_FUNCTION_LIST **function_list_ptr,
			const char *calling_fname )
{
	const char fname[] = "mx_scaler_get_pointers()";

	if ( scaler_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"The scaler_record pointer passed by '%s' was NULL",
			calling_fname );
	}

	if ( scaler_record->class != MXC_SCALER ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
			"The record '%s' passed by '%s' is not a scaler.",
			scaler_record->name, calling_fname );
	}

	if ( scaler != (MX_SCALER **) NULL ) {
		*scaler = (MX_SCALER *) (scaler_record->record_class_struct);

		if ( *scaler == (MX_SCALER *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCALER pointer for record '%s' passed by '%s' is NULL",
				scaler_record->name, calling_fname );
		}
	}

	if ( function_list_ptr != (MX_SCALER_FUNCTION_LIST **) NULL ) {
		*function_list_ptr = (MX_SCALER_FUNCTION_LIST *)
				(scaler_record->class_specific_function_list);

		if ( *function_list_ptr == (MX_SCALER_FUNCTION_LIST *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_SCALER_FUNCTION_LIST pointer for record '%s' passed by '%s' is NULL.",
				scaler_record->name, calling_fname );
		}
	}

	return MX_SUCCESSFUL_RESULT;
}

/*=======================================================================*/

MX_EXPORT mx_status_type
mx_scaler_clear( MX_RECORD *scaler_record )
{
	const char fname[] = "mx_scaler_clear()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *clear_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	clear_fn = function_list->clear;

	if ( clear_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"clear function ptr for MX_SCALER ptr 0x%p is NULL.",
			scaler);
	}

	/* All is well, so invoke the function. */

	status = (*clear_fn)( scaler );

	/* The "overflow_set" field should _not_ be set to FALSE here,
	 * since for some scaler types, mx_scaler_read_and_clear() is
	 * implemented as the equivalent of a call to mx_scaler_read()
	 * by the equivalent of a call to mx_scaler_clear().
	 */

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_overflow_set( MX_RECORD *scaler_record, int *overflow_set )
{
	const char fname[] = "mx_scaler_overflow_set()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *overflow_set_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	overflow_set_fn = function_list->overflow_set;

	if ( overflow_set_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"overflow_set function ptr for MX_SCALER ptr 0x%p is NULL.",
			scaler);
	}

	/* All is well, so invoke the function. */

	status = (*overflow_set_fn)( scaler );

	*overflow_set = scaler->overflow_set;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_read( MX_RECORD *scaler_record, long *value )
{
	const char fname[] = "mx_scaler_read()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *read_fn ) ( MX_SCALER * );
	int overflow_set;
	mx_status_type status, status2;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_fn = function_list->read;

	if ( read_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_SCALER ptr 0x%p is NULL.",
			scaler);
	}

	*value = 0L;

	/* All is well, so invoke the function. */

	status = (*read_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Did the scaler overflow? 
	 *
	 * Note there is a race condition here if the scaler is
	 * currently gated on.
	 */

	if ( scaler->automatic_overflow_check ) {

		status2 = mx_scaler_overflow_set(scaler_record, &overflow_set);

		switch ( status2.code ) {
		case MXE_SUCCESS:    /* do nothing */
			break;
		case MXE_UNSUPPORTED:
			status2 = MX_SUCCESSFUL_RESULT;
			overflow_set = FALSE;
			break;
		default:
			return status2;
		}
	}

	*value = scaler->value;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_read_and_clear( MX_RECORD *scaler_record, long *value )
{
	const char fname[] = "mx_scaler_read_and_clear()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *read_and_clear_fn ) ( MX_SCALER * );
	int overflow_set;
	mx_status_type status, status2;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	read_and_clear_fn = function_list->read_and_clear;

	if ( read_and_clear_fn == NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_SCALER ptr 0x%p is NULL.",
			scaler);
	}

	/* All is well, so invoke the function. */

	*value = 0L;

	status = (*read_and_clear_fn)( scaler );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Did the scaler overflow? 
	 *
	 * Note there is a race condition here if the scaler is
	 * currently gated on.
	 */

	if ( scaler->automatic_overflow_check ) {

		status2 = mx_scaler_overflow_set(scaler_record, &overflow_set);

		switch ( status2.code ) {
		case MXE_SUCCESS:    /* do nothing */
			break;
		case MXE_UNSUPPORTED:
			status2 = MX_SUCCESSFUL_RESULT;
			overflow_set = FALSE;
			break;
		default:
			return status2;
		}
	}

	*value = scaler->value;

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_write( MX_RECORD *scaler_record, long value )
{
	const char fname[] = "mx_scaler_write()";

	MX_SCALER *scaler;
	MX_SCALER_FUNCTION_LIST *function_list;
	mx_status_type ( *write_fn ) ( MX_SCALER * );
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, &function_list, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	write_fn = function_list->write;

	if ( write_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"read function ptr for MX_SCALER ptr 0x%p is NULL.",
			scaler);
	}

	/* All is well, so invoke the function. */

	scaler->value = value;

	status = (*write_fn)( scaler );

	if ( scaler->automatic_overflow_check == TRUE
	  && status.code == MXE_SUCCESS ) {

		scaler->overflow_set = FALSE;
	}

	return status;
}

MX_EXPORT mx_status_type
mx_scaler_get_dark_current( MX_RECORD *scaler_record, double *dark_current )
{
	const char fname[] = "mx_scaler_get_dark_current()";

	MX_SCALER *scaler;
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	*dark_current = scaler->dark_current;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scaler_set_dark_current( MX_RECORD *scaler_record, double dark_current )
{
	const char fname[] = "mx_scaler_set_dark_current()";

	MX_SCALER *scaler;
	mx_status_type status;

	status = mx_scaler_get_pointers( scaler_record,
					&scaler, NULL, fname );

	if ( status.code != MXE_SUCCESS )
		return status;

	scaler->dark_current = dark_current;

	return MX_SUCCESSFUL_RESULT;
}

