/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mx_scan.c --- Scan record support for MX library. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_key.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_var.h"
#include "mx_relay.h"
#include "mx_scan.h"
#include "mx_sclin.h"
#include "mx_scseg.h"
#include "mx_ain.h"
#include "mx_din.h"
#include "mx_scalr.h"
#include "mx_alloc.h"

/* --------------- */

#define DEFAULT_SHUTTER_NAME	"shutter"
#define DEFAULT_POLICY_NAME	"shutter_policy"

MX_EXPORT mx_status_type
mx_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mx_prepare_for_scan_start()";

	MX_RECORD *policy_record;
	MX_RECORD_FIELD *value_field;
	MX_SCAN_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_SCAN * );
	void *pointer_to_value;
	mx_status_type status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	/* ==== Setup scan shutter policy. ==== */

	/* Does a shutter exist? */

	scan->shutter_record
		= mx_get_record( scan->record, DEFAULT_SHUTTER_NAME );

	if ( scan->shutter_record == NULL ) {
		scan->shutter_policy = MXF_SHUTTER_IGNORE;
	} else {
		/* Find out what the scan shutter policy should be. */

		policy_record = mx_get_record( scan->record,
						DEFAULT_POLICY_NAME );

		if ( policy_record == NULL ) {
			scan->shutter_record = NULL;
			scan->shutter_policy = MXF_SHUTTER_IGNORE;
		} else {
			/* Check to see if this is the right kind of record. */

			if ( policy_record->superclass != MXR_VARIABLE ) {
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"The record '%s' is not a variable record.",
					DEFAULT_POLICY_NAME );
			}
			/* Get the value of the shutter policy. */

			status = mx_find_record_field( policy_record,
						"value", &value_field );

			if ( status.code != MXE_SUCCESS )
				return status;

			if ( value_field->datatype != MXFT_INT ) {
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"The variable '%s' is not of type MXFT_INT.",
					DEFAULT_POLICY_NAME );
			}
			if ( value_field->num_dimensions != 1 ) {
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The variable '%s' must have only one dimension.",
					DEFAULT_POLICY_NAME );
			}

			pointer_to_value
				= mx_get_field_value_pointer( value_field );

			scan->shutter_policy = *( (int *) pointer_to_value );

			/* Check that the shutter policy has a valid value. */

			switch( scan->shutter_policy ) {
			case MXF_SHUTTER_IGNORE:
			case MXF_SHUTTER_OPEN_FOR_SCAN:
			case MXF_SHUTTER_OPEN_FOR_DATAPOINT:
				break;
			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"The '%s' record is set to an illegal value of %d.  "
"Allowed values are 0=ignore, 1=open for scan, 2=open for datapoint",
					DEFAULT_POLICY_NAME,
					scan->shutter_policy );
				break;
			}
		}
	}

	/* ==== Initialize graphical plotting of scan results. ==== */

	if ( mx_plotting_is_enabled( scan->record ) ) {
		if ( scan->plot == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Plotting is enabled, but the MX_PLOT pointer is NULL." );
		}

		status = mx_plot_open( scan->plot );

		if ( status.code != MXE_SUCCESS )
			return status;

		scan->plot->section_number = 0;
	}

	/* ===== Clear all the input devices that need it. ===== */

	status = mx_clear_scan_input_devices( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Invoke the scan type specific initialization function. ==== */

	fl_ptr = (MX_SCAN_FUNCTION_LIST *)
			(scan->record->superclass_specific_function_list);

	if ( fl_ptr == (MX_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Scan function list pointer is NULL.");
	}

	fptr = fl_ptr->prepare_for_scan_start;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Prepare for scan function pointer is NULL.");
	}

	status = ( *fptr ) ( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* ==== Write out the main scan header. ==== */

	status = mx_datafile_write_main_header( scan->datafile );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_perform_scan( MX_SCAN *scan )
{
	const char fname[] = "mx_perform_scan()";

	MX_SCAN_FUNCTION_LIST *fl_ptr;
	mx_status_type ( *fptr ) ( MX_SCAN * );
	int enable_flag;
	mx_status_type status, perform_scan_status;

	MX_DEBUG( 2, ( "%s invoked.", fname ) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"Scan routine was passed a NULL pointer.");
	}

	if ( scan->datafile == (MX_DATAFILE *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DATAFILE pointer is NULL." );
	}

	status = mx_datafile_open( scan->datafile );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_prepare_for_scan_start( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	fl_ptr = (MX_SCAN_FUNCTION_LIST *)
			(scan->record->superclass_specific_function_list);

	if ( fl_ptr == (MX_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Scan function list pointer is NULL.");
	}

	fptr = fl_ptr->perform_scan;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Compute motor positions function pointer is NULL.");
	}

	/* Open the shutter now if requested. */

	if ( scan->shutter_policy == MXF_SHUTTER_OPEN_FOR_SCAN ) {
		status = mx_relay_command( scan->shutter_record,
						MXF_OPEN_RELAY );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Perform the scan. */

	perform_scan_status = ( *fptr ) ( scan );

	/* Close the shutter now if requested. */

	if ( scan->shutter_policy != MXF_SHUTTER_IGNORE ) {
		(void) mx_relay_command( scan->shutter_record,
						MXF_CLOSE_RELAY );
	}

	/* Check the status of the scan that just completed. */

	switch( perform_scan_status.code ) {
	case MXE_SUCCESS:
		mx_info( "Scan '%s' complete.", scan->record->name );
		break;
	case MXE_INTERRUPTED:
		mx_info( "Scan '%s' aborted.", scan->record->name );
		break;
	default:
		mx_info( "Scan '%s' terminated with error code %ld.",
			scan->record->name,
			perform_scan_status.code );
		break;
	}

	enable_flag = mx_plotting_is_enabled( scan->record );

	if ( enable_flag == 1 ) {
		if (scan->plot->type != MXP_CHILD
		 && scan->plot->type != MXP_NONE ) {

			mx_info_dialog("\aHit any key to close the plot...",
		"Scan Complete.  Click the OK button to close the plot", "OK");

		}
		(void) mx_plot_close( scan->plot );

	} else if ( enable_flag == 2 ) {
		/* Give the plotting program a little time to finish
		 * plotting the last data point.
		 */

		mx_msleep(500);

		(void) mx_plot_close( scan->plot );
	}
	
	status = mx_datafile_close( scan->datafile );

	if ( status.code != MXE_SUCCESS ) {
		if ( perform_scan_status.code == MXE_SUCCESS
		  || perform_scan_status.code == MXE_INTERRUPTED ) {
			return status;
		}
	}

	return perform_scan_status;
}

/* --------------- */

MX_EXPORT mx_status_type
mx_scan_measure_data( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_measure_data()";

	MX_RECORD **input_device_array;
	MX_RECORD *input_device;
	MX_RECORD *timer_record;
	MX_SCALER *scaler;
	MX_SCAN *child_scan;
	double double_value, remaining_time;
	long long_value;
	unsigned long ulong_value;
	int i, busy;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.",fname));

	input_device_array = scan->input_device_array;

	if ( input_device_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"input_device_array pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	timer_record = (MX_RECORD *)( scan->timer_record );

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Timer MX_RECORD pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	/* Open the shutter if requested. */

	if ( scan->shutter_policy == MXF_SHUTTER_OPEN_FOR_DATAPOINT ) {
		status = mx_relay_command( scan->shutter_record,
							MXF_OPEN_RELAY );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	/* Wait for the settling time. */

	status = mx_timer_start( timer_record, scan->settling_time );

	/* Wait for the settling time to be over. */

	busy = TRUE;

	while ( busy ) {
		if ( mx_user_requested_interrupt() ) {
			(void) mx_timer_stop( timer_record, &remaining_time );

			return mx_error( MXE_INTERRUPTED, fname,
			"Settling time was interrupted.");
		}

		status = mx_timer_is_busy( timer_record, &busy );

		if ( status.code != MXE_SUCCESS )
			return status;

		mx_msleep(1);	/* Wait at least 1 millisecond. */
	}

	/* Clear the input devices. */

	status = mx_clear_scan_input_devices( scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Start the measurement. */

	status = mx_timer_start( timer_record, scan->integration_time );

	if ( status.code != MXE_SUCCESS )
		return status;

	/* Wait for the measurement to be over. */

	busy = TRUE;

	while ( busy ) {
		if ( mx_user_requested_interrupt() ) {
			(void) mx_timer_stop( timer_record, &remaining_time );

			return mx_error( MXE_INTERRUPTED, fname,
			"Measurement integration time was interrupted.");
		}

		status = mx_timer_is_busy( timer_record, &busy );

		if ( status.code != MXE_SUCCESS )
			return status;

		mx_msleep(1);	/* Wait at least 1 millisecond. */
	}

	/* Close the shutter if requested. */

	if ( scan->shutter_policy == MXF_SHUTTER_OPEN_FOR_DATAPOINT ) {
		(void) mx_relay_command( scan->shutter_record,
							MXF_CLOSE_RELAY );
	}

	/* Read out and save the values from the input devices,
	 * or perform child scans, or whatever.
	 */

	for ( i = 0; i < scan->num_input_devices; i++ ) {
		input_device = input_device_array[i];

		switch ( input_device->superclass ) {
		case MXR_VARIABLE:
			status = mx_receive_variable( input_device );

			if ( status.code != MXE_SUCCESS ) {
				return status;
			}
			break;

		case MXR_SCAN:
			child_scan = (MX_SCAN *)
				(input_device->record_superclass_struct);

			status = mx_perform_scan( child_scan );

			if ( status.code != MXE_SUCCESS ) {
				return status;
			}
			break;

		case MXR_DEVICE:
			switch( input_device->class ) {
			case MXC_ANALOG_INPUT:
				status = mx_analog_input_read(
						input_device, &double_value );
				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;
			case MXC_DIGITAL_INPUT:
				status = mx_digital_input_read(
						input_device, &ulong_value );
				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;
			case MXC_MOTOR:
				status = mx_motor_get_position(
						input_device, &double_value );
				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;
			case MXC_SCALER:
				status = mx_scaler_read(
						input_device, &long_value );
				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				scaler = (MX_SCALER *)
					(input_device->record_class_struct);
				if ( scaler->subtract_dark_current ) {
					long_value -= mx_round(
						scaler->dark_current
						* scan->integration_time);
				}
				scaler->value = long_value;
				break;
			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Device type %ld cannot be a scan input device.",
					input_device->class );
				break;
			}
			break;

		default:
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record superclass %ld cannot be a scan input device.",
				input_device->superclass );
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scan_display_scan_progress( MX_SCAN *scan )
{
	const char fname[] = "mx_scan_display_scan_progress()";

	MX_RECORD **input_device_array;
	MX_RECORD *input_device;
	MX_LINEAR_SCAN *linear_scan;
	double *motor_position;
	long *motor_is_independent_variable;
	char buffer[250];
	char little_buffer[50];
	int i, num_motors, num_input_devices;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	num_motors = scan->num_motors;
	motor_position = scan->motor_position;
	motor_is_independent_variable = scan->motor_is_independent_variable;

	num_input_devices = scan->num_input_devices;
	input_device_array = scan->input_device_array;

	strcpy(buffer, "");

	if ( num_motors <= 0 ) {
		if ( scan->record->class == MXS_LINEAR_SCAN ) {
			linear_scan = (MX_LINEAR_SCAN *)
					(scan->record->record_class_struct);
			sprintf( little_buffer, " %ld",
					(linear_scan->step_number)[0] );
			strcat(buffer, little_buffer);
		}
	} else {
		for ( i = 0; i < num_motors; i++ ) {
			if ( motor_is_independent_variable[i] ) {
				sprintf( little_buffer, " %-.3f",
							motor_position[i] );
				strcat(buffer, little_buffer);
			}
		}
	}
	strcat(buffer, " - ");

	for ( i = 0; i < num_input_devices; i++ ) {
		input_device = input_device_array[i];

		status = mx_convert_device_value_to_string( input_device,
				little_buffer, sizeof(little_buffer) - 1 );

		if ( status.code != MXE_SUCCESS )
			return status;

		strcat(buffer, little_buffer);
		strcat(buffer, " ");
	} 

	/* Send the formatted progress string to the user. */

	mx_info(buffer);

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_clear_scan_input_devices( MX_SCAN *scan )
{
	const char fname[] = "mx_clear_scan_input_devices()";

	MX_RECORD *input_device;
	long i;
	mx_status_type status;

	/* ===== Clear all the input devices that need it. ===== */

	/* At present, only scalers need to be cleared. */

	if ( scan->num_input_devices > 0 ) {
		if ( scan->input_device_array == (MX_RECORD **) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
				"Scaler array pointer is NULL.");
		}
	}

	for ( i = 0; i < scan->num_input_devices; i++ ) {

		input_device = (scan->input_device_array)[i];

		if ( input_device == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Input device %d pointer is NULL", input_device);
		}

		switch( input_device->superclass ) {
		case MXR_VARIABLE:
		case MXR_SCAN:
			/* No clear operation needed. */
			break;

		case MXR_DEVICE:
			switch( input_device->class ) {
			case MXC_SCALER:
				status = mx_scaler_clear( input_device );

				if ( status.code != MXE_SUCCESS ) {
					return status;
				}
				break;

			default:
				/* All other input devices do not need
				 * to be cleared.
				 */
				break;
			}
			break;

		default:
			/* Interface and other record superclasses
			 * cannot be scan input devices.
			 */

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record superclass %d cannot be a scan input.",
				input_device->superclass );
			break;
		}
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scan_fixup_varargs_record_field_defaults(
		MX_RECORD_FIELD_DEFAULTS *record_field_defaults_array,
		long num_record_fields,
		long *num_independent_variables_varargs_cookie,
		long *num_motors_varargs_cookie,
		long *num_input_devices_varargs_cookie )
{
	const char fname[] = "mx_scan_fixup_varargs_record_field_defaults()";

	MX_RECORD_FIELD_DEFAULTS *field;
	long referenced_field_index;
	mx_status_type status;

	MX_DEBUG( 8,("%s invoked.", fname));

	/****
	 **** Put the 'num_motors' value in the appropriate places.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_motors", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0, num_motors_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_motors varargs cookie = %ld",
			fname, *num_motors_varargs_cookie));

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_position", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"motor_is_independent_variable", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_motors_varargs_cookie;

	/****
	 **** Put the 'num_input_devices' value in the
	 **** 'input_device_array' field.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_input_devices", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0,
			num_input_devices_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_input_devices varargs cookie = %ld",
			fname, *num_input_devices_varargs_cookie));

	/*---*/

	status = mx_find_record_field_defaults(
			record_field_defaults_array, num_record_fields,
			"input_device_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = *num_input_devices_varargs_cookie;

	/****
	 **** Construct the 'num_independent_variables' varargs cookie
	 **** which is used by the individual device classes.
	 ****/

	status = mx_find_record_field_defaults_index(
			record_field_defaults_array, num_record_fields,
			"num_independent_variables", &referenced_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
			referenced_field_index, 0,
			num_independent_variables_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	MX_DEBUG( 8,("%s: num_independent_variables varargs_cookie = %ld",
			fname, *num_independent_variables_varargs_cookie));

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_convert_device_value_to_string( MX_RECORD *input_device,
				char *buffer, size_t buffer_length )
{
	const char fname[] = "mx_convert_device_value_to_string()";

	MX_RECORD_FIELD *record_field;
	mx_status_type (*token_constructor) ( void *, char *, size_t,
					MX_RECORD *, MX_RECORD_FIELD * );
	MX_ANALOG_INPUT *analog_input;
	MX_DIGITAL_INPUT *digital_input;
	MX_MOTOR *motor;
	MX_SCALER *scaler;
	void *field_value_ptr;
	double position, raw_position;
	mx_status_type status;

	if ( buffer_length <= 0 ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Buffer length %ld is too short.", buffer_length );
	}

	switch( input_device->superclass ) {
	case MXR_SCAN:
		strcpy(buffer, "");
		break;
	case MXR_DEVICE:
		if ( buffer_length <= 10 ) {
			return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"Buffer length %ld <= 10 is too short.", buffer_length );
		}
		switch( input_device->class ) {
		case MXC_ANALOG_INPUT:
			analog_input = (MX_ANALOG_INPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%-10.*g", input_device->precision,
					analog_input->value);
			break;
		case MXC_DIGITAL_INPUT:
			digital_input = (MX_DIGITAL_INPUT *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10ld", digital_input->value);
			break;
		case MXC_MOTOR:
			motor = (MX_MOTOR *)
				(input_device->record_class_struct);
			switch( motor->subclass ) {
			case MXC_MTR_ANALOG:
				raw_position = motor->raw_position.analog;
				break;
			case MXC_MTR_STEPPER:
				raw_position
				    = (double)(motor->raw_position.stepper);
				break;
			default:
				return mx_error(MXE_NOT_YET_IMPLEMENTED, fname,
				"Motor subclass %ld not yet implemented.",
					motor->subclass );
				break;
			}
			position = motor->offset + motor->scale * raw_position;

			sprintf(buffer, "%-10.*g", input_device->precision,
					position);
			break;
		case MXC_SCALER:
			scaler = (MX_SCALER *)
				(input_device->record_class_struct);
			sprintf(buffer, "%10ld", scaler->value);
			break;
		default:
			return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Record class %ld not yet implemented.",
				input_device->class );
			break;
		}
		break;
	case MXR_VARIABLE:
		status = mx_find_record_field( input_device, "value",
						&record_field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field_value_ptr = mx_get_field_value_pointer( record_field );

		if ( field_value_ptr == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Field value pointer for record field '%s' in record '%s' is NULL.",
				record_field->name, input_device->name );
		}

		status = mx_get_token_constructor( record_field->datatype,
						&token_constructor );

		if ( status.code != MXE_SUCCESS )
			return status;

		strcpy( buffer, "" );

		status = mx_create_array_description(
				field_value_ptr,
				(record_field->num_dimensions - 1),
				buffer, buffer_length,
				input_device, record_field,
				token_constructor );

		if ( status.code != MXE_SUCCESS )
			return status;

		break;
	default:
		return mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"Record superclass %ld not yet implemented.",
				input_device->superclass );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_scan_delete_pointer_from_parent_to_child(
	MX_RECORD *parent_scan_record,
	MX_RECORD *child_scan_record )
{
	const char fname[] = "mx_scan_delete_pointer_from_parent_to_child()";

	MX_SEGMENTED_SCAN *segmented_scan;
	MX_RECORD_FIELD *field;
	MX_RECORD **old_scan_segment_record_array;
	MX_RECORD **new_scan_segment_record_array;
	long old_num_segments, new_num_segments;
	long i, j;
	mx_status_type status;

	if ( parent_scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"parent_scan_record pointer passed was NULL." );
	}
	if ( child_scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"child_scan_record pointer passed was NULL." );
	}
	if ( parent_scan_record->superclass != MXR_SCAN ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"Parent record '%s' is not a scan record.  Instead, it is of superclass %ld",
			parent_scan_record->name,
			parent_scan_record->superclass );
	}
	if ( child_scan_record->superclass != MXR_SCAN ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"Child record '%s' is not a scan record.  Instead, it is of superclass %ld",
			child_scan_record->name,
			child_scan_record->superclass );
	}
	if ( parent_scan_record->class != MXS_SEGMENTED_SCAN ) {
		return mx_error( MXE_TYPE_MISMATCH, fname,
"Parent record '%s' is not a segmented scan.  Instead, it is of class %ld ",
			parent_scan_record->name,
			parent_scan_record->class );
	}

	segmented_scan = (MX_SEGMENTED_SCAN *)
				(parent_scan_record->record_class_struct);

	if ( segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_SEGMENTED_SCAN pointer for parent record '%s' is NULL.",
			parent_scan_record->name );
	}

	/* Find the MX_RECORD_FIELD for the scan_segment_record_array
	 * field, since we will need it when modify the size of the array.
	 */

	status = mx_find_record_field( parent_scan_record,
			"scan_segment_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	if ( field->num_dimensions != 1 ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"scan_segment_record_array dimension size != 1.  Instead = %ld",
			field->num_dimensions );
	}

	/****/

	old_scan_segment_record_array
				= segmented_scan->scan_segment_record_array;
	old_num_segments        = segmented_scan->num_segments;

	if ( old_scan_segment_record_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"scan_segment_record_array pointer for parent segmented scan '%s' is NULL.",
			parent_scan_record->name );
	}

	for ( i = 0; i < old_num_segments; i++ ) {
		if ( child_scan_record == old_scan_segment_record_array[i] ) {
			break;        /* Exit the for() loop. */
		}
	}

	if ( i >= old_num_segments ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Parent scan record '%s' does not have record '%s' in its list of children.",
			parent_scan_record->name, child_scan_record->name );
	}

	/* Create a new scan_segment_record_array to replace the old one. */

	new_num_segments = old_num_segments - 1;

	new_scan_segment_record_array = (MX_RECORD **)
			malloc( new_num_segments * sizeof(MX_RECORD *) );

	if ( new_scan_segment_record_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating new scan_segment_record_array." );
	}

	for ( j = 0; j < new_num_segments; j++ ) {
		if ( j < i ) {
			new_scan_segment_record_array[j]
				= old_scan_segment_record_array[j];
		} else {
			new_scan_segment_record_array[j]
				= old_scan_segment_record_array[j+1];
		}
	}

	/* Update the values in the MX_SEGMENTED_SCAN structure. */

	segmented_scan->num_segments = new_num_segments;
	segmented_scan->scan_segment_record_array
					= new_scan_segment_record_array;

	/* We must also update the values in the MX_RECORD_FIELD structure
	 * to match.
	 */

	field->dimension[0] = new_num_segments;

	mx_write_void_pointer_to_memory_location( field->data_pointer,
				(void *) new_scan_segment_record_array );

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_find_child_scan_record_pointer_in_parent(
		MX_SCAN *child_scan,
		MX_RECORD ***address_of_pointer_to_record_pointer_in_parent )
{
	const char fname[] = "mx_find_child_scan_record_pointer_in_parent()";

	MX_RECORD *parent_scan_record;
	MX_SEGMENTED_SCAN *parent_segmented_scan;
	MX_RECORD **parent_scan_segment_record_array;
	long i, parent_num_segments;

	parent_scan_record = child_scan->parent_scan_record;

	if ( parent_scan_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The scan record '%s' does not have a parent.\n",
			child_scan->record->name );
	}

	parent_segmented_scan = (MX_SEGMENTED_SCAN *)
		(parent_scan_record->record_class_struct);

	if ( parent_segmented_scan
			== (MX_SEGMENTED_SCAN *) NULL ) {

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Parent '%s' of this child scan has a NULL MX_SEGMENTED_SCAN structure.\n",
			parent_scan_record->name );
	}

	parent_num_segments = parent_segmented_scan->num_segments;

	parent_scan_segment_record_array =
	    parent_segmented_scan->scan_segment_record_array;

	if ( parent_scan_segment_record_array == (MX_RECORD **) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Scan segment record array for parent scan '%s' is NULL.\n",
			parent_scan_record->name );
	}

	for ( i = 0; i < parent_num_segments; i++ ) {
		if (child_scan->record == parent_scan_segment_record_array[i]){
			break;     /* Exit the for() loop. */
		}
	}

	if ( i >= parent_num_segments ) {
		*address_of_pointer_to_record_pointer_in_parent = NULL;

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Could not find pointer to child scan '%s' in parent scan '%s'.",
			child_scan->record->name,
			parent_scan_record->name );
	}

	*address_of_pointer_to_record_pointer_in_parent
			= &parent_scan_segment_record_array[i];

	return MX_SUCCESSFUL_RESULT;
}

