/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mx_scan.h --- Header file for generic scan superclass support. */

#ifndef _MX_SCAN_H_
#define _MX_SCAN_H_

#include "mx_util.h"
#include "mx_rec.h"
#include "mx_motor.h"
#include "mx_scalr.h"
#include "mx_timer.h"
#include "mx_dfile.h"
#include "mx_plot.h"

#define MXF_SHUTTER_IGNORE		0
#define MXF_SHUTTER_OPEN_FOR_SCAN	1
#define MXF_SHUTTER_OPEN_FOR_DATAPOINT	2

typedef struct {
	MX_RECORD *record; /* Pointer to the MX_RECORD structure that points
                            * to this MX_SCAN structure.
                            */
	MX_RECORD *parent_scan_record;

	long num_independent_variables;
	long num_motors;
	double *motor_position;
	MX_RECORD **motor_record_array;
	long *motor_is_independent_variable;
	long num_input_devices;
	MX_RECORD **input_device_array;
	double settling_time;
	double integration_time;
	MX_RECORD *timer_record;
	long shutter_policy;
	MX_RECORD *shutter_record;
	MX_DATAFILE *datafile;
	MX_PLOT *plot;
} MX_SCAN;

typedef struct {
	mx_status_type ( *prepare_for_scan_start ) ( MX_SCAN *scan );
	mx_status_type ( *perform_scan ) ( MX_SCAN *scan );
} MX_SCAN_FUNCTION_LIST;

MX_API mx_status_type mx_prepare_for_scan_start( MX_SCAN *scan );

MX_API mx_status_type mx_perform_scan( MX_SCAN *scan );

MX_API mx_status_type mx_scan_measure_data( MX_SCAN *scan );

MX_API mx_status_type mx_scan_display_scan_progress( MX_SCAN *scan );

MX_API mx_status_type mx_clear_scan_input_devices( MX_SCAN *scan );

MX_API mx_status_type mx_scan_fixup_varargs_record_field_defaults(
			MX_RECORD_FIELD_DEFAULTS *record_field_defaults,
			long num_record_field,
			long *num_independent_variables_varargs_cookie,
			long *num_motors_varargs_cookie,
			long *num_input_devices_varargs_cookie );

MX_API mx_status_type mx_convert_device_value_to_string(
				MX_RECORD *input_device,
				char *buffer, size_t buffer_length );

MX_API mx_status_type mx_scan_delete_pointer_from_parent_to_child(
		MX_RECORD *parent_scan_record, MX_RECORD *child_scan_record );

MX_API mx_status_type mx_find_child_scan_record_pointer_in_parent(
		MX_SCAN *child_scan,
		MX_RECORD ***address_of_pointer_to_record_pointer_in_parent );

#define MX_SCAN_STANDARD_FIELDS  \
  {-1, -1, "num_independent_variables", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, \
	offsetof(MX_SCAN, num_independent_variables), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "num_motors", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, num_motors), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "motor_position", MXFT_DOUBLE, NULL, 1, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, motor_position), \
	{sizeof(double)}, NULL, MXFF_VARARGS}, \
  \
  {-1, -1, "motor_array", MXFT_RECORD, NULL, 1, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, motor_record_array), \
	{sizeof(MX_RECORD *)}, NULL, \
	(MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY | MXFF_VARARGS)}, \
  \
  {-1, -1, "motor_is_independent_variable", MXFT_LONG, \
	NULL, 1, {MXU_VARARGS_LENGTH}, MXF_REC_SUPERCLASS_STRUCT, \
			offsetof(MX_SCAN, motor_is_independent_variable), \
	{sizeof(long)}, NULL, MXFF_VARARGS}, \
  \
  {-1, -1, "num_input_devices", MXFT_LONG, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, num_input_devices), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "input_device_array", MXFT_RECORD, NULL, 1, {MXU_VARARGS_LENGTH}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, input_device_array), \
	{sizeof(MX_RECORD *)}, NULL, (MXFF_IN_DESCRIPTION | MXFF_VARARGS)}, \
  \
  {-1, -1, "settling_time", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, settling_time), \
	{0}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "integration_time", MXFT_DOUBLE, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, integration_time), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "timer", MXFT_RECORD, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, timer_record), \
	{sizeof(MX_RECORD *)}, NULL, MXFF_IN_DESCRIPTION}, \
  \
  {-1, -1, "datafile", MXFT_DATAFILE, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, datafile), \
	{0}, NULL, (MXFF_IN_DESCRIPTION | MXFF_IN_SUMMARY)}, \
  \
  {-1, -1, "plot", MXFT_PLOT, NULL, 0, {0}, \
	MXF_REC_SUPERCLASS_STRUCT, offsetof(MX_SCAN, plot), \
	{0}, NULL, MXFF_IN_DESCRIPTION}

#endif /* _MX_SCAN_H_ */

