/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_sclin.c
 *
 * Purpose: Driver for MX_LINEAR_SCAN step scan records.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_scan.h"
#include "mx_sclin.h"

MX_RECORD_FUNCTION_LIST mxs_linear_scan_record_function_list = {
	mxs_linear_scan_initialize_type,
	NULL,
	NULL,
	mxs_linear_scan_create_record_structures,
	mxs_linear_scan_finish_record_initialization,
	mxs_linear_scan_delete_record,
	NULL,
	NULL,
	mxs_linear_scan_dummy_function,
	mxs_linear_scan_dummy_function,
	mxs_linear_scan_dummy_function,
	mxs_linear_scan_dummy_function
};

MX_SCAN_FUNCTION_LIST mxs_linear_scan_scan_function_list = {
	mxs_linear_scan_prepare_for_scan_start,
	mxs_linear_scan_perform_scan
};

/* Prototype an internal use only function. */

static mx_status_type
mxs_linear_scan_perform_scan_level( MX_SCAN *scan,
		MX_LINEAR_SCAN *linear_scan,
		long dimension_level,
		mx_status_type (*compute_motor_positions_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *),
		mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_DISPLAY_MOTOR_MOVE_FPTR, int)
		);

/*=========*/

#define NUM_LINEAR_SCAN_FIELDS	4

MX_EXPORT mx_status_type
mxs_linear_scan_initialize_type( long record_type )
{
	const char fname[] = "mxs_linear_scan_initialize_type()";

	const char field_name[NUM_LINEAR_SCAN_FIELDS][MXU_FIELD_NAME_LENGTH+1]
	    = {
		"start_position",
		"step_size",
		"num_measurements",
		"step_number"};

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	int i;
	long num_record_fields;
	long num_independent_variables_varargs_cookie;
	long num_motors_varargs_cookie;
	long num_input_devices_varargs_cookie;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr
			= driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	/**** Fix up the record fields common to all scan types. ****/

	num_record_fields = *(driver->num_record_fields);

	status = mx_scan_fixup_varargs_record_field_defaults(
			record_field_defaults, num_record_fields,
			&num_independent_variables_varargs_cookie,
			&num_motors_varargs_cookie,
			&num_input_devices_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	/**** Fix up the record fields specific to MX_LINEAR_SCAN records ****/

	for ( i = 0; i < NUM_LINEAR_SCAN_FIELDS; i++ ) {
		status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			field_name[i], &field );

		if ( status.code != MXE_SUCCESS )
			return status;

		field->dimension[0] = num_independent_variables_varargs_cookie;
	}

	return MX_SUCCESSFUL_RESULT;
}

#undef NUM_LINEAR_SCAN_FIELDS		/* Just for safety's sake. */

MX_EXPORT mx_status_type
mxs_linear_scan_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_create_record_structures()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	MX_DRIVER *driver;
	MX_DATAFILE *datafile;
	MX_PLOT *plot;
	mx_status_type (*fptr)( MX_RECORD *, MX_SCAN *, MX_LINEAR_SCAN * );
	mx_status_type status;

	/* Allocate memory for the necessary structures. */

	scan = (MX_SCAN *) malloc( sizeof(MX_SCAN) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAN structure." );
	}

	linear_scan = (MX_LINEAR_SCAN *) malloc( sizeof(MX_LINEAR_SCAN) );

	if ( linear_scan == (MX_LINEAR_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_LINEAR_SCAN structure." );
	}

	datafile = (MX_DATAFILE *) malloc( sizeof(MX_DATAFILE) );

	if ( datafile == (MX_DATAFILE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_DATAFILE structure." );
	}

	plot = (MX_PLOT *) malloc( sizeof(MX_PLOT) );

	if ( plot == (MX_PLOT *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PLOT structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_superclass_struct = scan;
	record->record_class_struct = linear_scan;
	record->record_type_struct = NULL;

	scan->record = record;
	scan->parent_scan_record = NULL;

	scan->datafile = datafile;
	scan->plot = plot;

	datafile->scan = scan;
	plot->scan = scan;

	record->superclass_specific_function_list
				= &mxs_linear_scan_scan_function_list;

	/* Find the MX_LINEAR_SCAN_FUNCTION_LIST pointer. */

	driver = mx_get_driver_by_type( mx_type_list, record->type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_DRIVER for record type %ld is NULL.",
			record->type );
	}

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
			(driver->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST for record type %ld is NULL.",
			record->type );
	}

	record->class_specific_function_list = flist;

	/* Call the type specific 'create_record_structures' function. */

	fptr = flist->create_record_structures;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Linear scan type specific 'create_record_structures' pointer is NULL.");
	}

	status = (*fptr)( record, scan, linear_scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_finish_record_initialization()";

	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_RECORD * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	/* Call the record type specific 'finish_record_initialization'
	 * function.
	 */

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST pointer is NULL.");
	}

	fptr = flist->finish_record_initialization;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Linear scan type specific 'finish_record_initialization' pointer is NULL.");
	}

	status = (*fptr)( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_delete_record( MX_RECORD *record )
{
	const char fname[] = "mxs_linear_scan_delete_record()";

	MX_SCAN *scan;
	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_RECORD * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	scan = (MX_SCAN *)(record->record_superclass_struct);

	if ( scan != NULL ) {
		/* Is this a child scan? */

		if ( scan->parent_scan_record != NULL ) {

			status = mx_scan_delete_pointer_from_parent_to_child(
				scan->parent_scan_record, record );

			switch( status.code ) {
			case MXE_SUCCESS:
				break;
			default:
				return status;
				break;
			}
		}

		/* Now delete the scan's data structures. */

		if ( scan->datafile != NULL ) {
			free( scan->datafile );

			scan->datafile = NULL;
		}

		if ( scan->plot != NULL ) {
			if ( scan->plot->plot_arguments != NULL ) {
				free( scan->plot->plot_arguments );

				scan->plot->plot_arguments = NULL;
			}

			free( scan->plot );

			scan->plot = NULL;
		}

		linear_scan = (MX_LINEAR_SCAN *)
					(scan->record->record_class_struct);

		if ( linear_scan != NULL ) {
			flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

			if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_LINEAR_SCAN_FUNCTION_LIST pointer is NULL.");
			}

			fptr = flist->delete_record;

			/* If fptr is NULL, this record type has no type
			 * specific structures to delete.  Otherwise,
			 * we invoke fptr to delete the type specific
			 * structures.
			 */

			if ( fptr != NULL ) {
				status = (*fptr)( record );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			record->record_type_struct = NULL;
			linear_scan->record_type_struct = NULL;

			free( linear_scan );

			record->record_class_struct = NULL;
		}

		free( scan );

		record->record_superclass_struct = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_dummy_function( MX_RECORD *record )
{
	/* This function serves as a placeholder for functions that are
	 * not needed by this scan class.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_linear_scan_prepare_for_scan_start()";

	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr) (MX_SCAN *);
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked", fname));

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
				(scan->record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN_FUNCTION_LIST pointer for scan '%s' is NULL.",
			scan->record->name );
	}

	fptr = flist->prepare_for_scan_start;

	/* It is not an error for flist->prepare_for_scan_start to be NULL,
	 * this just means that this particular scan type does not need
	 * to do anything special when a scan is starting.
	 */

	if ( fptr != NULL ) {
		status = (*fptr) (scan);

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_linear_scan_perform_scan( MX_SCAN *scan )
{
	const char fname[] = "mxs_linear_scan_perform_scan()";

	MX_LINEAR_SCAN *linear_scan;
	MX_LINEAR_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*compute_motor_positions_fptr)
					( MX_SCAN *, MX_LINEAR_SCAN * );
	mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_DISPLAY_MOTOR_MOVE_FPTR, int);
	mx_status_type status;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_SCAN pointer passed was NULL." );
	}

	linear_scan = (MX_LINEAR_SCAN *)(scan->record->record_class_struct);

	if ( linear_scan == (MX_LINEAR_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_LINEAR_SCAN pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	flist = (MX_LINEAR_SCAN_FUNCTION_LIST *)
			(scan->record->class_specific_function_list);

	if ( flist == (MX_LINEAR_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_LINEAR_SCAN_FUNCTION_LIST ptr for scan record '%s' is NULL.",
			scan->record->name );
	}

	compute_motor_positions_fptr = flist->compute_motor_positions;

	if ( compute_motor_positions_fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"'compute_motor_position' function pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	/* It is not an error for flist->motor_record_array_move_special
	 * to be NULL.  If this function ptr is NULL, this just means that
	 * mx_motor_record_array_move_absolute will be called directly
	 * rather than being preprocessed by motor_record_array_move_special.
	 */

	move_special_fptr = flist->motor_record_array_move_special;

	status = mxs_linear_scan_perform_scan_level( scan, linear_scan,
			scan->num_independent_variables - 1,
			compute_motor_positions_fptr,
			move_special_fptr );

	return status;
}

static mx_status_type
mxs_linear_scan_perform_scan_level( MX_SCAN *scan,
		MX_LINEAR_SCAN *linear_scan,
		long dimension_level,
		mx_status_type (*compute_motor_positions_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *),
		mx_status_type (*move_special_fptr)
					(MX_SCAN *, MX_LINEAR_SCAN *,
					int, MX_RECORD **, double *,
					MX_DISPLAY_MOTOR_MOVE_FPTR, int)
		)
{
	const char fname[] = "mxs_linear_scan_perform_scan_level()";

	long i, array_index, num_dimension_steps, new_dimension_level;
	mx_status_type status;

	array_index = scan->num_independent_variables - dimension_level - 1;

	num_dimension_steps = linear_scan->num_measurements[ array_index ];

	new_dimension_level = dimension_level - 1;

	MX_DEBUG( 2,
("%s: dimension_level = %ld, array_index = %ld, num_dimension_steps = %ld",
		fname, dimension_level, array_index, num_dimension_steps ));

	if ( dimension_level > 0 ) {
		/** More levels in the array to descend through. **/

		MX_DEBUG( 2,("%s: -> Go to next dimension level = %ld",
					fname, new_dimension_level));

		for ( i = 0; i < num_dimension_steps; i++ ) {

			linear_scan->step_number[ array_index ] = i;

			status = mxs_linear_scan_perform_scan_level(scan,
					linear_scan,
					new_dimension_level,
					compute_motor_positions_fptr,
					move_special_fptr);

			if ( status.code != MXE_SUCCESS )
				return status;
		}
	} else {
		/** At bottom level of independent variables. **/

		MX_DEBUG( 2,
		("%s: At bottom level of independent variables.", fname ));

		if ( scan->num_motors > 0 ) {
			mx_info("Moving to start position for scan section.");
		}

		for ( i = 0; i < num_dimension_steps; i++ ) {

			/* Compute the next set of motor positions. */

			linear_scan->step_number[ array_index ] = i;

			status = (*compute_motor_positions_fptr)
						( scan, linear_scan );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* === Move to the computed motor positions. === */

			/* If move_special_fptr is NULL, we just invoke
			 * mx_motor_record_array_move_absolute directly.
			 * Otherwise, the computed move must be preprocessed
			 * by move_special_fptr before the motors are moved.
			 * See how slit scans are implemented in
			 * 'libmx/s_slit.c' for an example of this.
			 */

			if ( move_special_fptr == NULL ) {
				if ( scan->num_motors > 0 ) {
					status = 
					   mx_motor_array_move_absolute(
						scan->num_motors,
						scan->motor_record_array,
						scan->motor_position,
						NULL,
						0 );
				}
			} else {
				status = (*move_special_fptr)
					( scan,
					linear_scan,
					scan->num_motors,
					scan->motor_record_array,
					scan->motor_position,
					NULL,
					0 );
			}

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Perform the measurement. */

			status = mx_scan_measure_data( scan );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Write the result out to the datafile. */

			status = mx_add_measurement_to_datafile(
							scan->datafile );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Report the progress of the scan back to the user. */

			status = mx_scan_display_scan_progress( scan );

			if ( status.code != MXE_SUCCESS )
				return status;

			/* Finally, if enabled, add the measurement to the
			 * running scan plot.
			 */

			if ( mx_plotting_is_enabled( scan->record ) ) {

				if ( i == 0 ) {
					/* Tell the plotting package that
					 * this is a new section of the scan.
					 */

					status = mx_plot_start_plot_section(
							scan->plot );

					if ( status.code != MXE_SUCCESS )
						return status;
				}

				status = mx_add_measurement_to_plot(
							scan->plot );

				if ( status.code != MXE_SUCCESS )
					return status;
			}
		}

		(scan->plot->section_number)++;
	}
	return MX_SUCCESSFUL_RESULT;
}

