/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_scseg.c
 *
 * Purpose: Driver for MX_SEGMENTED_SCAN scan records.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_scan.h"
#include "mx_scseg.h"

MX_RECORD_FUNCTION_LIST mxs_segmented_scan_record_function_list = {
	mxs_segmented_scan_initialize_type,
	NULL,
	NULL,
	mxs_segmented_scan_create_record_structures,
	mxs_segmented_scan_finish_record_initialization,
	mxs_segmented_scan_delete_record,
	NULL,
	NULL,
	mxs_segmented_scan_dummy_function,
	mxs_segmented_scan_dummy_function,
	mxs_segmented_scan_dummy_function,
	mxs_segmented_scan_dummy_function
};

MX_SCAN_FUNCTION_LIST mxs_segmented_scan_scan_function_list = {
	mxs_segmented_scan_prepare_for_scan_start,
	mxs_segmented_scan_perform_scan
};

MX_EXPORT mx_status_type
mxs_segmented_scan_initialize_type( long record_type )
{
	const char fname[] = "mxs_segmented_scan_initialize_type()";

	MX_DRIVER *driver;
	MX_RECORD_FIELD_DEFAULTS *record_field_defaults;
	MX_RECORD_FIELD_DEFAULTS **record_field_defaults_ptr;
	MX_RECORD_FIELD_DEFAULTS *field;
	long num_record_fields;
	long num_independent_variables_varargs_cookie;
	long num_motors_varargs_cookie;
	long num_input_devices_varargs_cookie;

	long num_segments_field_index;
	long num_segments_varargs_cookie;
	mx_status_type status;

	driver = mx_get_driver_by_type( mx_type_list, record_type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Record type %ld not found in mx_type_list.",
			record_type );
	}

	record_field_defaults_ptr = driver->record_field_defaults_ptr;

	if (record_field_defaults_ptr == (MX_RECORD_FIELD_DEFAULTS **) NULL) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	record_field_defaults = *record_field_defaults_ptr;

	if ( record_field_defaults == (MX_RECORD_FIELD_DEFAULTS *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'record_field_defaults_ptr' for record type '%s' is NULL.",
			driver->name );
	}

	if ( driver->num_record_fields == (long *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"'num_record_fields' pointer for record type '%s' is NULL.",
			driver->name );
	}

	/**** Fix up the record fields common to all scan types. ****/

	num_record_fields = *(driver->num_record_fields);

	status = mx_scan_fixup_varargs_record_field_defaults(
			record_field_defaults, num_record_fields,
			&num_independent_variables_varargs_cookie,
			&num_motors_varargs_cookie,
			&num_input_devices_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	/** Fix up the record fields specific to MX_SEGMENTED_SCAN records **/

	status = mx_find_record_field_defaults_index(
			record_field_defaults, num_record_fields,
			"num_segments", &num_segments_field_index );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_construct_varargs_cookie(
		num_segments_field_index, 0, &num_segments_varargs_cookie );

	if ( status.code != MXE_SUCCESS )
		return status;

	status = mx_find_record_field_defaults(
			record_field_defaults, num_record_fields,
			"scan_segment_array", &field );

	if ( status.code != MXE_SUCCESS )
		return status;

	field->dimension[0] = num_segments_varargs_cookie;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxs_segmented_scan_create_record_structures()";

	MX_SCAN *scan;
	MX_SEGMENTED_SCAN *segmented_scan;
	MX_SEGMENTED_SCAN_FUNCTION_LIST *flist;
	MX_DRIVER *driver;
	MX_DATAFILE *datafile;
	MX_PLOT *plot;
	mx_status_type (*fptr)( MX_RECORD *, MX_SCAN *, MX_SEGMENTED_SCAN * );
	mx_status_type status;

	/* Allocate memory for the necessary structures. */

	scan = (MX_SCAN *) malloc( sizeof(MX_SCAN) );

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SCAN structure." );
	}

	segmented_scan
		= (MX_SEGMENTED_SCAN *) malloc( sizeof(MX_SEGMENTED_SCAN) );

	if ( segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SEGMENTED_SCAN structure." );
	}

	datafile = (MX_DATAFILE *) malloc( sizeof(MX_DATAFILE) );

	if ( datafile == (MX_DATAFILE *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_DATAFILE structure." );
	}

	plot = (MX_PLOT *) malloc( sizeof(MX_PLOT) );

	if ( plot == (MX_PLOT *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_PLOT structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_superclass_struct = scan;
	record->record_class_struct = segmented_scan;
	record->record_type_struct = NULL;	/* Not used by this class. */

	scan->record = record;
	scan->parent_scan_record = NULL;

	scan->datafile = datafile;
	scan->plot = plot;

	datafile->scan = scan;
	plot->scan = scan;

	record->superclass_specific_function_list
				= &mxs_segmented_scan_scan_function_list;

	/* Find the MX_SEGMENTED_SCAN_FUNCTION_LIST pointer. */

	driver = mx_get_driver_by_type( mx_type_list,
				record->type );

	if ( driver == (MX_DRIVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TYPE_LIST_ENTRY for record type %ld is NULL.",
			record->type );
	}

	flist = (MX_SEGMENTED_SCAN_FUNCTION_LIST *)
			(driver->class_specific_function_list);

	if ( flist == (MX_SEGMENTED_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_SEGMENTED_SCAN_FUNCTION_LIST for record type %ld is NULL.",
			record->type );
	}

	record->class_specific_function_list = flist;

	/* Call the type specific 'create_record_structures' function. */

	fptr = flist->create_record_structures;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Segmented scan type specific 'create_record_structures' pointer is NULL.");
	}

	status = (*fptr)( record, scan, segmented_scan );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_finish_record_initialization( MX_RECORD *record )
{
	const char fname[]="mxs_segmented_scan_finish_record_initialization()";

	MX_SCAN *scan;
	MX_SEGMENTED_SCAN *segmented_scan;
	MX_RECORD *child_scan_record;
	MX_SCAN *child_scan;
	MX_SEGMENTED_SCAN_FUNCTION_LIST *flist;
	long i;
	mx_status_type (*fptr)( MX_RECORD * );
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked for record '%s'.", fname, record->name));

	/* Setup 'parent_scan' pointers in each of the children of this
	 * segmented scan.
	 */

	scan = (MX_SCAN *)(record->record_superclass_struct);

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SCAN pointer for scan '%s' is NULL.",
			record->name );
	}

	segmented_scan = (MX_SEGMENTED_SCAN *)(record->record_class_struct);

	if ( segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SEGMENTED_SCAN pointer for scan '%s' is NULL.",
			record->name );
	}

	for ( i = 0; i < segmented_scan->num_segments; i++ ) {
		child_scan_record
			= (segmented_scan->scan_segment_record_array)[i];

		if ( child_scan_record == (MX_RECORD *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"Child scan record %ld for segmented scan '%s' is NULL.",
				i, record->name );
		}

		child_scan =
		    (MX_SCAN *)(child_scan_record->record_superclass_struct);
		
		if ( child_scan == (MX_SCAN *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MX_SCAN pointer for child scan record '%s' of segmented scan '%s' is NULL.",
				child_scan_record->name, record->name );
		}

		child_scan->parent_scan_record = record;
	}

	/* Call the record type specific 'finish_record_initialization'
	 * function.
	 */

	flist = (MX_SEGMENTED_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

	if ( flist == (MX_SEGMENTED_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_SEGMENTED_SCAN_FUNCTION_LIST pointer is NULL.");
	}

	fptr = flist->finish_record_initialization;

	if ( fptr == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"Segmented scan type specific 'finish_record_initialization' pointer is NULL.");
	}

	status = (*fptr)( record );

	if ( status.code != MXE_SUCCESS )
		return status;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_delete_record( MX_RECORD *record )
{
	const char fname[] = "mxs_segmented_scan_delete_record()";

	MX_SCAN *scan, *child_scan;
	MX_RECORD **scan_segment_record_array;
	MX_SEGMENTED_SCAN *segmented_scan;
	MX_SEGMENTED_SCAN_FUNCTION_LIST *flist;
	mx_status_type (*fptr)( MX_RECORD * );
	long i, num_segments;
	mx_status_type status;

	MX_DEBUG( 2,("%s invoked.", fname));

	if ( record == NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	scan = (MX_SCAN *)(record->record_superclass_struct);

	if ( scan != NULL ) {
		/* Is this a child scan? */

		if ( scan->parent_scan_record != NULL ) {

			status = mx_scan_delete_pointer_from_parent_to_child(
				scan->parent_scan_record, record );

			switch( status.code ) {
			case MXE_SUCCESS:
				break;
			default:
				return status;
				break;
			}
		}

		/* Now delete the scan's data structures. */

		if ( scan->datafile != NULL ) {
			free( scan->datafile );

			scan->datafile = NULL;
		}

		if ( scan->plot != NULL ) {
			if ( scan->plot->plot_arguments != NULL ) {
				free( scan->plot->plot_arguments );

				scan->plot->plot_arguments = NULL;
			}

			free( scan->plot );

			scan->plot = NULL;
		}

		segmented_scan = (MX_SEGMENTED_SCAN *)
					scan->record->record_class_struct;

		if ( segmented_scan != NULL ) {

			/* If the scan has a list of children available,
			 * delete them too.
			 */

			num_segments = segmented_scan->num_segments;

			scan_segment_record_array =
				    segmented_scan->scan_segment_record_array;

			if ( num_segments > 0 ) {
				if (scan_segment_record_array
							== (MX_RECORD **)NULL){
					return mx_error(
					    MXE_CORRUPT_DATA_STRUCTURE, fname,
				"scan_segment_record_array pointer is NULL.");
				}
			}

			for ( i = 0; i < num_segments; i++ ) {
				if ( scan_segment_record_array[i] == NULL ) {
					/* Just skip NULL entries.  There's
					 * not much we could do about them
					 * anyway.
					 */

					continue;  /* cycle the for() loop. */
				}

				/* Set the parent_scan_record pointer in the
				 * child to NULL before invoking the delete
				 * function for the scan.  Otherwise, the
				 * child will attempt to modify the 
				 * scan_segment_record_array in the parent
				 * to reflect the fact that it has been
				 * deleted.  Since we are also using the
				 * scan_segment_record_array pointer in
				 * the body of the current loop, we must
				 * suppress this behavior on the part
				 * of the child.
				 */

				child_scan = (MX_SCAN *)
		(scan_segment_record_array[i]->record_superclass_struct);

				child_scan->parent_scan_record = NULL;

				/* Now it is safe to invoke the delete
				 * function for the child scan record.
				 */

				status = mx_delete_record(
					scan_segment_record_array[i] );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			/* Note that it is possible to have num_segments == 0,
			 * but scan_segment_record_array != NULL.  Normally,
			 * this will only happen during a modify scan
			 * operation.  This is because 'modify scan' has
			 * been implemented by deleting the old record
			 * and putting the new record in its place.
			 *
			 * During this operation, num_segments is set to 0
			 * so that the children of a parent scan are not
			 * deleted when the original parent is deleted.
			 * scan_segment_record_array will still be non-NULL
			 * in this case, and still needs to be freed.
			 *
			 * Regardless, of all this, if the pointer
			 * scan_segment_record_array is not NULL, then
			 * scan_segment_record_array will be freed
			 * no matter what the value of num_segments is.
			 */

			if (scan_segment_record_array != (MX_RECORD **) NULL){
				free( scan_segment_record_array );
			}

			/****/

			flist = (MX_SEGMENTED_SCAN_FUNCTION_LIST *)
				(record->class_specific_function_list);

			if (flist == (MX_SEGMENTED_SCAN_FUNCTION_LIST *) NULL){
				return mx_error(
					MXE_CORRUPT_DATA_STRUCTURE, fname,
			"MX_SEGMENTED_SCAN_FUNCTION_LIST pointer is NULL.");
			}

			fptr = flist->delete_record;

			/* If fptr is NULL, this record type has no type
			 * specific structures to delete.  Otherwise,
			 * we invoke fptr to delete the type specific
			 * structures.
			 */

			if ( fptr != NULL ) {
				status = (*fptr)( record );

				if ( status.code != MXE_SUCCESS )
					return status;
			}

			record->record_type_struct = NULL;
			segmented_scan->record_type_struct = NULL;

			free( segmented_scan );

			record->record_class_struct = NULL;
		}

		free( scan );

		scan = NULL;
	}
	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_dummy_function( MX_RECORD *record )
{
	/* This function serves as a placeholder for functions that are
	 * not needed by this scan class.
	 */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_prepare_for_scan_start( MX_SCAN *scan )
{
	const char fname[] = "mxs_segmented_scan_prepare_for_scan_start()";

	MX_DEBUG( 2,("%s invoked", fname));

	/* This function may be superfluous. */

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxs_segmented_scan_perform_scan( MX_SCAN *scan )
{
	const char fname[] = "mxs_segmented_scan_perform_scan()";

	MX_SEGMENTED_SCAN *segmented_scan;
	MX_SEGMENTED_SCAN_FUNCTION_LIST *flist;
	MX_RECORD *scan_segment_record;
	MX_SCAN *scan_segment;
	long i;
	mx_status_type status;

	if ( scan == (MX_SCAN *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_SCAN pointer passed was NULL." );
	}

	segmented_scan = (MX_SEGMENTED_SCAN *)
				scan->record->record_class_struct;

	if ( segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_SEGMENTED_SCAN pointer for scan record '%s' is NULL.",
			scan->record->name );
	}

	flist = (MX_SEGMENTED_SCAN_FUNCTION_LIST *)
			(scan->record->class_specific_function_list);

	if ( flist == (MX_SEGMENTED_SCAN_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MX_SEGMENTED_SCAN_FUNCTION_LIST ptr for scan record '%s' is NULL.",
			scan->record->name );
	}

	for ( i = 0; i < segmented_scan->num_segments; i++ ) {
		scan_segment_record
			= (segmented_scan->scan_segment_record_array)[i];

		if ( scan_segment_record == (MX_RECORD *) NULL ) {
#if 0
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"Scan segment %ld has a NULL record pointer.", i );
#else
			mx_info(
			"Scan segment %ld has been deleted.  Skipping...", i);

			/* Cycle back to the beginning of the for() loop. */

			continue;
#endif
		}

		MX_DEBUG( 2,("%s: *** Scan segment %ld, Name = '%s' ***",
			fname, i, scan_segment_record->name));

		scan_segment =
		  (MX_SCAN *)(scan_segment_record->record_superclass_struct);

		if ( scan_segment == (MX_SCAN *) NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"Scan segment %ld ('%s') has a NULL MX_SCAN structure pointer.",
				i, scan_segment_record->name );
		}

		status = mx_perform_scan( scan_segment );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	return MX_SUCCESSFUL_RESULT;
}

