/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_timer.c
 *
 * Purpose: MX function library of generic timer operations.
 */

#include <stdio.h>
#include <stdlib.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_drivr.h"
#include "mx_timer.h"

MX_EXPORT mx_status_type
mx_timer_finish_record_initialization( MX_RECORD *timer_record )
{
	const char fname[] = "mx_timer_finish_record_initialization()";

	MX_TIMER *timer;

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	timer = (MX_TIMER *)(timer_record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER pointer for record '%s' is NULL.",
			timer_record->name );
	}

	timer->busy = 0;
	timer->start = 0.0;
	timer->stop = 0;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_timer_is_busy( MX_RECORD *timer_record, int *busy )
{
	const char fname[] = "mx_timer_is_busy()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_is_busy_fn ) ( MX_TIMER *, int * );
	mx_status_type status;

	MX_DEBUG(2, ("%s invoked.", fname));

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	timer = (MX_TIMER *)(timer_record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER pointer for record '%s' is NULL.",
			timer_record->name );
	}

	function_list = (MX_TIMER_FUNCTION_LIST *)
				(timer_record->class_specific_function_list);

	if ( function_list == (MX_TIMER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER_FUNCTION_LIST ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	timer_is_busy_fn = function_list->timer_is_busy;

	if ( timer_is_busy_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"timer_is_busy function ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	/* All is well, so invoke the function. */

	status = (*timer_is_busy_fn)( timer, busy );

	return status;
}

MX_EXPORT mx_status_type
mx_timer_start( MX_RECORD *timer_record, double seconds )
{
	const char fname[] = "mx_timer_start()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_start ) ( MX_TIMER *, double );
	mx_status_type status;

	MX_DEBUG(2, ("%s invoked.", fname));

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	timer = (MX_TIMER *)(timer_record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER pointer for record '%s' is NULL.",
			timer_record->name );
	}

	function_list = (MX_TIMER_FUNCTION_LIST *)
				(timer_record->class_specific_function_list);

	if ( function_list == (MX_TIMER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER_FUNCTION_LIST ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	timer_start = function_list->start;

	if ( timer_start == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"timer_is_busy function ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	/* All is well, so invoke the function. */

	status = (*timer_start)( timer, seconds );

	return status;
}

MX_EXPORT mx_status_type
mx_timer_stop( MX_RECORD *timer_record, double *seconds_left )
{
	const char fname[] = "mx_timer_stop()";

	MX_TIMER *timer;
	MX_TIMER_FUNCTION_LIST *function_list;
	mx_status_type ( *timer_stop ) ( MX_TIMER *, double * );
	mx_status_type status;

	MX_DEBUG( 2, ("%s invoked.", fname));

	if ( timer_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed is NULL.");
	}

	timer = (MX_TIMER *)(timer_record->record_class_struct);

	if ( timer == (MX_TIMER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER pointer for record '%s' is NULL.",
			timer_record->name );
	}

	function_list = (MX_TIMER_FUNCTION_LIST *)
				(timer_record->class_specific_function_list);

	if ( function_list == (MX_TIMER_FUNCTION_LIST *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_TIMER_FUNCTION_LIST ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	timer_stop = function_list->stop;

	if ( timer_stop == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"timer_is_busy function ptr for MX_TIMER ptr 0x%p is NULL.",
			timer);
	}

	/* All is well, so invoke the function. */

	status = (*timer_stop)( timer, seconds_left );

	return status;
}

