/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    mx_util.c
 *
 * Purpose: Define some of the utility functions.
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <float.h>
#include <math.h>

#ifdef OS_WIN32
#include <winsock.h>
#endif

#include "mx_util.h"
#include "mx_rec.h"

MX_EXPORT int
mx_get_max_file_descriptors( void )
{
	int result;

	/* This operation varies from OS to OS, so I will fill this in as time
	 * goes on.
	 */

#if defined( OS_LINUX ) || defined( OS_SOLARIS2 ) \
      || defined( OS_SGI ) || defined( OS_SUNOS4 )

	result = getdtablesize();

#elif defined( OS_WIN32 )

	result = FD_SETSIZE;

#elif defined( OS_DJGPP )

	result = 64;

#else

#error mx_get_max_file_descriptors() has not been implemented for this platform.

#endif

	return result;
}

MX_EXPORT long
mx_round( double value )
{
	long result;

	if ( value >= 0.0 ) {
		result = (long) ( 0.5 + value );
	} else {
		result = (long) ( -0.5 + value );
	}

	return result;
}

/*
 * mx_divide_safely divides two floating point numbers by each other
 * while testing for and avoiding division by zero.
 */

MX_EXPORT double
mx_divide_safely( double numerator, double denominator )
{
	double result;

	if ( fabs(denominator) >= 1.0 ) {
		result = numerator / denominator;
	} else {
		if ( fabs(numerator) < fabs(denominator * DBL_MAX) ) {
			result = numerator / denominator;
		} else {
			if ( (numerator >= 0.0 && denominator >= 0.0)
			  || (numerator < 0.0 && denominator < 0.0) )
			{
				result = DBL_MAX;
			} else {
				result = -DBL_MAX;
			}
		}
	}
	return result;
}

/* mx_difference() computes a relative difference function.
 *
 * Derived from pages 250-251 in "C Programming FAQs" by Steve Summit,
 * Addison Wesley, 1996.
 */

MX_EXPORT double
mx_difference( double value1, double value2 )
{
	double abs1, abs2, maxabs, result;

	if ( value1 >= 0.0 ) {
		abs1 = value1;
	} else {
		abs1 = - value1;
	}

	if ( value2 >= 0.0 ) {
		abs2 = value2;
	} else {
		abs2 = - value2;
	}

	if ( abs1 >= abs2 ) {
		maxabs = abs1;
	} else {
		maxabs = abs2;
	}
	
	if ( maxabs == 0.0 ) {
		result = 0.0;
	} else {
		result = ( value1 - value2 ) / maxabs;

		if ( result < 0.0 ) {
			result = - result;
		}
	}

	return result;
}

/* Another function from "C Programming FAQs" is the following simple
 * wildcard matcher from page 228 which is attributed to Arjan Kenter.
 *
 * mx_match() does what I need without requiring the inclusion
 * of complicated regular expression matching functions.
 *
 * As an example, the call mx_match( "a*b.c", "aplomb.c" ) would return 1.
 */

MX_EXPORT int
mx_match( char *pattern, char *string )
{
	switch( *pattern ) {
	case '\0':	return ! *string;

	case '*':	return mx_match( pattern+1, string ) ||
				( *string && mx_match( pattern, string+1 ) );

	case '?':	return *string && mx_match( pattern+1, string+1 );

	default:	return *pattern == *string &&
					mx_match( pattern+1, string+1 );
	}
}

MX_EXPORT char *
mx_current_time_string( void )
{
	char *ptr;
	size_t length;
	time_t seconds_since_1970_started;

	seconds_since_1970_started = time(NULL);

	ptr = ctime( &seconds_since_1970_started );

	/* Zap the newline at the end of the string returned by ctime(). */

	length = strlen(ptr);

	ptr[length - 1] = '\0';

	return ptr;
}

MX_EXPORT char *
mx_skip_string_fields( char *buffer, int num_fields )
{
	char *ptr;
	int num_skipped;
	size_t segment_length;

	ptr = buffer;

	/* Skip over any leading whitespace. */

	segment_length = strspn( ptr, MX_RECORD_FIELD_SEPARATORS );

	ptr += segment_length;

	if ( *ptr == '\0' )
		return ptr;

	/* Now loop over string fields. */

	num_skipped = 0;

	while ( num_skipped < num_fields ) {

		/* Skip the string. */

		segment_length = strcspn( ptr, MX_RECORD_FIELD_SEPARATORS );

		ptr += segment_length;

		num_skipped++;

		if ( *ptr == '\0' )
			return ptr;

		/* Skip the trailing whitespace. */

		segment_length = strspn( ptr, MX_RECORD_FIELD_SEPARATORS );

		ptr += segment_length;

		if ( *ptr == '\0' )
			return ptr;
	}

	return ptr;
}

