/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mcopy.c -- Motor scan copy function. */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "motor.h"
#include "mx_scseg.h"

static int
motor_copy_scan_record( MX_RECORD *old_scan_record,
		MX_RECORD **new_scan_record, char *new_scan_name );


int
motor_copy_fn( int argc, char *argv[] )
{
	const char cname[] = "copy";

	MX_RECORD *old_scan_record, *new_scan_record;
	MX_SCAN *old_scan;
	MX_SEGMENTED_SCAN *old_segmented_scan;
	MX_SEGMENTED_SCAN *new_segmented_scan;
	MX_RECORD **old_scan_segment_record_array;
	MX_RECORD **new_scan_segment_record_array;
	MX_RECORD **new_child_record_list;
	MX_RECORD *old_child_record, *new_child_record;
	MX_SCAN *new_child_scan;
	char new_child_name[MXU_RECORD_NAME_LENGTH + 1];
	char *old_scan_name, *new_scan_name;
	char buffer[MXU_FILENAME_LENGTH + 12];
	long i, num_segments;
	size_t length;
	int status;
	mx_status_type mx_status;

	static char usage[]
		= "Usage:  copy scan 'scan_name' 'new_scan_name'\n";

	if ( argc <= 2 ) {
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	length = strlen( argv[2] );

	if ( length <= 0 )
		length = 1;

	if ( strncmp( argv[2], "scan", length ) == 0 )
	{
		switch( argc ) {
		case 3:
			fprintf( output,
			    "%s: No source or destination scan name given\n",
				cname );
			fprintf( output, "%s\n", usage );
			return FAILURE;
		case 4:
			fprintf( output,"%s: No destination scan name given\n",
				cname );
			fprintf( output, "%s\n", usage );
			return FAILURE;
		case 5:
			old_scan_name = argv[3];
			new_scan_name = argv[4];
			break;
		default:
			fprintf( output,
			"%s: Too many arguments on command line.\n", cname );
			fprintf( output, "%s\n", usage );
			return FAILURE;
		}

		/* Does an old scan record with this name exist? */

		old_scan_record = mx_get_record(
					motor_record_list, old_scan_name );

		if ( old_scan_record == (MX_RECORD *) NULL ) {
			fprintf( output,
			"%s: Old scan record '%s' does not exist.\n",
				cname, old_scan_name );
			return FAILURE;
		}

		if ( old_scan_record->superclass != MXR_SCAN ) {
			fprintf( output,
			"%s: The old record '%s' is not a scan record.\n",
				cname, old_scan_name );
			return FAILURE;
		}

		old_scan = (MX_SCAN *)
				(old_scan_record->record_superclass_struct);

		if ( old_scan == (MX_SCAN *) NULL ) {
			fprintf( output,
				"MX_SCAN structure for scan '%s' is NULL.\n",
				old_scan_name );
			return FAILURE;
		}

		/* Is this a child scan?  Child scans should not be
		 * directly copied.
		 */

		if ( old_scan->parent_scan_record != NULL ) {
			fprintf( output,
	"'%s' is a child scan record and cannot be copied directly.\n",
				old_scan_name );
			fprintf( output,
	"Its parent scan appears to be '%s'.  Copy that instead.\n",
				old_scan->parent_scan_record->name );
			return FAILURE;
		}

		/*** If this is not a segmented scan, then just copy
		 *** the record directly.
		 ***/

		if ( old_scan_record->class != MXS_SEGMENTED_SCAN ) {
			status = motor_copy_scan_record( old_scan_record,
					&new_scan_record, new_scan_name );

			if ( status != SUCCESS )
				return status;

			mx_status = mx_finish_record_initialization(
							new_scan_record );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;
		} else {
			old_segmented_scan = (MX_SEGMENTED_SCAN *)
				(old_scan_record->record_class_struct);

			if (old_segmented_scan == (MX_SEGMENTED_SCAN *)NULL){
				fprintf( output,
			"MX_SEGMENTED_SCAN pointer for scan '%s' is NULL!\n",
					old_scan_record->name );
				return FAILURE;
			}

			num_segments = old_segmented_scan->num_segments;

			old_scan_segment_record_array =
				old_segmented_scan->scan_segment_record_array;

			if ( old_scan_segment_record_array
					== (MX_RECORD **) NULL ){
				fprintf( output,
		"scan_segment_record_array pointer for scan '%s' is NULL.\n",
					old_scan_record->name );
				return FAILURE;
			}

			new_child_record_list = (MX_RECORD **)
				malloc( num_segments * sizeof(MX_RECORD *) );

			if ( new_child_record_list == (MX_RECORD **) NULL ) {
				fprintf( output,
		"Ran out of memory allocating new_child_record_list array.\n");
				return FAILURE;
			}

			/*** Copy the child scan records before copying
			 *** the parent record.
			 ***/

			for ( i = 0; i < num_segments; i++ ) {
				old_child_record
					= old_scan_segment_record_array[i];

				/* Generate the new child's name. */
				sprintf( new_child_name, "%s.%03ld",
					new_scan_name, i );

				/* Copy the child scan. */

				status = motor_copy_scan_record(
						old_child_record,
						&new_child_record,
						new_child_name );

				if ( status != SUCCESS ) {
					free( new_child_record_list );

					return status;
				}

				/* The new child copy may have MX_DATAFILE
				 * and MX_PLOT fields that refer to the
				 * old parent scan's record name.  We must
				 * fix this up before finishing record
				 * initialization of the child record.
				 */

				new_child_scan = (MX_SCAN *)
				  (new_child_record->record_superclass_struct);

				if ( new_child_scan->datafile->type
							== MXDF_CHILD ) {
				    strcpy(
					new_child_scan->datafile->filename,"");
				    strncat(
					new_child_scan->datafile->filename,
					new_scan_name, MXU_FILENAME_LENGTH );
				}

				if ( new_child_scan->plot->type
							== MXP_CHILD ) {

				    if ( new_child_scan->plot->plot_arguments
						!= NULL ) {
					free(
					 new_child_scan->plot->plot_arguments);
				    }

				    new_child_scan->plot->plot_arguments
					= (char *) malloc(
					    strlen(new_scan_name) + 1 );

				    if ( new_child_scan->plot->plot_arguments
						== NULL ) {
					fprintf( output,
	"Ran out of memory allocating plot arguments field for child plot.\n");
					free( new_child_record_list );
					return FAILURE;
				    }

				    strcpy(
					new_child_scan->plot->plot_arguments,
					new_scan_name );
				}

				/* Now it is safe to finish the record setup.*/

				mx_status = mx_finish_record_initialization(
							new_child_record );

				if ( mx_status.code != MXE_SUCCESS ) {
					free( new_child_record_list );
					return FAILURE;
				}

				new_child_record_list[i] = new_child_record;
			}

			/*** Now that the children are set up, we may now
			 *** set up the parent scan record.
			 ***/

			status = motor_copy_scan_record( old_scan_record,
				&new_scan_record, new_scan_name );

			if ( status != SUCCESS ) {
				free( new_child_record_list );
				return status;
			}

			/* At this point, the new scan record will have
			 * its scan_segment_record_array elements
			 * pointing to child records in the old segmented
			 * scan.  We must fix this before finishing the
			 * record initialization, since that step will
			 * overwrite the scan->parent_scan_record pointer
			 * in each of the children's records.
			 */

			for ( i = 0; i < num_segments; i++ ) {
				new_segmented_scan = (MX_SEGMENTED_SCAN *)
				    (new_scan_record->record_class_struct);

				new_scan_segment_record_array = (MX_RECORD **)
			    (new_segmented_scan->scan_segment_record_array);

				new_scan_segment_record_array[i]
					= new_child_record_list[i];
			}

			free( new_child_record_list );

			/* Now we can finish the record setup of the parent. */

			mx_status = mx_finish_record_initialization(
							new_scan_record );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;
		}
	} else {
		fprintf( output,
		"%s: unrecognized argument '%s'\n\n", cname, argv[2]);
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	if ( do_scan_database_updates ) {
		sprintf(buffer, "save scan %s", scan_savefile);

		cmd_execute_command_line( command_list_length,
						command_list, buffer );

		motor_has_unsaved_scans = FALSE;
	} else {
		motor_has_unsaved_scans = TRUE;
	}

	return SUCCESS;
}

int
motor_copy_scan_record( MX_RECORD *old_scan_record,
			MX_RECORD **new_scan_record, char *new_scan_name )
{
	const char fname[] = "motor_copy_scan_record()";

	MX_RECORD *record;
	char old_scan_description[MXU_RECORD_DESCRIPTION_LENGTH+1];
	char new_scan_description[MXU_RECORD_DESCRIPTION_LENGTH+1];
	int i, new_scan_name_length;
	mx_status_type mx_status;

	*new_scan_record = NULL;

	/* Does a record with the new scan's name already exist? */

	record = mx_get_record( motor_record_list, new_scan_name );

	if ( record != (MX_RECORD *) NULL ) {
		/* If the record that previously had the same
		 * name as 'new_scan_name', is a scan record,
		 * then delete it.  Otherwise, give an error
		 * message.
		 */

		if ( record->superclass != MXR_SCAN ) {
			fprintf( output,
	"%s: The record '%s' already exists and is not a scan record.\n",
					fname, new_scan_name );
			fprintf( output,
"Cannot overwrite a previously existing record unless it is a scan record.\n");
			return FAILURE;
		}

		mx_status = mx_delete_record( record );

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;
	}

	/* Get the record description for this old record. */

	mx_status = mx_create_description_from_record(
		old_scan_record,
		old_scan_description,
		sizeof(old_scan_description)-1 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	/* Create the new scan description from the old one. */

	strcpy( new_scan_description, old_scan_description );

	new_scan_name_length = strlen( new_scan_name );

	if ( new_scan_name_length > MXU_RECORD_NAME_LENGTH ) {
		new_scan_name_length = MXU_RECORD_NAME_LENGTH;
	}

	for ( i = 0; i < new_scan_name_length; i++ ) {
		new_scan_description[i] = new_scan_name[i];
	}
	for (i = new_scan_name_length; i<MXU_RECORD_NAME_LENGTH; i++) {
		new_scan_description[i] = ' ';
	}

	/* Create a scan record from the newly formatted scan 
	 * description string.
	 */

	mx_status = mx_create_record_from_description(
		motor_record_list,
		new_scan_description,
		new_scan_record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

