/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mcount.c - Scaler count function. */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "motor.h"
#include "mx_ain.h"
#include "mx_din.h"

int
motor_count_fn( int argc, char *argv[] )
{
	const char cname[] = "count";

	MX_RECORD **input_device_array;
	MX_RECORD *input_device;
	MX_RECORD *timer_record;
	int i, j, num_input_devices, busy;
	double counting_time, remaining_time, double_value;
	unsigned long ulong_value;
	long long_value;
	mx_status_type status;

	static char usage[]
		= "Usage:  count 'counting_time' 'scaler1' 'scaler2' ...\n";

	if ( argc <= 3 ) {
		fprintf( output, "%s\n", usage );
		return FAILURE;
	}

	counting_time = atof( argv[2] );

	/* Do we have a timer named 'timer1'? */

	timer_record = mx_get_record( motor_record_list, "timer1" );

	if ( timer_record == NULL ) {
		fprintf( output, "Timer record 'timer1' does not exist.\n" );
		return FAILURE;
	}

	/* Create an array of the input device records. */

	num_input_devices = argc - 3;

	input_device_array = (MX_RECORD **) malloc( num_input_devices
						* sizeof(MX_RECORD *) );

	if ( input_device_array == NULL ) {
		fprintf( output,
		"Attempt to allocate array of input devices failed.\n" );

		return FAILURE;
	}

	for ( i = 3; i < argc; i++ ) {
		j = i - 3;

		input_device_array[j] = mx_get_record( motor_record_list,
							argv[i] );

		if ( input_device_array[j] == NULL ) {
			fprintf( output,
			"Input device record '%s' does not exist.\n", argv[i]);

			free( input_device_array );
			return FAILURE;
		}
		if ( input_device_array[j]->superclass != MXR_DEVICE ) {
			fprintf( output,
			"Record '%s' is not an input device.\n", argv[i]);

			free( input_device_array );
			return FAILURE;
		}

		/* Clear any input devices that need it. */

		switch( input_device_array[j]->class ) {
		case MXC_SCALER:
			status = mx_scaler_clear( input_device_array[j] );

			if ( status.code != MXE_SUCCESS ) {
				free( input_device_array );
				return FAILURE;
			}
			break;
		case MXC_ANALOG_INPUT:
		case MXC_DIGITAL_INPUT:
		case MXC_MOTOR:
			break;
		default:
			fprintf( output,
			"Record '%s' cannot be used by the %s command.\n", 
				argv[i], cname );
			free( input_device_array );
			return FAILURE;
			break;
		}
	}

	/* Perform the measurement. */

	status = mx_timer_start( timer_record, counting_time );

	if ( status.code != MXE_SUCCESS ) {
		free( input_device_array );
		return FAILURE;
	}

	busy = TRUE;

	while ( busy ) {
		if ( mx_user_requested_interrupt() ) {
			(void) mx_timer_stop( timer_record, &remaining_time );

			free( input_device_array );

			fprintf( output,
			"Measurement integration time was interrupted.\n" );
		}
		status = mx_timer_is_busy( timer_record, &busy );

		if ( status.code != MXE_SUCCESS ) {
			free( input_device_array );
			return FAILURE;
		}
		mx_msleep(1);
	}

	/* Read out the input devices. */

	for ( i = 0; i < num_input_devices; i++ ) {
		input_device = input_device_array[i];

		switch( input_device->class ) {
		case MXC_ANALOG_INPUT:
			status = mx_analog_input_read(
						input_device, &double_value );
			if ( status.code == MXE_SUCCESS ) {
				fprintf( output, "%.*g ",
						input_device->precision,
						double_value );
			}
			break;
		case MXC_DIGITAL_INPUT:
			status = mx_digital_input_read(
						input_device, &ulong_value );
			if ( status.code == MXE_SUCCESS ) {
				fprintf( output, "%lu ", ulong_value );
			}
			break;
		case MXC_MOTOR:
			status = mx_motor_get_position(
						input_device, &double_value );
			if ( status.code == MXE_SUCCESS ) {
				fprintf( output, "%.*g ",
						input_device->precision,
						double_value );
			}
			break;
		case MXC_SCALER:
			status = mx_scaler_read( input_device, &long_value );
			if ( status.code == MXE_SUCCESS ) {
				fprintf( output, "%lu ", long_value );
			}
			break;
		}
		if ( status.code != MXE_SUCCESS ) {
			free( input_device_array );
			return FAILURE;
		}
	}
	fprintf( output, "\n" );

	free( input_device_array );

	return SUCCESS;
}

