/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* msclst.c  --  Motor list scan setup and modify function. */

/* If old_scan == NULL, we are doing a new scan setup.  Otherwise, we are
 * modifying a previous scan.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "motor.h"
#include "mdialog.h"
#include "mx_alloc.h"
#include "mx_sclst.h"
#include "sl_file.h"

int
motor_setup_list_scan_parameters(
	char *scan_name,
	MX_SCAN *old_scan,
	char *record_description_buffer,
	size_t record_description_buffer_length,
	char *input_devices_string,
	char *measurement_times_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_list_scan_parameters()";

	MX_RECORD *record;
	MX_RECORD **motor_record_array;
	MX_FILE_LIST_SCAN *file_list_scan;
	MX_RECORD *parent_scan_record;
	MX_SCAN *child_scan;
	MX_RECORD **pointer_to_record_pointer_in_parent;
	mx_status_type mx_status;
	char prompt[100];
	char buffer[100];
	char scan_position_filename[MXU_FILENAME_LENGTH + 1];
	int i, status;
	int string_length;
	size_t buffer_left;
	long scan_class, scan_type;
	long scan_num_independent_variables;
	long scan_num_motors;
	long old_scan_num_motors;
	int *motor_is_independent_variable;

	char **motor_name_array;
	long motor_name_dimension_array[2];

	static size_t name_element_size_array[2] = {
		sizeof(char), sizeof(char *) };

	long default_long;
	char *default_string_ptr;

	scan_class = MXS_LIST_SCAN;

	if ( old_scan == (MX_SCAN *) NULL ) {
		/* Only file list scans are currently supported. */

		scan_type = MXS_LST_FILE;
	} else {
		/* 'modify scan' is not allowed to change the scan type.
		 * Use 'setup scan' if you want a different scan type.
		 */

		scan_type = old_scan->record->type;
	}

	switch( scan_type ) {
	case MXS_LST_FILE:
		if ( old_scan == NULL ) {
			default_long = 1;
		} else {
			default_long = old_scan->num_motors;
		}
		status = motor_get_long( output, "Enter number of motors -> ",
			TRUE, default_long,
			&scan_num_motors, 1, 1000000 );

		if ( status != SUCCESS ) {
			return status;
		}
		scan_num_independent_variables = scan_num_motors;
		break;
	default:
		fprintf( output, "%s: Unrecognized scan type %ld.\n",
			fname, scan_type );

		return FAILURE;
		break;
	}

	/* Add the record type info to the record description. */

	sprintf( record_description_buffer, "%s scan list_scan ", scan_name );

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length = string_length;

	switch( scan_type ) {
	case MXS_LST_FILE:
		strncat( record_description_buffer, "file_list_scan \"\" ",
			buffer_left );
		break;
	default:
		fprintf( output, "Unknown linear scan type = %ld\n",
			scan_type );
		break;
	}

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;
	sprintf( buffer, "%ld %ld ",
			scan_num_independent_variables, scan_num_motors );
	strncat( record_description_buffer, buffer, buffer_left );

	if ( scan_num_independent_variables <= 0 ) {
		fprintf( output,
	"%s: This scan has %ld independent variables.  That can't be right.\n",
		fname, scan_num_independent_variables );

		return FAILURE;
	}
	if ( scan_num_motors <= 0 ) {
		fprintf( output,
		"%s: This scan has %ld motors.  That can't be right.\n",
		fname, scan_num_motors );

		return FAILURE;
	}

	motor_is_independent_variable
			= (int *) malloc( scan_num_motors * sizeof(int));

	if ( motor_is_independent_variable == NULL ) {
		fprintf( output,
"%s: Out of memory allocating 'motor_is_independent_variable' array.\n",
			fname );
		return FAILURE;
	}

	switch( scan_type ) {
	default:
		for ( i = 0; i < scan_num_motors; i++ ) {
			motor_is_independent_variable[i] = TRUE;
		}
		break;
	}

	if ( old_scan == (MX_SCAN *) NULL ) {
		motor_record_array = NULL;
		old_scan_num_motors = 0;

		switch( scan_type ) {
		case MXS_LST_FILE:
			file_list_scan = NULL;
			break;
		default:
			fprintf( output, "%s: Unrecognized scan type %ld.\n",
				fname, scan_type );
			return FAILURE;
			break;
		}
	} else {
		motor_record_array = old_scan->motor_record_array;
		old_scan_num_motors = old_scan->num_motors;

		switch( scan_type ) {
		case MXS_LST_FILE:
			file_list_scan = (MX_FILE_LIST_SCAN *)
				(old_scan->record->record_type_struct);

			if ( file_list_scan == (MX_FILE_LIST_SCAN *) NULL ) {
				fprintf( output,
		"%s: MX_FILE_LIST_SCAN pointer in MX_SCAN is NULL.\n", fname );
				return FAILURE;
			}
			break;
		default:
			fprintf( output, "%s: Unrecognized scan type %ld.\n",
				fname, scan_type );
			return FAILURE;
			break;
		}
	}

	motor_name_dimension_array[0] = scan_num_motors;
	motor_name_dimension_array[1] = MXU_RECORD_NAME_LENGTH + 1;

	motor_name_array = (char **) mx_allocate_array(
		2, motor_name_dimension_array,
		name_element_size_array );

	if ( motor_name_array == (char **) NULL ) {
		fprintf( output,
	"%s: Unable to allocate memory for list of motor names.\n", fname);
			return FAILURE;
	}

	/* Prompt for the names of the motors and add them to
	 * the record description.
	 */

	status = motor_setup_scan_motor_names(
			record_description_buffer,
			record_description_buffer_length,
			old_scan_num_motors,
			scan_num_motors,
			motor_record_array,
			motor_name_array );

	(void) mx_free_array( motor_name_array,
		2, motor_name_dimension_array,
		name_element_size_array );

	if ( status != SUCCESS ) {
		return status;
	}

	/* We need to identify where the list of motor positions is to 
	 * be obtained from.  This will vary from scan type to scan type.
	 *
	 * This information is added to the record description after
	 * the common scan parameters.
	 */

	switch( scan_type ) {
	case MXS_LST_FILE:
		sprintf( prompt,
			"Enter name of file containing motor positions -> " );

		if ( file_list_scan ) {
			default_string_ptr = NULL;
		} else {
			default_string_ptr = file_list_scan->position_filename;
		}
		status = motor_get_string( output, prompt,
				default_string_ptr,
				&string_length, scan_position_filename );

		if ( status != SUCCESS ) {
			return status;
		}
		break;
	default:
		fprintf( output, "%s: Unrecognized scan type %ld.\n",
			fname, scan_type );

		return FAILURE;
		break;
	}

	/* Prompt for the common scan parameters such as input devices,
	 * data files and plot types.
	 */

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( input_devices_string != NULL ) {
		strncat( record_description_buffer,
			input_devices_string, buffer_left );
	} else {
		status = motor_setup_input_devices( old_scan,
						buffer, sizeof(buffer) );
		if ( status != SUCCESS )
			return status;
		strncat( record_description_buffer, buffer, buffer_left );
	}

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( measurement_times_string != NULL ) {
		strncat( record_description_buffer,
			measurement_times_string, buffer_left );
	} else {
		status = motor_setup_measurement_times( old_scan,
						buffer, sizeof(buffer) );
		if ( status != SUCCESS )
			return status;
		strncat( record_description_buffer, buffer, buffer_left );
	}

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( timer_name_string != NULL ) {
		strncat( record_description_buffer,
			timer_name_string, buffer_left );
	} else {
		status = motor_setup_timer_name( old_scan,
						buffer, sizeof(buffer) );
		if ( status != SUCCESS )
			return status;
		strncat( record_description_buffer, buffer, buffer_left );
	}

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( datafile_and_plot_parameters_string != NULL ) {
		strncat( record_description_buffer,
			datafile_and_plot_parameters_string, buffer_left );
	} else {
		status = motor_setup_datafile_and_plot_parameters( old_scan,
						scan_class, scan_type,
						buffer, sizeof(buffer) );
		if ( status != SUCCESS )
			return status;
		strncat( record_description_buffer, buffer, buffer_left );
	}

	/* Now add the source of the position list to the record
	 * description.
	 */

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	switch( scan_type ) {
	case MXS_LST_FILE:
		strncat( record_description_buffer,
			scan_position_filename, buffer_left );

		break;
	default:
		fprintf( output, "%s: Unrecognized scan type %ld.\n",
			fname, scan_type );

		return FAILURE;
		break;
	}

#if 0
	fprintf(output,
		"\n*** description = '%s'\n\n", record_description_buffer);
#endif

	parent_scan_record = NULL;
	pointer_to_record_pointer_in_parent = NULL;

	if ( old_scan != (MX_SCAN *) NULL ) {
		/* Was the old scan a child scan?  If it was, we need
		 * to preserve the link between the parent and the child.
		 */

		parent_scan_record = old_scan->parent_scan_record;

		if ( parent_scan_record != NULL ) {
			mx_status =
			    mx_find_child_scan_record_pointer_in_parent(
			      old_scan, &pointer_to_record_pointer_in_parent );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			/* If old_scan->parent_record != NULL when the
			 * delete operation is invoked on old_scan->record,
			 * the delete operation will modify the 
			 * scan_segment_record_array in the parent to 
			 * reflect the deletion.
			 *
			 * Since we don't want the scan_segment_record_array
			 * to be modified, we stop this from happening by
			 * setting old_scan->parent_scan_record to NULL.
			 */

			old_scan->parent_scan_record = NULL;
		}

		/* Delete the old scan. */

		mx_status = mx_delete_record( old_scan->record );

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;
	}

	/* Add the scan to the record list. */

	mx_status = mx_create_record_from_description( motor_record_list,
		record_description_buffer, &record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	if ( parent_scan_record != NULL ) {
		*pointer_to_record_pointer_in_parent = record;

		child_scan = (MX_SCAN *)(record->record_superclass_struct);

		child_scan->parent_scan_record = parent_scan_record;
	}

	mx_status = mx_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

