/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mscoxafs.c  --  XAFS _segmented_ scan setup and modify function. */

/* If old_scan == NULL, we are doing a new scan setup.  Otherwise, we are
 * modifying a previous scan.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "motor.h"
#include "mdialog.h"
#include "mx_alloc.h"
#include "mx_var.h"
#include "mx_scseg.h"
#include "mx_sclin.h"
#include "d_xafswn.h"
#include "seg_xafs.h"

/* Forward references to XAFS only setup routines. */

static int motor_setup_xafs_segments( char *parent_scan_name,
				MX_SCAN *parent_old_scan,
				long num_energy_regions,
				long num_k_regions,
				MX_RECORD **scan_segment_record_array,
				char **segment_name_array,
				char *input_devices_string,
				char *timer_name_string,
				char *datafile_and_plot_parameters_string );
static int motor_create_child_xafs_scan( char *child_scan_name,
				MX_SCAN *old_child_scan,
				char *child_scan_record_description );


int
motor_setup_old_xafs_scan_parameters(
	char *scan_name,
	MX_SCAN *old_scan,
	char *record_description_buffer,
	size_t record_description_buffer_length,
	char *input_devices_string,
	char *measurement_times_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_old_xafs_scan_parameters()";

	MX_RECORD *record;
	MX_RECORD **scan_segment_record_array;
	MX_SEGMENTED_SCAN *old_segmented_scan;
	MX_RECORD *parent_scan_record;
	MX_SCAN *child_scan;
	MX_RECORD **pointer_to_record_pointer_in_parent;
	mx_status_type mx_status;
	char *ptr;
	int status;
	int i, string_length;
	size_t buffer_left;
	long scan_num_segments, old_num_segments;
	long num_energy_regions, num_k_regions;
	MX_RECORD *old_child_record;
	MX_SCAN *old_child_scan;
	MX_RECORD **old_motor_record_array;

	int invoke_setup_scan_segments;
	char yes_no_string[20];
	int yes_no_string_length;

	char *child_input_devices_ptr;
	char *child_timer_name_ptr;

	char child_input_devices_buffer[100];
	char child_timer_name_buffer[100];
	char child_datafile_and_plot_buffer[100];
	char parent_datafile_and_plot_buffer[100];

	char **segment_name_array;
	long segment_name_dimension_array[2];

	static size_t name_element_size_array[2] = {
		sizeof(char), sizeof(char *) };

	if ( old_scan == (MX_SCAN *) NULL ) {
		old_segmented_scan = NULL;
	} else {
		old_segmented_scan = (MX_SEGMENTED_SCAN *)
				(old_scan->record->record_class_struct);

		if ( old_segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
			fprintf( output,
			"MX_SEGMENTED_SCAN pointer for old scan '%s' is NULL.",
				old_scan->record->name );
			return FAILURE;
		}
	}

	/* Add the record_type info to the record description. */

	sprintf( record_description_buffer,
			"%s scan segmented_scan ", scan_name );

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	strncat( record_description_buffer,
			"old_xafs_scan \"\" 0 0 0 0 0 ", buffer_left );

	/* Prompt for the common scan parameters such as input devices,
	 * data files and plot types.
	 */

	/*=== Input devices (used only in child) ===*/

	if ( input_devices_string != NULL ) {
		child_input_devices_ptr = input_devices_string;
	} else {
		/* The old parent scan does not have a record of the
		 * list of input devices, so we get it from the 
		 * first old child scan.
		 */

		if ( old_segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
			old_child_scan = NULL;
		} else {
			old_child_record
			  = old_segmented_scan->scan_segment_record_array[0];
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);
		}

		status = motor_setup_input_devices( old_child_scan,
				child_input_devices_buffer,
				sizeof( child_input_devices_buffer ) - 1 );

		if ( status != SUCCESS )
			return status;

		child_input_devices_ptr = child_input_devices_buffer;
	}

	/*=== Timer name (used both in parent and child) ===*/

	if ( timer_name_string != NULL ) {
		child_timer_name_ptr = timer_name_string;
	} else {
		status = motor_setup_timer_name( old_scan,
				child_timer_name_buffer,
				sizeof( child_timer_name_buffer ) - 1 );

		if ( status != SUCCESS )
			return status;

		child_timer_name_ptr = child_timer_name_buffer;
	}
	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	strncat( record_description_buffer,
			child_timer_name_ptr, buffer_left );

	/*=== Datafile and plot parameters (used only in parent) ===*/

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( datafile_and_plot_parameters_string != NULL ) {
		strncat( record_description_buffer,
			datafile_and_plot_parameters_string, buffer_left );
	} else {
		status = motor_setup_datafile_and_plot_parameters( old_scan,
				MXS_SEGMENTED_SCAN, MXS_SEG_XAFS,
				parent_datafile_and_plot_buffer,
				sizeof(parent_datafile_and_plot_buffer) - 1 );

		if ( status != SUCCESS )
			return status;

		strncat( record_description_buffer,
			parent_datafile_and_plot_buffer, buffer_left );
	}

	sprintf( child_datafile_and_plot_buffer, "child %s child \"%s\"",
			scan_name, scan_name );

	/* 
	 * Note from WML (07/22/97)
	 *
	 * For now it is not possible to modify the number of energy regions
	 * or k regions in a scan.  I am deferring implementing this until
	 * later since it requires that a lot of pointer manipulations be
	 * done correctly in such a way that there is no chance of the
	 * program crashing.
	 */

	if ( old_scan == (MX_SCAN *) NULL ) {

		num_energy_regions = 0;
		num_k_regions = 0;

		status = motor_get_long( output,
			"Enter number of energy regions -> ",
			TRUE, 1,
			&num_energy_regions, 1, 1000000 );
		if ( status != SUCCESS ) {
			return status;
		}
		status = motor_get_long( output,
			"Enter number of k regions -> ",
			TRUE, 1,
			&num_k_regions, 1, 1000000 );
		if ( status != SUCCESS ) {
			return status;
		}
		scan_num_segments = num_energy_regions + num_k_regions;

	} else {    /* old_scan != NULL */

		/* Determine how many energy and k regions there were
		 * in the old scan.
		 */

		num_energy_regions = 0;
		num_k_regions = 0;
		old_num_segments = old_segmented_scan->num_segments;

		/* Determine the number of energy regions first. */

		for ( i = 0; i < old_num_segments; i++ ) {
			old_child_record
			  = old_segmented_scan->scan_segment_record_array[i];
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);
			old_motor_record_array
				= old_child_scan->motor_record_array;

			if ( old_motor_record_array == (MX_RECORD **) NULL ) {
				fprintf( output,
	"Child scan record '%s' had no motors defined.\n",
					old_child_record->name );
				fprintf( output,
	"This means that '%s' is not a correctly set up XAFS scan.\n",
					old_scan->record->name );
				fprintf( output,
	"In this case, you can only use 'setup scan' and not 'modify scan'.\n");
				return FAILURE;
			}
			if ( strcmp("e_minus_e0",
				old_motor_record_array[0]->name) == 0 ) {

				num_energy_regions++;
			} else {
				/* We must have gone past the last energy
				 * region, so exit the for() loop.
				 */
				break;
			}
		}

		/* Now, determine the number of k regions. */

		for ( ; i < old_num_segments; i++ ) {
			old_child_record
			  = old_segmented_scan->scan_segment_record_array[i];
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);
			old_motor_record_array
				= old_child_scan->motor_record_array;

			if ( old_motor_record_array == (MX_RECORD **) NULL ) {
				fprintf( output,
	"Child scan record '%s' had no motors defined.\n",
					old_child_record->name );
				fprintf( output,
	"This means that '%s' is not a correctly set up XAFS scan.\n",
					old_scan->record->name );
				fprintf( output,
	"In this case, you can only use 'setup scan' and not 'modify scan'.\n");
				return FAILURE;
			}
			if ( strcmp("k", old_motor_record_array[0]->name)
								== 0 ) {
				num_k_regions++;
			} else {
				fprintf( output,
	"One of the motors in the child scan '%s' is neither "
	"'e_minus_e0' nor 'k'.\n",
					old_child_record->name );
				fprintf( output,
	"Instead, the motor's name is '%s'\n",old_motor_record_array[0]->name);
				fprintf( output,
	"This means that '%s' is not a correctly set up XAFS scan.\n",
					old_scan->record->name );
				fprintf( output,
	"In this case, you can only use 'setup scan' and not 'modify scan'.\n");
				return FAILURE;
			}
		}
		scan_num_segments = num_energy_regions + num_k_regions;
	}

	/**** Setup the individual scan segments and add their names to
	 **** the record description.
	 ****/

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;
	ptr = record_description_buffer + string_length;

	sprintf( ptr, "%ld ", scan_num_segments );

#if 0
	fprintf( output, "\n*** parent description so far = '%s'\n",
			record_description_buffer );
#endif

	if ( old_scan == (MX_SCAN *) NULL ) {
		scan_segment_record_array = NULL;
	} else {
		scan_segment_record_array
			= old_segmented_scan->scan_segment_record_array;
	}

	segment_name_dimension_array[0] = scan_num_segments;
	segment_name_dimension_array[1] = MXU_RECORD_NAME_LENGTH + 1;

	segment_name_array = (char **) mx_allocate_array(
		2, segment_name_dimension_array,
		name_element_size_array );

	if ( segment_name_array == (char **) NULL ) {
		fprintf( output,
	"%s: Unable to allocate memory for list of scan segment names.\n",
			fname);
	}

	invoke_setup_scan_segments = TRUE;

	if ( old_scan != (MX_SCAN *) NULL ) {
		yes_no_string_length = sizeof( yes_no_string ) - 1;

		status = motor_get_string( output,
			"Modify individual scan regions? (yes or no) -> ",
			NULL, &yes_no_string_length, yes_no_string );

		if ( status != SUCCESS )
			return status;

		for ( i = 0; i < yes_no_string_length; i++ ) {
			if ( isupper( yes_no_string[i] ) ) {
				yes_no_string[i] = tolower( yes_no_string[i] );
			}
		}

		if (strncmp(yes_no_string, "yes", yes_no_string_length) == 0){
			invoke_setup_scan_segments = TRUE;
		} else {
			invoke_setup_scan_segments = FALSE;
		}
	}

	if ( invoke_setup_scan_segments == TRUE ) {
		status = motor_setup_xafs_segments(
			scan_name,
			old_scan,
			num_energy_regions,
			num_k_regions,
			scan_segment_record_array,
			segment_name_array,
			child_input_devices_ptr,
			child_timer_name_ptr,
			child_datafile_and_plot_buffer );

		if ( status != SUCCESS ) {
			(void) mx_free_array( segment_name_array,
				2, segment_name_dimension_array,
				name_element_size_array );

			fprintf( output,
				"Setup of individual scan segments failed.\n");
			return status;
		}
	} else {
		for ( i = 0; i < scan_num_segments; i++ ) {
			strcpy( segment_name_array[i], "" );

			strncat( segment_name_array[i],
				scan_segment_record_array[i]->name,
				MXU_RECORD_NAME_LENGTH );
		}
	}

	/* Add the names of the scan segments to the end of the record
	 * description.
	 */

	for ( i = 0; i < scan_num_segments; i++ ) {
		string_length = strlen(record_description_buffer);
		buffer_left = record_description_buffer_length - string_length;

		strncat( record_description_buffer,
				segment_name_array[i], buffer_left );

		string_length = strlen(record_description_buffer);
		buffer_left = record_description_buffer_length - string_length;

		strncat( record_description_buffer, " ", buffer_left );
	}

	(void) mx_free_array( segment_name_array,
		2, segment_name_dimension_array,
		name_element_size_array );

#if 0
	fprintf(output,
		"\n*** description = '%s'\n\n", record_description_buffer);
#endif

	parent_scan_record = NULL;
	pointer_to_record_pointer_in_parent = NULL;

	if ( old_scan != (MX_SCAN *) NULL ) {
		/* Was the old scan a child scan?  If it was, we need
		 * to preserve the link between the parent and the child.
		 */

		parent_scan_record = old_scan->parent_scan_record;

		if ( parent_scan_record != NULL ) {
			mx_status =
			    mx_find_child_scan_record_pointer_in_parent(
			      old_scan, &pointer_to_record_pointer_in_parent );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			/* If old_scan->parent_record != NULL when the
			 * delete operation is invoked on old_scan->record,
			 * the delete operation will modify the 
			 * scan_segment_record_array in the parent to 
			 * reflect the deletion.
			 *
			 * Since we don't want the scan_segment_record_array
			 * to be modified, we stop this from happening by
			 * setting old_scan->parent_scan_record to NULL.
			 */

			old_scan->parent_scan_record = NULL;
		}

		/* Set the number of segments in this scan to zero.  This
		 * will have the side effect of preventing any of the 
		 * child scan records from being deleted when the old
		 * parent scan is deleted.
		 */

		old_segmented_scan->num_segments = 0;

		/* Delete the old scan. */

		mx_status = mx_delete_record( old_scan->record );

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;
	}

	/* Add the scan to the record list */

	mx_status = mx_create_record_from_description( motor_record_list,
		record_description_buffer, &record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	if ( parent_scan_record != NULL ) {
		*pointer_to_record_pointer_in_parent = record;

		child_scan = (MX_SCAN *)(record->record_superclass_struct);

		child_scan->parent_scan_record = parent_scan_record;
	}

	mx_status = mx_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

#define FREE_DOUBLE_ARRAYS \
	do { \
		if ( scan_region_boundary != NULL ) { \
			free( scan_region_boundary ); \
			scan_region_boundary = NULL; \
		} \
		if ( scan_region_step_size != NULL ) { \
			free( scan_region_step_size ); \
			scan_region_step_size = NULL; \
		} \
		if ( scan_region_measurement_time != NULL ) { \
			free( scan_region_measurement_time ); \
			scan_region_measurement_time = NULL; \
		} \
	} while(0)

static int
motor_setup_xafs_segments( char *parent_scan_name,
	MX_SCAN *parent_old_scan,
	long num_energy_regions,
	long num_k_regions,
	MX_RECORD **scan_segment_record_array,
	char **segment_name_array,
	char *input_devices_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_xafs_segments()";

	MX_SCAN *old_child_scan;
	MX_LINEAR_SCAN *old_child_linear_scan;
	MX_RECORD *old_child_record;
	MX_RECORD *another_old_record;
	MX_SEGMENTED_SCAN *old_segmented_scan;
	char child_scan_name[MXU_RECORD_NAME_LENGTH + 1];
	char child_motor_name[MXU_RECORD_NAME_LENGTH + 1];
	char child_scan_record_description[MXU_RECORD_DESCRIPTION_LENGTH+1];
	char measurement_times_string[MXU_BUFFER_LENGTH + 1];
	char motor_scan_parameters_string[MXU_BUFFER_LENGTH + 1];
	long i, old_scan_num_segments, scan_num_segments;
	int status;
	char prompt[80];
	double *scan_region_boundary;
	double *scan_region_step_size;
	double *scan_region_measurement_time;
	double scan_region_settling_time;
	double region_start, region_end, region_step_size;
	long region_num_measurements;

	scan_region_boundary = NULL;
	scan_region_step_size = NULL;
	scan_region_measurement_time = NULL;

	scan_num_segments = num_energy_regions + num_k_regions;

#if 0
	fprintf( output, "********\n" );
	fprintf( output, "*** parent_scan_name = '%s'\n", parent_scan_name );
	fprintf( output, "*** scan_num_segments = %ld\n", scan_num_segments );
	fprintf( output, "*** num_energy_regions = %ld\n", num_energy_regions);
	fprintf( output, "*** num_k_regions = %ld\n", num_k_regions);
	fprintf( output, "*** input_devices_string = '%s'\n",
					input_devices_string );
	fprintf( output, "*** timer_name_string = '%s'\n",
					timer_name_string );
	fprintf( output, "*** datafile_and_plot_parameters_string = '%s'\n",
					datafile_and_plot_parameters_string );
#endif

	status = SUCCESS;

	/* Setup arrays to contain scan region parameters. */

	scan_region_boundary = (double *)
		malloc( (scan_num_segments + 1) * sizeof(double) );

	if ( scan_region_boundary == NULL ) {
		fprintf( output,
		"%s: can't allocate memory for scan_region_boundary array.\n",
			fname );
		return FAILURE;
	}

	scan_region_step_size = (double *)
		malloc( scan_num_segments * sizeof(double) );

	if ( scan_region_step_size == NULL ) {
		fprintf( output,
		"%s: can't allocate memory for scan_region_step_size array.\n",
			fname );
		FREE_DOUBLE_ARRAYS;
		return FAILURE;
	}

	scan_region_measurement_time = (double *)
		malloc( scan_num_segments * sizeof(double) );

	if ( scan_region_measurement_time == NULL ) {
		fprintf( output,
	"%s: can't allocate memory for scan_region_measurement_time array.\n",
			fname );
		FREE_DOUBLE_ARRAYS;
		return FAILURE;
	}

	if ( parent_old_scan == NULL ) {
		old_scan_num_segments = 0;

		old_segmented_scan = NULL;
	} else {
		old_segmented_scan = (MX_SEGMENTED_SCAN *)
			(parent_old_scan->record->record_class_struct);

		if ( old_segmented_scan == NULL ) {
			fprintf( output,
			"%s: Program bug - old_segmented_scan == NULL "
			"when parent_old_scan != NULL\n",
				fname );
			FREE_DOUBLE_ARRAYS;
			return FAILURE;
		}

		old_scan_num_segments
			= old_segmented_scan->num_segments;
	}

	/* Setup the individual scan segments. */

	for ( i = 0; i < scan_num_segments; i++ ) {

		/* Construct or find out the child scan name. */

		if ( i >= old_scan_num_segments ) {

			old_child_record = NULL;

		} else if ( old_segmented_scan->scan_segment_record_array
				== NULL ) {

			old_child_record = NULL;
		} else {

			old_child_record =
			  (old_segmented_scan->scan_segment_record_array)[i];
		}

		if ( old_child_record == NULL ) {
			sprintf( child_scan_name, "%s.%03ld",
					parent_scan_name, i );
		} else {
			strcpy( child_scan_name, old_child_record->name );
		}

#if 0
		fprintf( output, "\n*** Child scan (%ld) name = '%s'\n\n",
				i, child_scan_name );
#endif

		/* Add the scan segment's name to the list of segments in
		 * the parent's data structures.
		 */

		strcpy( segment_name_array[i], "" );
		strncat( segment_name_array[i],
				child_scan_name, MXU_RECORD_NAME_LENGTH );

		/* Get a pointer to the child scan's MX_SCAN structure. */

		if ( old_child_record != NULL ) {
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);
			old_child_linear_scan = (MX_LINEAR_SCAN *)
				(old_child_record->record_class_struct);
		} else {
			/* If old_child_record == NULL, then does a record
			 * with this name already exist?  This is unlikely,
			 * but possible.  In that case, we must not
			 * accidentally delete an unrelated record.
			 */

			another_old_record = mx_get_record(
				motor_record_list, child_scan_name );

			if ( another_old_record != NULL ) {
				fprintf( output,
"%s: Child scan name '%s' would collide with an already existing record.\n",
					fname, child_scan_name );
				fprintf( output,
"%s: Delete that other record and try again.\n", fname );
				FREE_DOUBLE_ARRAYS;
				return FAILURE;
			}

			old_child_scan = NULL;
			old_child_linear_scan = NULL;
		}

		/* Get the position information for all the regions. */

		if ( old_child_linear_scan != NULL ) {
			scan_region_boundary[i]
				= old_child_linear_scan->start_position[0];
			scan_region_step_size[i]
				= old_child_linear_scan->step_size[0];
			scan_region_measurement_time[i]
				= old_child_scan->integration_time;
		} else {
			if ( i == 0 ) {
				scan_region_boundary[0] = 0;
				scan_region_step_size[0] = 0;
				scan_region_measurement_time[0] = 0;
			} else {
				scan_region_boundary[i]
					= scan_region_boundary[i-1];
				scan_region_step_size[i]
					= scan_region_step_size[i-1];
				scan_region_measurement_time[i]
					= scan_region_measurement_time[i-1];
			}
		}

		fprintf( output, "\n" );

		if ( i < num_energy_regions ) {
			sprintf(prompt,
				"Enter region %ld start (E - E0) -> ", i );
			status = motor_get_double( output, prompt,
				TRUE, scan_region_boundary[i],
				&scan_region_boundary[i],
				-1.0e8, 1.0e8 );
		} else {
			sprintf(prompt,
				"Enter region %ld start (k) -> ", i );
			status = motor_get_double( output, prompt,
				TRUE, scan_region_boundary[i],
				&scan_region_boundary[i],
				0.0, 1.0e8 );
		}
		if ( status != SUCCESS ) {
			FREE_DOUBLE_ARRAYS;
			return status;
		}

		if ( i < num_energy_regions ) {
			sprintf(prompt,
				"Enter region %ld step size (energy) -> ", i );
		} else {
			sprintf(prompt,
				"Enter region %ld step size (k) -> ", i );
		}
		status = motor_get_double( output, prompt,
			TRUE, scan_region_step_size[i],
			&scan_region_step_size[i],
			0.0, 1.0e8 );
		if ( status != SUCCESS ) {
			FREE_DOUBLE_ARRAYS;
			return status;
		}

		sprintf(prompt, "Enter region %ld measurement time -> ", i );
		status = motor_get_double( output, prompt,
			TRUE, scan_region_measurement_time[i],
			&scan_region_measurement_time[i],
			0.0, 1.0e8 );
		if ( status != SUCCESS ) {
			FREE_DOUBLE_ARRAYS;
			return status;
		}
			
		/* Get the position information for the end position. */

		if ( i == (scan_num_segments - 1) ) {
			if ( old_child_linear_scan != NULL ) {
				scan_region_boundary[scan_num_segments]
				    = old_child_linear_scan->start_position[0]
				    + old_child_linear_scan->step_size[0]
	  *(double)(old_child_linear_scan->num_measurements[0] - 1);

			} else {
				scan_region_boundary[scan_num_segments]
				= scan_region_boundary[scan_num_segments - 1];
			}
			fprintf( output, "\n" );

			status = motor_get_double( output,
				"Enter last region end position (k) -> ",
				TRUE, scan_region_boundary[scan_num_segments],
				&scan_region_boundary[scan_num_segments],
				0.0, 1.0e8 );
			if ( status != SUCCESS ) {
				FREE_DOUBLE_ARRAYS;
				return status;
			}

			if ( old_child_scan != NULL ) {
				scan_region_settling_time
					= old_child_scan->settling_time;
			} else {
				scan_region_settling_time = 0.0;
			}
			status = motor_get_double( output,
				"Enter motor settling time -> ",
				TRUE, scan_region_settling_time,
				&scan_region_settling_time,
				0.0, 1.0e8 );
			if ( status != SUCCESS ) {
				FREE_DOUBLE_ARRAYS;
				return status;
			}
		}

	}

#if 0
	fprintf( output, "*** scan_region_boundary[] = " );

	for ( i = 0; i <= scan_num_segments; i++ ) {
		fprintf( output, "%g ", scan_region_boundary[i] );
	}

	fprintf( output, "\n*** scan_region_step_size[] = " );

	for ( i = 0; i < scan_num_segments; i++ ) {
		fprintf( output, "%g ", scan_region_step_size[i] );
	}

	fprintf( output, "\n*** scan_region_measurement_time[] = " );

	for ( i = 0; i < scan_num_segments; i++ ) {
		fprintf( output, "%g ", scan_region_measurement_time[i] );
	}

	fprintf( output, "\n" );
#endif

	/* We now have all the information we need to setup the child scans. */

	for ( i = 0; i < scan_num_segments; i++ ) {

		sprintf( child_scan_name, "%s.%03ld", parent_scan_name, i );

		region_start = scan_region_boundary[i];
		region_step_size = scan_region_step_size[i];

		if ( i < num_energy_regions ) {
			strcpy( child_motor_name, "e_minus_e0" );

			if ( i < num_energy_regions - 1 ) {
				region_end = scan_region_boundary[i+1];
			} else {
				/* Next region start is specified in
				 * units of k, so we must convert it
				 * to energy.
				 */
				region_end = MX_HBAR_SQUARED_OVER_2M_ELECTRON
					* scan_region_boundary[i+1]
					* scan_region_boundary[i+1];
			}
		} else {
			strcpy( child_motor_name, "k" );

			region_end = scan_region_boundary[i+1];
		}

		if ( fabs(region_step_size) < 1.0e-32 ) {
			fprintf( output, "Region step size %g is too small.\n",
				region_step_size );

			FREE_DOUBLE_ARRAYS;
			return FAILURE;
		} else {
			region_num_measurements = 1L + (long) 
			    ((region_end - region_start) / region_step_size);
		}

		sprintf( motor_scan_parameters_string, "%g %g %ld",
			region_start, region_step_size,
			region_num_measurements );

		sprintf( measurement_times_string, "%g %g",
			scan_region_settling_time,
			scan_region_measurement_time[i] );

		sprintf( child_scan_record_description,
		"%s scan linear_scan motor_scan \"\" 1 1 %s %s %s %s %s %s",
				child_scan_name,
				child_motor_name,
				input_devices_string,
				measurement_times_string,
				timer_name_string,
				datafile_and_plot_parameters_string,
				motor_scan_parameters_string );

		/* Create the child XAFS scan. */

		if ( old_segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
			old_child_scan = NULL;
		} else {
			old_child_record
			  = old_segmented_scan->scan_segment_record_array[i];
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);
		}

		status = motor_create_child_xafs_scan( child_scan_name,
				old_child_scan,
				child_scan_record_description );

		if ( status == FAILURE ) {
			fprintf( output, "Child scan setup aborted.\n");

			break;	/* exit the for() loop. */
		}
	}

	FREE_DOUBLE_ARRAYS;

	return status;
}


static int
motor_create_child_xafs_scan(
	char *scan_name,
	MX_SCAN *old_scan,
	char *record_description_buffer )
{
	MX_RECORD *record;
	MX_RECORD *parent_scan_record;
	MX_RECORD **pointer_to_record_pointer_in_parent;
	MX_SCAN *child_scan;
	mx_status_type mx_status;

#if 0
	fprintf(output,
		"\n*** description = '%s'\n\n",record_description_buffer);
	fprintf(output,
		"*** old_scan = '%p'\n", old_scan);
	if ( old_scan != NULL ) {
		fprintf(output,
			"*** old_scan = '%s'\n", old_scan->record->name);
	}
#endif

	parent_scan_record = NULL;
	pointer_to_record_pointer_in_parent = NULL;

	if ( old_scan != (MX_SCAN *) NULL ) {
		/* Was the old scan a child scan?  If it was, we need
		 * to preserve the link between the parent and the child.
		 */

		parent_scan_record = old_scan->parent_scan_record;

#if 0
		fprintf(output, "*** parent_scan_record = %p\n",
					parent_scan_record );
#endif

		if ( parent_scan_record != NULL ) {
			mx_status =
			    mx_find_child_scan_record_pointer_in_parent(
			      old_scan, &pointer_to_record_pointer_in_parent );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			/* If old_scan->parent_record != NULL when the
			 * delete operation is invoked on old_scan->record,
			 * the delete operation will modify the 
			 * scan_segment_record_array in the parent to 
			 * reflect the deletion.
			 *
			 * Since we don't want the scan_segment_record_array
			 * to be modified, we stop this from happening by
			 * setting old_scan->parent_scan_record to NULL.
			 */

			old_scan->parent_scan_record = NULL;
		}
#if 0
		fprintf( output, "*** about to delete record '%s'\n",
					old_scan->record->name );
#endif

		/* Delete the old scan. */

		mx_status = mx_delete_record( old_scan->record );

#if 0
		fprintf( output,
			"*** status code from mx_delete_record = %ld\n",
			mx_status.code );
#endif

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;
	}

	/* Add the scan to the record list. */

	mx_status = mx_create_record_from_description( motor_record_list,
		record_description_buffer, &record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	if ( parent_scan_record != NULL ) {
		*pointer_to_record_pointer_in_parent = record;

		child_scan = (MX_SCAN *)(record->record_superclass_struct);

		child_scan->parent_scan_record = parent_scan_record;
	}

	mx_status = mx_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

