/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mscreg.c  --  Motor generic region scan setup and modify function. */

/* If old_scan == NULL, we are doing a new scan setup.  Otherwise, we are
 * modifying a previous scan.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "motor.h"
#include "mdialog.h"
#include "mx_alloc.h"
#include "mx_scseg.h"
#include "seg_reg.h"

int
motor_setup_region_scan_parameters(
	char *scan_name,
	MX_SCAN *old_scan,
	char *record_description_buffer,
	size_t record_description_buffer_length,
	char *input_devices_string,
	char *measurement_times_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_region_scan_parameters()";

	MX_RECORD *record;
	MX_RECORD **scan_segment_record_array;
	MX_SEGMENTED_SCAN *old_segmented_scan;
	MX_RECORD *parent_scan_record;
	MX_SCAN *child_scan;
	MX_RECORD **pointer_to_record_pointer_in_parent;
	mx_status_type mx_status;
	char *ptr;
	int status;
	int i, string_length;
	size_t buffer_left;
	long scan_num_segments;

	int invoke_setup_scan_segments;
	char yes_no_string[20];
	int yes_no_string_length;

	char *child_input_devices_ptr;
	char *child_measurement_times_ptr;
	char *child_timer_name_ptr;

#if 0
	char child_input_devices_buffer[100];
	char child_measurement_times_buffer[100];
#endif
	char child_timer_name_buffer[100];
	char child_datafile_and_plot_buffer[100];
	char parent_datafile_and_plot_buffer[100];

	char **segment_name_array;
	long segment_name_dimension_array[2];

	long default_long;

	static size_t name_element_size_array[2] = {
		sizeof(char), sizeof(char *) };

	if ( old_scan == (MX_SCAN *) NULL ) {
		old_segmented_scan = NULL;
	} else {
		old_segmented_scan = (MX_SEGMENTED_SCAN *)
				(old_scan->record->record_class_struct);

		if ( old_segmented_scan == (MX_SEGMENTED_SCAN *) NULL ) {
			fprintf( output,
			"MX_SEGMENTED_SCAN pointer for old scan '%s' is NULL.",
				old_scan->record->name );
			return FAILURE;
		}
	}

	/* Add the record_type info to the record description. */

	sprintf( record_description_buffer,
			"%s scan segmented_scan ", scan_name );

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	strncat( record_description_buffer,
			"region_scan \"\" 0 0 0 0 0 ", buffer_left );

	/* Prompt for the common scan parameters such as input devices,
	 * data files and plot types.
	 */

#if 1
	/* For now, we force the input devices and measurement times
	 * to be set in the child scan, if they were not set in the
	 * current scan's parent.
	 */

	child_input_devices_ptr = input_devices_string;
	child_measurement_times_ptr = measurement_times_string;
#else

	/*=== Input devices (used only in child) ===*/

	if ( input_devices_string != NULL ) {
		child_input_devices_ptr = input_devices_string;
	} else {
		status = motor_setup_input_devices( old_scan,
				child_input_devices_buffer,
				sizeof( child_input_devices_buffer ) - 1 );

		if ( status != SUCCESS )
			return status;

		child_input_devices_ptr = child_input_devices_buffer;
	}

	/*=== Measurement times (used only in child) ===*/

	if ( measurement_times_string != NULL ) {
		child_measurement_times_ptr = measurement_times_string;
	} else {
		status = motor_setup_measurement_times( old_scan,
				child_measurement_times_buffer,
				sizeof( child_measurement_times_buffer ) - 1 );

		if ( status != SUCCESS )
			return status;

		child_measurement_times_ptr = child_measurement_times_buffer;
	}
#endif

	/*=== Timer name (used both in parent and child) ===*/

	if ( timer_name_string != NULL ) {
		child_timer_name_ptr = timer_name_string;
	} else {
		status = motor_setup_timer_name( old_scan,
				child_timer_name_buffer,
				sizeof( child_timer_name_buffer ) - 1 );

		if ( status != SUCCESS )
			return status;

		child_timer_name_ptr = child_timer_name_buffer;
	}
	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	strncat( record_description_buffer,
			child_timer_name_ptr, buffer_left );

	/*=== Datafile and plot parameters (used only in parent) ===*/

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;

	if ( datafile_and_plot_parameters_string != NULL ) {
		strncat( record_description_buffer,
			datafile_and_plot_parameters_string, buffer_left );
	} else {
		status = motor_setup_datafile_and_plot_parameters( old_scan,
				MXS_SEGMENTED_SCAN, MXS_SEG_REGION,
				parent_datafile_and_plot_buffer,
				sizeof(parent_datafile_and_plot_buffer) - 1 );

		if ( status != SUCCESS )
			return status;

		strncat( record_description_buffer,
			parent_datafile_and_plot_buffer, buffer_left );
	}

	sprintf( child_datafile_and_plot_buffer, "child %s child \"%s\"",
			scan_name, scan_name );

	/* 
	 * Note from WML (06/08/97)
	 *
	 * For now it is not possible to modify the number of segments
	 * in a scan.  I am deferring implementing this until later since
	 * it requires that a lot of pointer manipulations be done 
	 * correctly in such a way that there is no chance of the
	 * program crashing.
	 */

	if ( old_scan == (MX_SCAN *) NULL ) {
		if ( old_segmented_scan == NULL ) {
			default_long = 1;
		} else {
			default_long = old_segmented_scan->num_segments;
		}
		status = motor_get_long( output,
			"Enter number of scan segments -> ",
			TRUE, default_long,
			&scan_num_segments, 1, 1000000 );

		if ( status != SUCCESS ) {
			return status;
		}
	} else {
		scan_num_segments = old_segmented_scan->num_segments;
	}

	/**** Setup the individual scan segments and add their names to
	 **** the record description.
	 ****/

	string_length = strlen(record_description_buffer);
	buffer_left = record_description_buffer_length - string_length;
	ptr = record_description_buffer + string_length;

	sprintf( ptr, "%ld ", scan_num_segments );

#if 0
	fprintf( output, "\n*** parent description so far = '%s'\n",
			record_description_buffer );
#endif

	if ( old_scan == (MX_SCAN *) NULL ) {
		scan_segment_record_array = NULL;
	} else {
		scan_segment_record_array
			= old_segmented_scan->scan_segment_record_array;
	}

	segment_name_dimension_array[0] = scan_num_segments;
	segment_name_dimension_array[1] = MXU_RECORD_NAME_LENGTH + 1;

	segment_name_array = (char **) mx_allocate_array(
		2, segment_name_dimension_array,
		name_element_size_array );

	if ( segment_name_array == (char **) NULL ) {
		fprintf( output,
	"%s: Unable to allocate memory for list of scan segment names.\n",
			fname);
	}

	invoke_setup_scan_segments = TRUE;

	if ( old_scan != (MX_SCAN *) NULL ) {
		yes_no_string_length = sizeof( yes_no_string ) - 1;

		status = motor_get_string( output,
			"Modify individual scan regions? (yes or no) -> ",
			NULL, &yes_no_string_length, yes_no_string );

		if ( status != SUCCESS )
			return status;

		for ( i = 0; i < yes_no_string_length; i++ ) {
			if ( isupper( yes_no_string[i] ) ) {
				yes_no_string[i] = tolower( yes_no_string[i] );
			}
		}

		if (strncmp(yes_no_string, "yes", yes_no_string_length) == 0){
			invoke_setup_scan_segments = TRUE;
		} else {
			invoke_setup_scan_segments = FALSE;
		}
	}

	if ( invoke_setup_scan_segments == TRUE ) {
		status = motor_setup_scan_segments(
			scan_name,
			old_scan,
			scan_num_segments,
			scan_segment_record_array,
			segment_name_array,
			child_input_devices_ptr,
			child_measurement_times_ptr,
			child_timer_name_ptr,
			child_datafile_and_plot_buffer );

		if ( status != SUCCESS ) {
			(void) mx_free_array( segment_name_array,
				2, segment_name_dimension_array,
				name_element_size_array );

			fprintf( output,
				"Setup of individual scan segments failed.\n");
			return status;
		}
	} else {
		for ( i = 0; i < scan_num_segments; i++ ) {
			strcpy( segment_name_array[i], "" );

			strncat( segment_name_array[i],
				scan_segment_record_array[i]->name,
				MXU_RECORD_NAME_LENGTH );
		}
	}

	/* Add the names of the scan segments to the end of the record
	 * description.
	 */

	for ( i = 0; i < scan_num_segments; i++ ) {
		string_length = strlen(record_description_buffer);
		buffer_left = record_description_buffer_length - string_length;

		strncat( record_description_buffer,
				segment_name_array[i], buffer_left );

		string_length = strlen(record_description_buffer);
		buffer_left = record_description_buffer_length - string_length;

		strncat( record_description_buffer, " ", buffer_left );
	}

	(void) mx_free_array( segment_name_array,
		2, segment_name_dimension_array,
		name_element_size_array );

#if 0
	fprintf(output,
		"\n*** description = '%s'\n\n", record_description_buffer);
#endif

	parent_scan_record = NULL;
	pointer_to_record_pointer_in_parent = NULL;

	if ( old_scan != (MX_SCAN *) NULL ) {
		/* Was the old scan a child scan?  If it was, we need
		 * to preserve the link between the parent and the child.
		 */

		parent_scan_record = old_scan->parent_scan_record;

		if ( parent_scan_record != NULL ) {
			mx_status =
			    mx_find_child_scan_record_pointer_in_parent(
			      old_scan, &pointer_to_record_pointer_in_parent );

			if ( mx_status.code != MXE_SUCCESS )
				return FAILURE;

			/* If old_scan->parent_record != NULL when the
			 * delete operation is invoked on old_scan->record,
			 * the delete operation will modify the 
			 * scan_segment_record_array in the parent to 
			 * reflect the deletion.
			 *
			 * Since we don't want the scan_segment_record_array
			 * to be modified, we stop this from happening by
			 * setting old_scan->parent_scan_record to NULL.
			 */

			old_scan->parent_scan_record = NULL;
		}

		/* Set the number of segments in this scan to zero.  This
		 * will have the side effect of preventing any of the 
		 * child scan records from being deleted when the old
		 * parent scan is deleted.
		 */

		old_segmented_scan->num_segments = 0;

		/* Delete the old scan. */

		mx_status = mx_delete_record( old_scan->record );

		if ( mx_status.code != MXE_SUCCESS )
			return FAILURE;
	}

	/* Add the scan to the record list */

	mx_status = mx_create_record_from_description( motor_record_list,
		record_description_buffer, &record, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	if ( parent_scan_record != NULL ) {
		*pointer_to_record_pointer_in_parent = record;

		child_scan = (MX_SCAN *)(record->record_superclass_struct);

		child_scan->parent_scan_record = parent_scan_record;
	}

	mx_status = mx_finish_record_initialization( record );

	if ( mx_status.code != MXE_SUCCESS )
		return FAILURE;

	return SUCCESS;
}

