/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/* mscseg.c  --  Motor segmented scan setup and modify function. */

/* If old_scan == NULL, we are doing a new scan setup.  Otherwise, we are
 * modifying a previous scan.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "motor.h"
#include "mdialog.h"
#include "mx_alloc.h"
#include "mx_scseg.h"
#include "seg_reg.h"

int
motor_setup_segmented_scan_parameters(
	char *scan_name,
	MX_SCAN *old_scan,
	char *record_description_buffer,
	size_t record_description_buffer_length,
	char *input_devices_string,
	char *measurement_times_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_segmented_scan_parameters()";

	int status;
	long scan_type;

	if ( old_scan == (MX_SCAN *) NULL ) {
		fprintf(output, "Enter scan type:\n"
			"    1.  general region scan\n"
			"    2.  old XAFS scan\n"
			"\n");

		status = motor_get_long( output,
			"--> ", TRUE, 1, &scan_type, 1, 2 );

		if ( status != SUCCESS ) {
			return status;
		}
		scan_type += (MXS_SEG_REGION - 1);
	} else {
		/* 'modify scan' is not allowed to change the scan type.
		 * Use 'setup scan' if you want a different scan type.
		 */

		scan_type = old_scan->record->type;
	}

	switch( scan_type ) {
	case MXS_SEG_REGION:
		status = motor_setup_region_scan_parameters(
			scan_name,
			old_scan,
			record_description_buffer,
			record_description_buffer_length,
			input_devices_string,
			measurement_times_string,
			timer_name_string,
			datafile_and_plot_parameters_string );
		break;
	case MXS_SEG_XAFS:
		status = motor_setup_old_xafs_scan_parameters(
			scan_name,
			old_scan,
			record_description_buffer,
			record_description_buffer_length,
			input_devices_string,
			measurement_times_string,
			timer_name_string,
			datafile_and_plot_parameters_string );
		break;
	default:
		fprintf( output, "%s: Unrecognized scan type %ld.\n",
			fname, scan_type );

		status = FAILURE;
	}

	return status;
}

int
motor_setup_scan_segments( char *parent_scan_name,
	MX_SCAN *parent_old_scan,
	long scan_num_segments,
	MX_RECORD **scan_segment_record_array,
	char **segment_name_array,
	char *input_devices_string,
	char *measurement_times_string,
	char *timer_name_string,
	char *datafile_and_plot_parameters_string )
{
	const char fname[] = "motor_setup_scan_segments()";

	MX_SCAN *old_child_scan;
	MX_RECORD *old_child_record;
	MX_RECORD *another_old_record;
	MX_SEGMENTED_SCAN *old_segmented_scan;
	char child_scan_name[MXU_RECORD_NAME_LENGTH + 1];
	char child_scan_record_description[MXU_RECORD_DESCRIPTION_LENGTH+1];
	long i, old_scan_num_segments;
	long child_scan_class;
	int status;

#if 0
	fprintf( output, "********\n" );
	fprintf( output, "*** parent_scan_name = '%s'\n", parent_scan_name );
	fprintf( output, "*** scan_num_segments = %ld\n", scan_num_segments );
	fprintf( output, "*** input_devices_string = '%s'\n",
					input_devices_string );
	fprintf( output, "*** measurement_times_string = '%s'\n",
					measurement_times_string );
	fprintf( output, "*** timer_name_string = '%s'\n",
					timer_name_string );
	fprintf( output, "*** datafile_and_plot_parameters_string = '%s'\n",
					datafile_and_plot_parameters_string );
#endif

	status = SUCCESS;

	if ( parent_old_scan == NULL ) {
		old_scan_num_segments = 0;
		old_segmented_scan = NULL;
	} else {
		old_segmented_scan = (MX_SEGMENTED_SCAN *)
			(parent_old_scan->record->record_class_struct);

		if ( old_segmented_scan == NULL ) {
			fprintf( output,
			"%s: Program bug - old_segmented_scan == NULL "
			"when parent_old_scan != NULL\n",
				fname );
			return FAILURE;
		}

		old_scan_num_segments
			= old_segmented_scan->num_segments;
	}

	for ( i = 0; i < scan_num_segments; i++ ) {

		/* Construct or find out the child scan name. */

		if ( i >= old_scan_num_segments ) {

			old_child_record = NULL;

		} else if ( old_segmented_scan->scan_segment_record_array
				== NULL ) {

			old_child_record = NULL;
		} else {

			old_child_record =
			  (old_segmented_scan->scan_segment_record_array)[i];
		}

		if ( old_child_record == NULL ) {
			sprintf( child_scan_name, "%s.%03ld",
					parent_scan_name, i );
		} else {
			strcpy( child_scan_name, old_child_record->name );
		}

#if 0
		fprintf( output, "\n*** Child scan (%ld) name = '%s'\n\n",
				i, child_scan_name );
#endif

		/* Add the scan segment's name to the list of segments in
		 * the parent's data structures.
		 */

		strcpy( segment_name_array[i], "" );
		strncat( segment_name_array[i],
				child_scan_name, MXU_RECORD_NAME_LENGTH );

		/* Get a pointer to the child scan's MX_SCAN structure. */

		if ( old_child_record != NULL ) {
			old_child_scan = (MX_SCAN *)
				(old_child_record->record_superclass_struct);

			child_scan_class = old_child_record->class;
		} else {
			/* If old_child_record == NULL, then does a record
			 * with this name already exist?  This is unlikely,
			 * but possible.  In that case, we must not
			 * accidentally delete an unrelated record.
			 */

			another_old_record = mx_get_record(
				motor_record_list, child_scan_name );

			if ( another_old_record != NULL ) {
				fprintf( output,
"%s: Child scan name '%s' would collide with an already existing record.\n",
					fname, child_scan_name );
				fprintf( output,
"%s: Delete that other record and try again.\n", fname );
				return FAILURE;
			}

			old_child_scan = NULL;

			/* Get the child's scan class. */

			fprintf(output,
			"Enter scan class for child scan '%s'\n"
			"    1 - Linear scan\n"
			"    2 - List scan\n"
			"    3 - XAFS scan\n"
			"    4 - Segmented scan\n"
			"\n", child_scan_name );

			status = motor_get_long( output, "--> ", TRUE, 1,
					&child_scan_class, 1, 4);

			child_scan_class += (MXS_LINEAR_SCAN - 1);
		}

		/* Call a class-dependent function to prompt for the
		 * scan parameters.
		 */

		child_scan_record_description[0] = '\0';

		switch( child_scan_class ) {
		case MXS_LINEAR_SCAN:
			status = motor_setup_linear_scan_parameters(
				child_scan_name, old_child_scan,
				child_scan_record_description,
				MXU_RECORD_DESCRIPTION_LENGTH,
				input_devices_string,
				measurement_times_string,
				timer_name_string,
				datafile_and_plot_parameters_string );
			break;
		case MXS_SEGMENTED_SCAN:
			status = motor_setup_segmented_scan_parameters(
				child_scan_name, old_child_scan,
				child_scan_record_description,
				MXU_RECORD_DESCRIPTION_LENGTH,
				input_devices_string,
				measurement_times_string,
				timer_name_string,
				datafile_and_plot_parameters_string );
			break;
		case MXS_LIST_SCAN:
			status = motor_setup_list_scan_parameters(
				child_scan_name, old_child_scan,
				child_scan_record_description,
				MXU_RECORD_DESCRIPTION_LENGTH,
				input_devices_string,
				measurement_times_string,
				timer_name_string,
				datafile_and_plot_parameters_string );
			break;
		default:
			fprintf(output, "%s: unrecognized scan class %ld.",
				fname, child_scan_class);
			return FAILURE;
			break;
		}

		if ( status == FAILURE ) {
			fprintf( output, "Child scan setup aborted.\n");

			break;	/* exit the for() loop. */
		}
	}

	return status;
}

