/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*  next_cmd.c  ---  A simple command line processor.  */

#include <stdio.h>
#include <string.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_key.h"
#include "keycodes.h"
#include "next_cmd.h"

#if USE_INTERNAL_CMD_LINE	/* Don't compile unless this is selected. */

/* Define array of saved commands. */

static char command_history[ NUM_COMMAND_LINES ][ COMMAND_LINE_LENGTH ];

static int first_command = -1;
static int last_command = -1;
static int total_num_commands = -1;

static int backspace = 0;	/* placeholder for backspace character. */

extern FILE *output;

static void
show_regular_cursor( void )
{
	/* not yet implemented */
}

static void
show_insert_cursor( void )
{
	/* not yet implemented */
}

void
add_command_to_history_buffer( char *command_string )
{
	/* Update the buffer indices. */

	total_num_commands++;

	if ( first_command < 0 ) {
		first_command = 0;
		last_command = 0;
	} else {
		last_command++;

		if ( last_command >= NUM_COMMAND_LINES ) {
			last_command = 0;
		}

		if ( last_command == first_command ) {
			first_command++;

			if ( first_command >= NUM_COMMAND_LINES ) {
				first_command = 0;
			}
		}
	}

	/* Store the command line. */

	strncpy( command_history[ last_command ],
		command_string, COMMAND_LINE_LENGTH );

#if 0
	fprintf( output, "'%s' added to history buffer line %d.\n",
		command_string, last_command );
#endif

	return;
}

int
get_first_command_index( void )
{
	return first_command;
}

int
get_last_command_index( void )
{
	return last_command;
}

int
get_total_num_commands( void )
{
	return total_num_commands;
}

char
*get_history_buffer_line( int line_number )
{
	static char *ptr;

	if ( line_number < 0 || line_number >= NUM_COMMAND_LINES ) {
		ptr = NULL;
	} else {
		ptr = &( command_history[ line_number ][0] );
	}

	return ptr;
}

char
*read_next_command_line( void )
{
	static char command_buffer[ COMMAND_LINE_LENGTH ];

	char *command_buffer_start, *command_buffer_end;
	char *end_of_string;
	char *ptr, *ptr_from, *ptr_to;
	int exit_loop, insert_mode;
	int non_blank_character_found;
	int num_backspaces;
	int old_command, old_string_length, new_string_length;
	int c, i, buffer_index;

	/* Define the backspace character, if we haven't already. */

	if ( backspace == 0 ) {
#ifdef OS_MSDOS
		backspace = BS;
#else
		backspace = BS;
#endif
	}

	/* Start reading keystrokes and acting on them. */

	command_buffer_start = &command_buffer[0];
	command_buffer_end = &command_buffer[ COMMAND_LINE_LENGTH - 1 ];

	ptr = end_of_string = command_buffer_start;

	exit_loop = FALSE;
	insert_mode = FALSE;

	old_command = total_num_commands + 1;

	show_regular_cursor();

	while (exit_loop == FALSE) {

		c = mx_getch();	/* Get the next character. */

#ifdef OS_MSDOS
		/* Mask off high byte of ordinary characters. 
		 * Characters with a low byte of all zeros are
		 * PC scan codes like KEY_LEFT.
		 */

		if ( c & 0xFF ) {
			c &= 0xFF;
		}
#endif /* OS_MSDOS */

		switch( c ) {

		case KEY_LEFT:
			if ( ptr > command_buffer_start ) {
				ptr--;

				fputc( backspace, output );
			}
			break;

		case KEY_RIGHT:
			if ( ptr < end_of_string ) {
				fputc( *ptr, output );

				ptr++;
			}
			break;

		case KEY_UP:
		case KEY_DOWN:
			insert_mode = FALSE;
			show_regular_cursor();

			old_string_length 
			    = (int) (end_of_string - command_buffer_start);

			/* Go to the start of the old string. */

			num_backspaces 
				= (int) (ptr - command_buffer_start);

			while( num_backspaces-- > 0 ) {
				fputc( backspace, output );
				fflush(output);
			}

			/* Compute the index of the command line to be
			 * recalled while ensuring that the command line
			 * index doesn't fall outside the range of 
			 * allowed values.
			 */

			if ( c == KEY_UP ) {
				old_command--;
			} else {
				old_command++;
			}

			if ( old_command < 0 ) {
				old_command = 0;
			}

			if ( total_num_commands - old_command
						> NUM_COMMAND_LINES ) {
				old_command 
				    = total_num_commands - NUM_COMMAND_LINES;
			}

			/* Now display the command line. */

			if ( old_command > total_num_commands ) {
				old_command = total_num_commands + 1;

				for ( i = 0; i < old_string_length; i++ ) {
					fputc( ' ', output );
				}
				for ( i = 0; i < old_string_length; i++ ) {
					fputc( backspace, output );
				}

				end_of_string = ptr = command_buffer_start;
			} else {
				buffer_index = old_command % NUM_COMMAND_LINES;

				strncpy( command_buffer,
					command_history[ buffer_index ],
					COMMAND_LINE_LENGTH );

				new_string_length = strlen(command_buffer);

				fprintf( output, "%s", command_buffer );

				if (new_string_length < old_string_length) {
					for ( i = new_string_length;
					  i < old_string_length; i++ )
					{
						fputc( ' ', output );
					}
					for ( i = new_string_length;
					  i < old_string_length; i++ )
					{
						fputc( backspace, output );
					}
				}
				ptr = end_of_string 
				  = command_buffer_start + new_string_length;
			}
			break;

		case KEY_INSERT:
			if ( insert_mode == FALSE ) {
				insert_mode = TRUE;
				show_insert_cursor();
			} else {
				insert_mode = FALSE;
				show_regular_cursor();
			}
			break;

		case KEY_DELETE:
			ptr_to = ptr;
			ptr_from = ptr_to + 1;

			num_backspaces = (int) (end_of_string - ptr);

			while ( ptr_from < end_of_string ) {
				*ptr_to = *ptr_from++;

				fputc( *ptr_to++, output );
				fflush(output);
			}

			if ( num_backspaces > 0 ) {
				fputc( ' ', output );

				end_of_string = ptr_from - 1;

				while ( num_backspaces-- > 0 ) {
					fputc( backspace, output );
					fflush(output);
				}
			}

			break;

		case BS:	/* Destructive backspace. */

			if ( ptr > command_buffer_start ) {
				ptr--;
				*ptr = '\0';
				end_of_string = ptr;

				fputc( backspace, output );
				fputc( ' ', output );
				fputc( backspace, output );
			}
			break;

		case '\n':	/* End of line. */
		case CR:
			*end_of_string = '\0';
			exit_loop = TRUE;

			fputc( '\n', output );
			break;

		default:	/* Everything else. */

			if ( insert_mode == FALSE || ptr >= end_of_string ) {
				*ptr = (char) c;
				ptr++;

				if ( ptr > end_of_string ) {
					end_of_string = ptr;
				}

				/* Display the character. */

				fputc( c, output );

			} else {
				/* Move everything to the right of 'ptr'
				 * one character further to the right.
				 */

				for ( ptr_to = end_of_string + 1;
				      ptr_to > ptr; ptr_to-- )
				{
					*ptr_to = *(ptr_to - 1);
				}

				*ptr = (char) c;
				ptr++;

				end_of_string++;

				/* Display the character. */

				fputc( c, output );

				/* Display the rest of the old string. */

				fprintf( output, "%s", ptr );

				num_backspaces = strlen(ptr);

				while (num_backspaces-- > 0) {
					fputc( backspace, output );
				}
			}

			break;
		}

		fflush( output );
	}

	/* Add the command to the command history unless the 
	 * command buffer only contains blank characters.
	 */

	new_string_length = strlen(command_buffer);

	non_blank_character_found = FALSE;

	for ( i = 0; i < new_string_length; i++ ) {
		if ( command_buffer[i] != ' ' && command_buffer[i] != '\t' ) {
			non_blank_character_found = TRUE;

			break;		/* exit the for() loop. */
		}
	}

	if ( non_blank_character_found ) {

		add_command_to_history_buffer( command_buffer );

	}

	return command_buffer_start;
}

#endif /* USE_INTERNAL_CMD_LINE	*/
