/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    ms_epics.c
 *
 * Purpose: This driver connects to one or more EPICS soft variables
 *          and starts to monitor changes to their values.  When certain
 *          EPICS variables are changed, the MX server interprets this
 *          as a command to perform some operation like moving a motor.
 *          The status of the operations being performed is written
 *          back to other EPICS variables.  In effect, MX uses an
 *          EPICS IOC as a mailbox through which EPICS clients send
 *          commands to an MX server and in which MX returns the result.
 */

#include <stdio.h>

#include "mxconfig.h"

#ifndef OS_UNIX
#undef  HAVE_EPICS
#define HAVE_EPICS 0
#endif

#if HAVE_EPICS

#include <stdlib.h>

#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_alloc.h"
#include "mx_sock.h"

#include "ms_serv.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"
#include "db_access.h"
#include "epicsVersion.h"

#include "mx_epics.h"

#include "ms_epics.h"
#include "ms_epmd.h"

/* Between EPICS 3.13.0.beta4 and EPICS 3.13.0.beta11, the definition of
 * struct event_handler_args in "cadef.h" was modified to change the field
 * "void *dbr" to "READONLY void *dbr".  READONLY usually evaluates to 'const'.
 * Since we have not yet given up support of EPICS versions prior to 
 * 3.13.0.beta11, we have to be ready to deal with both cases.
 */

#if ( (EPICS_VERSION < 3) || (EPICS_VERSION > 3) )
#   error "Not yet programmed to deal with EPICS versions other than version 3."
#else

#   if ( EPICS_REVISION < 13 )

#	define MXEP_CONST

#   elif ( EPICS_REVISION > 13 )

#	define MXEP_CONST		READONLY

#   else /* EPICS_REVISION == 13 */

#	if ( EPICS_MODIFICATION > 0 )

#	    define MXEP_CONST		READONLY

#	else /* EPICS_MODIFICATION == 0 */

#	    if ( EPICS_UPDATE_LEVEL > 4 )

#		define MXEP_CONST	READONLY

#	    else

#		define MXEP_CONST

#	    endif
#	endif
#   endif
#endif

/* The following is a place to store a pointer to the EPICS socket
 * event handler structure.
 */

static MXSRV_EVENT_HANDLER *mxsrv_epics_event_handler;

/* The following are two other variables that are global to this file. */

static int               mxsrv_epics_num_channels;
static MX_EPICS_CHANNEL *mxsrv_epics_channel_list;

mx_status_type
mxsrv_epics_init( MX_RECORD *record_list,
		MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
		MXSRV_EVENT_HANDLER *event_handler )
{
        const char fname[] = "mxsrv_epics_init()";

	MX_RECORD *current_record, *list_head_record;
	MX_LIST_HEAD *list_head;
	long server_protocols_active;
	int current_channel_number;
	int epics_status;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked.", fname));

	list_head_record = record_list->list_head;

	list_head = (MX_LIST_HEAD *)list_head_record->record_superclass_struct;

	server_protocols_active = list_head->server_protocols_active;

	if ( ( server_protocols_active & MXPROT_EPICS_MAILBOX ) == 0 ) {

		/* If the server is not configured to use EPICS mailbox
		 * records, then we return without setting them up.
		 * This is necessary to avoid problems that would occur
		 * if both EPICS server side and EPICS client side records
		 * were used in the same server.
		 */

		return MX_SUCCESSFUL_RESULT;
	}

	/* How many EPICS channels will we be managing? */

	current_record = record_list->next_record;

	mxsrv_epics_num_channels = 0;

	while ( current_record != list_head_record ) {
		if ( current_record->superclass == MXR_EPICS_MAILBOX ) {
			switch( current_record->type ) {
			case MXM_EPICS_MOTOR_DB:
				mxsrv_epics_num_channels
				  += mxsrv_epics_motor_db_num_channel_defaults;

				break;
			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Record '%s' (type = %ld) is not a known type of EPICS mailbox record.",
					current_record->name,
					current_record->type );
				break;
			}
		}
		current_record = current_record->next_record;
	}

	if ( mxsrv_epics_num_channels <= 0 ) {
		mxsrv_epics_channel_list = NULL;

		return MX_SUCCESSFUL_RESULT;
	}

	mxsrv_epics_channel_list = (MX_EPICS_CHANNEL *)
		malloc( mxsrv_epics_num_channels * sizeof(MX_EPICS_CHANNEL));

	if ( mxsrv_epics_channel_list == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Couldn't allocate the mxsrv_epics_channel_list array." );
	}

	/* Now set up the connections to the EPICS channels. */

	current_channel_number = 0;

	current_record = list_head_record->next_record;

	while ( current_record != list_head_record ) {
		if ( current_record->superclass == MXR_EPICS_MAILBOX ) {
			switch( current_record->type ) {
			case MXM_EPICS_MOTOR_DB:
				status = mxsrv_epics_motor_db_add_channels(
						current_record,
						mxsrv_epics_channel_list,
						current_channel_number );

				if ( status.code != MXE_SUCCESS )
					return status;

				current_channel_number
				  += mxsrv_epics_motor_db_num_channel_defaults;
				break;
			default:
				return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Record '%s' (type = %ld) is not a known type of EPICS mailbox record.",
					current_record->name,
					current_record->type );
				break;
			}
		}
		current_record = current_record->next_record;
	}

	/* Allow a little time for any remaining EPICS I/O to finish. */

	epics_status = ca_pend_event( 0.1 );

	switch( epics_status ) {
	case ECA_NORMAL:
	case ECA_TIMEOUT:
		break;
	default:
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"ca_pend_event() failed with status = %d, error_text = '%s'",
			epics_status, ca_message(epics_status) );
	}

	MX_DEBUG( 1,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_epics_process_event( MX_RECORD *record_list,
		MXSRV_SOCKET_HANDLER *socket_handler,
		MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
		MXSRV_EVENT_HANDLER *event_handler )
{
        const char fname[] = "mxsrv_epics_process_event()";

	int epics_status;

	MX_DEBUG( 1,("%s invoked.", fname));

	/* Give Channel Access a chance to process any queued events. */

	epics_status = ca_pend_event( 0.001 );

	switch( epics_status ) {
	case ECA_NORMAL:
	case ECA_TIMEOUT:
		break;
	default:
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"ca_pend_event() failed with status = %d, error_text = '%s'",
			epics_status, ca_message(epics_status) );
	}

	MX_DEBUG( 1,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/*=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=*/

mx_status_type
mxsrv_epics_channel_access_initialize(
		MX_RECORD *record_list,
		MXSRV_SOCKET_HANDLER_LIST *socket_handler_list )
{
	const char fname[] = "mxsrv_epics_channel_access_initialize()";

	int i, epics_status;

	MX_DEBUG( 1,("%s invoked.", fname));

	/* Get a pointer to the MX server event handler structure for EPICS. */

	mxsrv_epics_event_handler = NULL;

	for ( i = 0; i < mxsrv_num_event_handlers; i++ ) {
		if ( mxsrv_event_handler_list[i].type
					== MXF_SRV_EPICS_EVENT_TYPE ) {

			mxsrv_epics_event_handler
					= &mxsrv_event_handler_list[i];

			break;		/* Exit the for() loop. */
		}
	}

	if ( mxsrv_epics_event_handler == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"No MX server event handler has been defined "
			"for EPICS in mxsrv_event_handler_list[]." );
	}

	/* Initialize channel access itself. */

	epics_status = ca_task_initialize();

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
	"ca_task_initialize() failed with status = %d, error_text = '%s'",
			epics_status, ca_message(epics_status) );
	}

	epics_status = ca_add_fd_registration( mxsrv_epics_fd_register_fn,
						socket_handler_list );

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_FUNCTION_FAILED, fname,
	"ca_add_fd_registration() failed with status = %d, error_text = '%s'",
			epics_status, ca_message(epics_status) );
	}

	MX_DEBUG( 1,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

void
mxsrv_epics_fd_register_fn( void *userarg, int fd, int opened )
{
	const char fname[] = "mxsrv_epics_fd_register_fn()";

	MXSRV_SOCKET_HANDLER_LIST *socket_handler_list;
	MXSRV_SOCKET_HANDLER *old_socket_handler, *new_socket_handler;
	MX_SOCKET current_socket;
	int i, handler_array_size;

	if ( userarg == NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
					"NULL userarg pointer passed." );
		return;
	}

	socket_handler_list = (MXSRV_SOCKET_HANDLER_LIST *) userarg;

	handler_array_size = socket_handler_list->handler_array_size;

#ifndef OS_WIN32
	if ( fd < 0 || fd >= socket_handler_list->max_sockets ) {
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"File descriptor %d passed is outside the allowed range: (0 - %d)",
			fd, socket_handler_list->max_sockets - 1 );
		return;
	}
#endif /* OS_WIN32 */

	if ( opened ) {    /* Adding a new socket. */

		new_socket_handler = (MXSRV_SOCKET_HANDLER *)
					malloc( sizeof(MXSRV_SOCKET_HANDLER) );

		if ( new_socket_handler == NULL ) {
			(void) mx_error( MXE_OUT_OF_MEMORY, fname,
"Attempt to allocate a new socket handler for new EPICS socket %d failed.",
				fd );
			return;
		}

		new_socket_handler->synchronous_socket = fd;
		new_socket_handler->callback_socket = MX_INVALID_SOCKET;
		new_socket_handler->event_handler = mxsrv_epics_event_handler;

		for ( i = 0; i < handler_array_size; i++ ) {

			/* Look for an empty slot in the array. */

			if ( socket_handler_list->array[i] == NULL ) {
				break;             /* Exit the for() loop. */
			}
		}

		if ( i >= handler_array_size ) {
			free( new_socket_handler );

			(void) mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Attempted to use more than the maximum number (%d) "
			"of socket handlers allowed.", handler_array_size );
		}

		new_socket_handler->handler_array_index = i;

		socket_handler_list->array[i] = new_socket_handler;

		socket_handler_list->num_sockets_in_use++;

		MX_DEBUG( 1,(
"%s: EPICS socket handler %d for socket %d added to socket handler list.",
			fname, i, fd ));

	} else {	/* Removing an old socket. */

		/* Have to search the socket handler list for the socket. */

		for ( i = 0; i < handler_array_size; i++ ) {
			if ( socket_handler_list->array[i] != NULL ) {
				current_socket
			   = socket_handler_list->array[i]->synchronous_socket;

				if ( current_socket == fd ) {
					break;    /* Exit the for() loop. */
				}
			}
		}

		if ( i >= handler_array_size ) {
			(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"EPICS has asked that we deregister socket %d, but we do not have an entry "
"for that socket in the socket handler array.", fd );
		}

		old_socket_handler = socket_handler_list->array[i];

		socket_handler_list->array[i] = NULL;

		socket_handler_list->num_sockets_in_use--;

		MX_DEBUG( 1,(
"%s: EPICS socket handler %d for socket %d removed from socket handler list.",
			fname, i, fd ));

		/* Invalidate the contents of the socket handler
		 * just in case someone has a pointer to it.
		 */

		old_socket_handler->synchronous_socket = MX_INVALID_SOCKET;
		old_socket_handler->callback_socket = MX_INVALID_SOCKET;
		old_socket_handler->handler_array_index = -1;
		old_socket_handler->event_handler = NULL;

		free( old_socket_handler );
	}
	return;
}

/*=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=*/

/*
 * mxsrv_epics_mx_value_changed_callback() is invoked when a value change
 * in the MX database needs to be propagated to the EPICS database.
 */

mx_status_type
mxsrv_epics_mx_value_changed_callback( MX_RECORD_FIELD *record_field,
					void *callback_arguments )
{
	const char fname[] = "mxsrv_epics_mx_value_changed_callback()";

	MX_EPICS_CHANNEL *epics_channel;
	mx_status_type ( *process_fn ) (MX_EPICS_CHANNEL *, int, void *);
	char pvname[35];
	void *value_ptr;
	mx_status_type status;

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );
	}

	/* For this callback, the callback arguments are an 
	 * MX_EPICS_CHANNEL structure pointer.
	 */

	epics_channel = (MX_EPICS_CHANNEL *) callback_arguments;

	if ( epics_channel == (MX_EPICS_CHANNEL *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"callback_arguments == NULL, but was expected to contain "
		"a pointer to an MX_EPICS_CHANNEL structure." );
	}

	/* Construct a pointer to the database value for this record field. */

	if ( record_field->flags & MXFF_VARARGS ) {
		value_ptr = mx_read_void_pointer_from_memory_location(
						record_field->data_pointer );
	} else {
		value_ptr = record_field->data_pointer;
	}

	process_fn = epics_channel->process_fn;

	if ( process_fn == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"EPICS process_fn pointer for PV '%s' is NULL.", pvname );
	}

	MX_DEBUG( 1,("%s >>> invoked for channel '%s'.",
			fname, ca_name(epics_channel->channel_id)));

	/* Send the value to EPICS. */

	status = ( *process_fn ) ( epics_channel, MXSRV_WRITE_TO_EPICS, NULL );

	MX_DEBUG( 1,("%s >>> complete for channel '%s'.",
			fname, ca_name(epics_channel->channel_id)));

	return status;
}

/*=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=*/

/*
 * mxsrv_epics_value_changed_callback() is invoked when a value change
 * in the EPICS database needs to be propagated to the MX database.
 */

void
mxsrv_epics_value_changed_callback( struct event_handler_args args )
{
	const char fname[] = "mxsrv_epics_value_changed_callback()";

	MX_EPICS_CHANNEL *channel;
	MXSRV_CALLBACK *callback_found;
	chid channel_id;
	chtype epics_type_from_ca;
	chtype epics_type_from_mx, epics_dbr_time_type_from_mx;
	long element_count;
	size_t bytes_to_move;
	void *mx_value_ptr;
	TS_STAMP *last_put_time;
	int epics_status, event_too_old;

	MXEP_CONST void *epics_value_ptr;
	MXEP_CONST TS_STAMP *timestamp;

	MX_RECORD *mx_record;
	MX_RECORD_FIELD *mx_record_field;
	mx_status_type (*mx_process_fn) (void *, void *, int);
	long mx_type;
	mx_status_type mx_status;

	channel            = (MX_EPICS_CHANNEL *) (args.usr);
	channel_id         = args.chid;
	epics_type_from_ca = args.type;
	element_count      = args.count;
	epics_status       = args.status;

	MX_DEBUG( 1,("%s <<< invoked for channel '%s'.",
			fname, ca_name(channel_id)));

	if ( channel->discard_next_event ) {
		MX_DEBUG( 1,("%s: discarding next event for channel '%s'.",
			fname, ca_name(channel_id)));

		channel->discard_next_event = FALSE;

		return;
	}
	if ( channel->discard_events ) {
		MX_DEBUG( 1,("%s: discarding event for channel '%s'.",
			fname, ca_name(channel_id)));
		return;
	}

	switch( epics_type_from_ca ) {
	case DBR_TIME_STRING:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_string *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_string *) args.dbr )->stamp );

		break;
	case DBR_TIME_CHAR:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_char *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_char *) args.dbr )->stamp );

		break;
	case DBR_TIME_SHORT:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_short *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_short *) args.dbr )->stamp );

		break;
	case DBR_TIME_LONG:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_long *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_long *) args.dbr )->stamp );

		break;
	case DBR_TIME_FLOAT:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_float *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_float *) args.dbr )->stamp );

		break;
	case DBR_TIME_DOUBLE:
		epics_value_ptr = ( MXEP_CONST void * )
	    &( ( (MXEP_CONST struct dbr_time_double *) args.dbr )->value );

		timestamp
	    = &( ( (MXEP_CONST struct dbr_time_double *) args.dbr )->stamp );

		break;
	default:
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"The EPICS DBR_... type %ld is not currently supported.",
			(long) epics_type_from_ca );
		return;
	}

	last_put_time = &( channel->last_put_time );

	event_too_old = FALSE;

	if ( timestamp->secPastEpoch < last_put_time->secPastEpoch ) {

		event_too_old = TRUE;

	} else if ( timestamp->secPastEpoch == last_put_time->secPastEpoch ) {

		if ( timestamp->nsec < last_put_time->nsec ) {

			event_too_old = TRUE;
		}
	}

	if ( event_too_old ) {
		MX_DEBUG( 1,("%s: discarding old event for channel '%s'.",
			fname, ca_name(channel_id)));
		return;
	}

	/*===*/

	mx_record = channel->mx_record;

	mx_record_field = channel->mx_record_field;

	mx_type = mx_record_field->datatype;

	mx_status = mx_epics_convert_mx_type_to_epics_type(
						mx_type, &epics_type_from_mx );

	if ( mx_status.code != MXE_SUCCESS )
		return;

	epics_dbr_time_type_from_mx =
	  mxsrv_epics_convert_dbr_type_to_dbr_time_type( epics_type_from_mx );

	if ( epics_type_from_ca != epics_dbr_time_type_from_mx ) {
		(void) mx_error( MXE_TYPE_MISMATCH, fname,
	"EPICS field type %ld returned by Channel Access does not match "
	"the EPICS field type %ld that would be predicted from the "
	"MX field type %ld",
			epics_type_from_ca,
			epics_dbr_time_type_from_mx,
			mx_type );
		return;
	}

	if ( element_count <= 0 ) {
		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Received less than or equal to zero values (element_count = %ld)",
			element_count );
		return;
	}
	if ( mx_record_field->num_dimensions == 0 ) {
		element_count = 1;
	} else {
		if ( element_count > mx_record_field->dimension[0] )
			element_count = mx_record_field->dimension[0];
	}
	if ( mx_record_field->flags & MXFF_VARARGS ) {
		mx_value_ptr = mx_read_void_pointer_from_memory_location(
					mx_record_field->data_pointer );
	} else {
		mx_value_ptr = mx_record_field->data_pointer;
	}

	switch( mx_type ) {
	case MXFT_STRING:
		bytes_to_move = 1 + element_count * sizeof(char);
		break;
	case MXFT_CHAR:
		bytes_to_move = element_count * sizeof(char);
		break;
	case MXFT_UCHAR:
		bytes_to_move = element_count * sizeof(unsigned char);
		break;
	case MXFT_SHORT:
		bytes_to_move = element_count * sizeof(short);
		break;
	case MXFT_USHORT:
		bytes_to_move = element_count * sizeof(unsigned short);
		break;
	case MXFT_INT:
		bytes_to_move = element_count * sizeof(int);
		break;
	case MXFT_UINT:
		bytes_to_move = element_count * sizeof(unsigned int);
		break;
	case MXFT_LONG:
		bytes_to_move = element_count * sizeof(long);
		break;
	case MXFT_ULONG:
		bytes_to_move = element_count * sizeof(unsigned long);
		break;
	case MXFT_FLOAT:
		bytes_to_move = element_count * sizeof(float);
		break;
	case MXFT_DOUBLE:
		bytes_to_move = element_count * sizeof(double);
		break;
	default:
		(void) mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"MX field type %ld may not be attached to an EPICS process variable.",
			mx_type );
		return;
		break;
	}

	/* Copy the data to the record field. */

	memcpy( mx_value_ptr, epics_value_ptr, bytes_to_move );

#if 0
	{
		double *double_ptr;

		if ( mx_type == MXFT_DOUBLE ) {
			double_ptr = (double *) mx_value_ptr;

			MX_DEBUG(-2,(
			"%s: *double_ptr = %g", fname, *double_ptr));
		}
	}
#endif

	/* If there is one, call the MX process function to write data
	 * to the hardware.
	 *
	 * The MX value changed callback to EPICS must be disabled during
	 * this call and reenabled afterwards.  Otherwise, you will get
	 * an infinite loop of callbacks between EPICS and MX.
	 */

	mx_process_fn = mx_record_field->process_function;

	if ( mx_process_fn != NULL ) {

		mx_status = mxsrv_find_callback( MXSF_VALUE_CHANGED_CALLBACK,
				mx_record_field,
				mxsrv_epics_mx_value_changed_callback,
				channel,
				&callback_found );

		if ( mx_status.code != MXE_SUCCESS )
			return;

		callback_found->callback_enabled = FALSE;

		mx_status = ( *mx_process_fn )
			( mx_record, mx_record_field, MX_PROCESS_PUT );

		callback_found->callback_enabled = TRUE;

		if ( mx_status.code != MXE_SUCCESS )
			return;
	}

	MX_DEBUG( 1,("%s <<< succeeded for channel '%s'.",
			fname, ca_name(channel_id)));

	return;
}

void
mxsrv_epics_change_connection_callback( struct connection_handler_args args )
{
	const char fname[] = "mxsrv_epics_change_connection_callback()";

	MX_EPICS_CHANNEL *channel, *current_channel;
	chid channel_id;
	long op;
	int i;

	channel_id = args.chid;
	op         = args.op;

	/* What kind of message was it? */

	if ( op == CA_OP_CONN_DOWN ) {
		(void) mx_error( MXE_NETWORK_IO_ERROR, fname,
			"EPICS channel '%s' connection is down.",
			ca_name(channel_id) );

		return;

	} else if ( op != CA_OP_CONN_UP ) {

		(void ) mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Unexpected Channel Access message for channel '%s', op = %ld.",
			ca_name(channel_id), op);

		return;
	}

	/* If we get here, it is because the connection is coming back up
	 * after having been down.
	 */

	mx_info( "EPICS channel '%s' connection is back up.",
			ca_name(channel_id) );

	/* Must do a linear search of the array mxsrv_epics_channel_list
	 * to find the MX_EPICS_CHANNEL structure corresponding to
	 * this channel id.
	 */

	channel = NULL;

	for ( i = 0; i < mxsrv_epics_num_channels; i++ ) {

		current_channel = &mxsrv_epics_channel_list[i];

		if ( current_channel->channel_id == channel_id ) {
			channel = current_channel;
			break;
		}
	}

	if ( channel == NULL ) {
		(void) mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
			"The channel id %p for channel '%s' was not found "
			"in the mxsrv_epics_channel_list array.",
			channel_id, ca_name(channel_id) );
		return;
	}

	MX_DEBUG( 1,("%s: record = '%s', field = '%s'",
		fname, channel->mx_record->name,
		channel->mx_record_field->name ));

	/* When the channel comes back up, we will be sent an unwanted
	 * value changed callback.  We must ensure that the next value
	 * changed callback for this channel is discarded.
	 */

	channel->discard_next_event = TRUE;

	return;
}

/*=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=+=*/

mx_status_type
mxsrv_epics_standard_process_fn( MX_EPICS_CHANNEL *channel,
			int direction,
			void *value_ptr )
{
	const char fname[] = "mxsrv_epics_standard_process_fn()";

	MX_RECORD_FIELD *record_field;
	char operation_string[30];
	void *field_value_ptr;
	unsigned long element_count;
	chtype epics_type;
	mx_status_type mx_status;

	switch( direction ) {
	case MXSRV_READ_FROM_EPICS:
		strcpy( operation_string, "(READ_FROM_EPICS)" );
		break;
	case MXSRV_WRITE_TO_EPICS:
		strcpy( operation_string, "(WRITE_TO_EPICS)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", direction );
		break;
	}

	MX_DEBUG( 1,("%s: Channel '%s' %s", fname,
			ca_name( channel->channel_id ), operation_string));

	record_field = channel->mx_record_field;

	if ( value_ptr != NULL ) {
		field_value_ptr = value_ptr;
	} else {
		if ( record_field->flags & MXFF_VARARGS ) {
			field_value_ptr
				= mx_read_void_pointer_from_memory_location(
					record_field->data_pointer );
		} else {
			field_value_ptr = record_field->data_pointer;
		}
	}

	mx_status = mx_epics_convert_mx_type_to_epics_type(
			record_field->datatype, &epics_type );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( record_field->num_dimensions == 0 ) {
		element_count = 1;
	} else {
		element_count = ca_element_count( channel->channel_id );

		if ( element_count > record_field->dimension[0] ) {
			element_count = record_field->dimension[0];
		}
	}

	switch( direction ) {
	case MXSRV_READ_FROM_EPICS:
		MX_DEBUG( 1,("%s: MXSRV_READ_FROM_EPICS not yet implemented.",
			fname ));
		break;
	case MXSRV_WRITE_TO_EPICS:
#if 0
		MX_DEBUG(-2,("%s: epics_type = %ld", fname, epics_type));
		MX_DEBUG(-2,("%s: element_count = %ld", fname, element_count));
		MX_DEBUG(-2,("%s: channel = '%s'",
				fname, ca_name(channel->channel_id) ));
		MX_DEBUG(-2,("%s: field_value_ptr = %p",
				fname, field_value_ptr));
#endif

#if 0
		{
			double *double_ptr;

			if ( record_field->datatype == MXFT_DOUBLE ) {
				double_ptr = (double *) field_value_ptr;

				MX_DEBUG(-2,(
				"%s: *double_ptr = %g", fname, *double_ptr));
			}
		}
#endif

		mx_status = mxsrv_epics_array_put(
					epics_type,
					element_count,
					channel,
					field_value_ptr );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unrecognized direction flag value %d", direction );
		break;
	}
	return MX_SUCCESSFUL_RESULT;
}

/* The purpose of mxsrv_epics_array_put() is to do a PUT to an EPICS PV
 * that we have placed a monitor on, but arrange for the resulting value
 * changed callback from EPICS to be discarded.
 *
 * The reason we want to do such a thing is that the MX server wants to 
 * know if any other EPICS clients change the PV in the EPICS server.
 * However, the MX server wants to discard value changed callbacks if they
 * were do to EPICS PUT operations done by the MX server itself.  Otherwise,
 * we could have bad things happen like infinite loops of events going back
 * and forth between the EPICS and MX server or have move commands from
 * MX clients to motor controllers be issued twice instead of once.
 */

mx_status_type
mxsrv_epics_array_put( chtype epics_type,
			unsigned long element_count,
			MX_EPICS_CHANNEL *channel,
			void *field_value_ptr )
{
	const char fname[] = "mxsrv_epics_array_put()";

	int epics_status;

	MX_DEBUG( 1,("%s invoked for channel '%s'",
				fname, ca_name( channel->channel_id ) ));

	/* Send the data to the EPICS server. */

	channel->discard_events = TRUE;
	channel->waiting_for_internal_callback = TRUE;

	epics_status = ca_array_put_callback(
					epics_type,
					element_count,
					channel->channel_id,
					field_value_ptr,
					mxsrv_epics_internal_put_callback,
					channel );

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"ca_array_put_callback() failed for PV '%s' "
			"with status = %d, error text = '%s'",
				ca_name( channel->channel_id ),
				epics_status,
				ca_message( epics_status ) );
	}

	MX_DEBUG( 2,("%s: ca_array_put_callback() complete.", fname));

	/* Synchronously wait for notification that the PUT operation
	 * has completed.
	 */

	while ( channel->waiting_for_internal_callback ) {
		MX_DEBUG( 2,("%s: Waiting for PUT callback.", fname));

		ca_pend_event( 0.001 );

		mx_msleep(1);
	}

	/* Get the status of the EPICS PV by asking for a single byte
	 * using channel type DBR_TIME_CHAR.
	 *
	 * This is done in order to get the timestamp for the last time
	 * that this process variable was processed.  The timestamp will
	 * probably be for the PUT operation that just completed.  It is
	 * possible that it will be due to a PUT done by a client of the
	 * EPICS server, but in any case this will be at a later time
	 * than when we did our put.
	 */

	MX_DEBUG( 2,("%s: Calling ca_array_get_callback().", fname));

	channel->waiting_for_internal_callback = TRUE;

	epics_status = ca_array_get_callback(
					DBR_TIME_CHAR,
					1,
					channel->channel_id,
					mxsrv_epics_internal_get_callback,
					channel );

	if ( epics_status != ECA_NORMAL ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"ca_array_get_callback() failed for PV '%s' "
			"with status = %d, error text = '%s'",
				ca_name( channel->channel_id ),
				epics_status,
				ca_message( epics_status ) );
	}

	MX_DEBUG( 2,("%s: ca_array_get_callback() complete.", fname));

	/* Synchronously wait for notification that the GET operation
	 * has completed.
	 */

	while ( channel->waiting_for_internal_callback ) {
		MX_DEBUG( 2,("%s: Waiting for GET callback.", fname));

		ca_pend_event( 0.001 );

		mx_msleep(1);
	}

	/* mxsrv_epics_internal_get_callback() will have stored the timestamp
	 * of the last time the EPICS record was processed by now, so it is
	 * now safe to turn callbacks on and return to our caller.
	 */

	channel->discard_events = FALSE;

	MX_DEBUG( 2,("%s returning.", fname));

	return MX_SUCCESSFUL_RESULT;
}

/* The following two callbacks are reserved for the exclusive use of
 * mxsrv_epics_array_put() defined above and should not be used by any
 * other functions in the server.
 */

void
mxsrv_epics_internal_put_callback( struct event_handler_args args )
{
	MX_EPICS_CHANNEL *channel;

	channel = (MX_EPICS_CHANNEL *) (args.usr);

	MX_DEBUG( 2, ("PUT callback seen for channel '%s'",
				ca_name( channel->channel_id ) ));

	channel->waiting_for_internal_callback = FALSE;

	return;
}

void
mxsrv_epics_internal_get_callback( struct event_handler_args args )
{
	MX_EPICS_CHANNEL *channel;

	MXEP_CONST struct dbr_time_char *time_field_struct;
	MXEP_CONST TS_STAMP *timestamp;

	channel = (MX_EPICS_CHANNEL *) (args.usr);

	MX_DEBUG( 2, ("GET callback seen for channel '%s'",
				ca_name( channel->channel_id ) ));

	/* Find the timestamp for the last time this EPICS record
	 * was processed.
	 */

	time_field_struct = (MXEP_CONST struct dbr_time_char *) (args.dbr);

	timestamp = &( time_field_struct->stamp );

	/* Copy the timestamp to the MX_EPICS_CHANNEL structure. */

	channel->last_put_time.secPastEpoch = timestamp->secPastEpoch;
	channel->last_put_time.nsec         = timestamp->nsec;

	channel->waiting_for_internal_callback = FALSE;

	return;
}

#endif /* HAVE_EPICS */

