/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    ms_epmd.c
 *
 * Purpose: Interface between EPICS mailbox motor records and the server.
 */

#include <stdio.h>

#include "mxconfig.h"

#ifndef OS_UNIX
#undef  HAVE_EPICS
#define HAVE_EPICS 0
#endif

#if HAVE_EPICS

#include <stdlib.h>

#include "mx_util.h"
#include "mx_rec.h"
#include "mx_alloc.h"
#include "mx_sock.h"

#include "ms_serv.h"

/* The following are include files from the EPICS distributions. */

#include "tsDefs.h"
#include "cadef.h"

#include "mx_epics.h"

#include "ms_epics.h"
#include "ms_epmd.h"
#include "m_epmtrd.h"

MX_EPICS_CHANNEL_DEFAULTS mxsrv_epics_motor_db_channel_defaults[] = {
	{"EGU",   "units",              mxsrv_epics_standard_process_fn},
	{"VAL",   "setpoint",           mxsrv_epics_standard_process_fn},
	{"RBV",   "position",           mxsrv_epics_standard_process_fn},
	{"ZDMOV", "busy",               mxsrv_epics_standard_process_fn},
	{"SET",   "set_mode",           mxsrv_epics_standard_process_fn},
	{"STOP",  "soft_abort",         mxsrv_epics_standard_process_fn},
	{"LLS",   "negative_limit_hit", mxsrv_epics_standard_process_fn},
	{"HLS",   "positive_limit_hit", mxsrv_epics_standard_process_fn},
	{"LLM",   "negative_limit",     mxsrv_epics_standard_process_fn},
	{"HLM",   "positive_limit",     mxsrv_epics_standard_process_fn},
};

int mxsrv_epics_motor_db_num_channel_defaults =
			sizeof( mxsrv_epics_motor_db_channel_defaults )
			/ sizeof( mxsrv_epics_motor_db_channel_defaults[0] );

mx_status_type
mxsrv_epics_motor_db_add_channels( MX_RECORD *record,
			MX_EPICS_CHANNEL *channel_list,
			int current_channel_number )
{
	const char fname[] = "mxsrv_epics_motor_db_add_channels()";

	chtype epics_dbr_type, epics_dbr_time_type;
	MX_EPICS_MAILBOX_MOTOR_DB *motor_db;
	MX_EPICS_CHANNEL *current_channel;
	MX_EPICS_CHANNEL_DEFAULTS *current_channel_defaults;
	MX_RECORD_FIELD *record_field;
	char pvname[35];
	int i, j, epics_status;
	unsigned long element_count;
	mx_status_type status;

	MX_DEBUG( 1,("%s invoked for record '%s'", fname, record->name));

	motor_db = (MX_EPICS_MAILBOX_MOTOR_DB *)(record->record_type_struct);

	if ( motor_db == (MX_EPICS_MAILBOX_MOTOR_DB *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_EPICS_MAILBOX_MOTOR_DB pointer for record '%s' is NULL.",
			record->name );
	}

	for ( i = 0, j = current_channel_number;
		i < mxsrv_epics_motor_db_num_channel_defaults; i++, j++ ) {

		current_channel = &channel_list[j];
		current_channel_defaults
			= &mxsrv_epics_motor_db_channel_defaults[i];

#if 0
		MX_DEBUG(-2,("%s: motor_record = '%s'",
					fname, motor_db->motor_record->name));
		MX_DEBUG(-2,("%s: epics_namestem = '%s'",
					fname, motor_db->epics_namestem));
#endif

		sprintf( pvname, "%s_%s", motor_db->epics_namestem,
			current_channel_defaults->epics_field_name );

		MX_DEBUG( 1,("%s: Setting up process variable '%s'.",
					fname, pvname));

		/* Setup the MX_EPICS_CHANNEL structure. */

		current_channel->process_fn
				= current_channel_defaults->process_fn;

		current_channel->mailbox_record = record;
		current_channel->mx_record = motor_db->motor_record;

		current_channel->discard_events = FALSE;
		current_channel->discard_next_event = TRUE;
		current_channel->put_callback_in_progress = FALSE;
		current_channel->last_put_time.secPastEpoch = 0;
		current_channel->last_put_time.nsec = 0;
		current_channel->waiting_for_internal_callback = FALSE;

		status = mx_find_record_field(
				motor_db->motor_record,
				current_channel_defaults->mx_field_name,
				&record_field );

		if ( status.code != MXE_SUCCESS )
			return status;

		if ( record_field->num_dimensions > 1 ) {
			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Record field %s.%s referred to by mailbox %s has > 1 dimensions "
"(num_dimensions = %ld).  EPICS database records cannot support this.",
					motor_db->motor_record->name,
					record_field->name,
					record->name,
					record_field->num_dimensions );
		}
		current_channel->mx_record_field = record_field;

		status = mx_epics_convert_mx_type_to_epics_type(
				record_field->datatype, &epics_dbr_type );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Try to find the specified pvname in an EPICS IOC. */

		MX_DEBUG( 1,("%s: Calling ca_search('%s',...)",
					fname, pvname));

		epics_status = ca_search(pvname,
					&(current_channel->channel_id));

		if ( epics_status != ECA_NORMAL ) {
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"ca_search() failed for PV '%s' with status = %d, error text = '%s'",
			pvname, epics_status, ca_message(epics_status) );
		}

		epics_status = ca_pend_io( 3.0 );

		if ( epics_status != ECA_NORMAL ) {
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"ca_pend_io() failed for PV '%s' with status = %d, error text = '%s'",
			pvname, epics_status, ca_message(epics_status) );
		}

		MX_DEBUG( 1,("%s: ca_search() succeeded.", fname));

		/* How big is the EPICS process variable? */

		if ( record_field->num_dimensions == 0 ) {
			element_count = 1;
		} else {
			element_count = ca_element_count(
						current_channel->channel_id);

			if ( element_count > record_field->dimension[0] ) {
				element_count = record_field->dimension[0];
			}
		}

		/* Set up an MX callback handler to send value changes
		 * from MX to EPICS.
		 */

		MX_DEBUG( 1,("%s: Adding MX value changed handler for '%s'",
				fname, pvname));

		status = mxsrv_add_callback(
				MXSF_VALUE_CHANGED_CALLBACK,
				record_field,
				mxsrv_epics_mx_value_changed_callback,
				current_channel );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Invoke the MX callback handler to send the current
		 * MX value to EPICS.
		 */

		MX_DEBUG( 1,("%s: Invoking MX value changed handler for '%s'",
				fname, pvname));

		status = mxsrv_epics_mx_value_changed_callback(
				record_field, current_channel );

		if ( status.code != MXE_SUCCESS )
			return status;

		/* Set up an EPICS callback handler to get value changes
		 * in EPICS sent back to MX.  Use a DBR_TIME_XXXX type
		 * so that the value changed callback is also sent a 
		 * timestamp, status, etc.
		 */

		MX_DEBUG( 1,("%s: Adding EPICS value changed handler for '%s'",
				fname, pvname));

		epics_dbr_time_type =
	mxsrv_epics_convert_dbr_type_to_dbr_time_type( epics_dbr_type );

		epics_status = ca_add_masked_array_event(
				epics_dbr_time_type,
				element_count,
				current_channel->channel_id,
				mxsrv_epics_value_changed_callback,
				current_channel,
				(float)0, (float)0, (float)0, 
				&(current_channel->event_id),
				DBE_VALUE );

		if ( epics_status != ECA_NORMAL ) {
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"ca_add_masked_array_event() failed for PV '%s' with status = %d, "
	"error text = '%s'",
			pvname, epics_status, ca_message(epics_status) );
		}

		/* Set up a change connection handler here. */

		MX_DEBUG( 1,("%s: Adding EPICS connection handler for '%s'",
				fname, pvname));

		epics_status = ca_change_connection_event(
			current_channel->channel_id,
			mxsrv_epics_change_connection_callback );

		if ( epics_status != ECA_NORMAL ) {
			return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"ca_change_connection_event() failed for PV '%s' with status = %d",
				pvname, epics_status );
		}
	}

	MX_DEBUG( 1,("%s returning for record '%s'.", fname, record->name ));

	return MX_SUCCESSFUL_RESULT;
}

#endif /* HAVE_EPICS */
