/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    ms_mxsrv.c
 *
 * Purpose: This file implements the server side of the MX network protocol
 *          for the generic MX server.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_types.h"
#include "mx_rec.h"
#include "mx_sock.h"
#include "mx_net.h"
#include "mx_alloc.h"

#include "ms_serv.h"
#include "ms_mxsrv.h"
#include "ms_secur.h"

#include "ms_prrec.h"
#include "ms_prain.h"
#include "ms_praou.h"
#include "ms_prdin.h"
#include "ms_prdou.h"
#include "ms_prmtr.h"
#include "ms_prscl.h"
#include "ms_prtim.h"
#include "ms_pramp.h"

MXSRV_MX_SERVER_SOCKET mxsrv_mx_server_socket_struct;

static mx_status_type ( *mxsrv_process_function_setup_array[] )(MX_RECORD *) =
	{
		mxsrv_setup_record_process_functions,
		mxsrv_setup_analog_input_process_functions,
		mxsrv_setup_analog_output_process_functions,
		mxsrv_setup_digital_input_process_functions,
		mxsrv_setup_digital_output_process_functions,
		mxsrv_setup_motor_process_functions,
		mxsrv_setup_scaler_process_functions,
		mxsrv_setup_timer_process_functions,
		mxsrv_setup_amplifier_process_functions,
	};

static int mxsrv_num_process_function_setup_elements =
			sizeof( mxsrv_process_function_setup_array )
			/ sizeof( mxsrv_process_function_setup_array[0] );

long
mxsrv_get_returning_message_type( long message_type )
{
	long returning_message_type;
	long i;

	static long message_pairs[][2] = {

    { MX_NETMSG_GET_ARRAY_ASCII_COMMAND, MX_NETMSG_GET_ARRAY_ASCII_RESPONSE },
    { MX_NETMSG_PUT_ARRAY_ASCII_COMMAND, MX_NETMSG_PUT_ARRAY_ASCII_RESPONSE },
    { MX_NETMSG_GET_FIELD_TYPE_COMMAND,  MX_NETMSG_GET_FIELD_TYPE_RESPONSE },

	};

	static long num_message_pairs = sizeof( message_pairs )
					/ sizeof( message_pairs[0] );

	for ( i = 0; i < num_message_pairs; i++ ) {
		if ( message_type == message_pairs[i][0] ) {
			break;		/* Exit the for() loop. */
		}
	}
	if ( i >= num_message_pairs ) {
		returning_message_type = MX_NETMSG_UNEXPECTED_ERROR;
	} else {
		returning_message_type = message_pairs[i][1];
	}
	return returning_message_type;
}

mx_status_type
mxsrv_free_client_socket_handler( MXSRV_SOCKET_HANDLER *socket_handler,
			MXSRV_SOCKET_HANDLER_LIST *socket_handler_list )
{
	int i;
	mx_status_type status;

	i = socket_handler->handler_array_index;

	socket_handler_list->array[i] = NULL;

	mx_info("Client %d (socket %d) disconnected.",
		i, socket_handler->synchronous_socket);

	/* If this handler had a callback socket, close it. */

	if ( socket_handler->callback_socket != MX_INVALID_SOCKET ) {
		status = mx_socket_close( socket_handler->callback_socket );

		socket_handler_list->num_sockets_in_use--;

		if ( status.code != MXE_SUCCESS ) {
			free( socket_handler );
			return status;
		}
	}

	/* Close our end of the synchronous socket. */

	status = mx_socket_close( socket_handler->synchronous_socket );

	socket_handler_list->num_sockets_in_use--;

	if ( status.code != MXE_SUCCESS ) {
		free( socket_handler );
		return status;
	}

	/* Invalidate the contents of the socket handler just in case
	 * someone has a pointer to it.
	 */

	socket_handler->synchronous_socket = MX_INVALID_SOCKET;
	socket_handler->callback_socket = MX_INVALID_SOCKET;
	socket_handler->handler_array_index = -1;
	socket_handler->event_handler = NULL;

	free( socket_handler );

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_server_socket_init( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	const char fname[] = "mxsrv_mx_server_socket_init()";

	MXSRV_MX_SERVER_SOCKET *server_socket_struct;
	MX_SOCKET server_socket;
	mx_status_type ( *process_function_setup ) ( MX_RECORD * );
	int i, port_number, max_sockets, handler_array_size;
	mx_status_type status;

	MX_DEBUG(1,("%s invoked.", fname));

	if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_SOCKET_HANDLER_LIST pointer passed is NULL." );
	}
	if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_EVENT_HANDLER pointer passed is NULL." );
	}

	max_sockets = socket_handler_list->max_sockets;
	handler_array_size = socket_handler_list->handler_array_size;

	/* Get the port number of the server socket we are supposed to open. */

	server_socket_struct = (MXSRV_MX_SERVER_SOCKET *)
				( event_handler->type_struct );

	if ( server_socket_struct == (MXSRV_MX_SERVER_SOCKET *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
"MXSRV_MX_SERVER_SOCKET pointer for server event handler '%s' is NULL.",
			event_handler->name );
	}

	port_number = server_socket_struct->port;

	if ( port_number < 0 || port_number > 65535 ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Illegal server port number %d for server event handler '%s'",
			port_number, event_handler->name );
	}

	/* Initialize the MX process functions before we create
	 * the server socket.
	 */

#if 0
	status = mxsrv_setup_motor_process_functions( record_list );

	if ( status.code != MXE_SUCCESS )
		return status;
#else
	for ( i = 0; i < mxsrv_num_process_function_setup_elements; i++ ) {

		process_function_setup = mxsrv_process_function_setup_array[i];

		if ( process_function_setup == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"mxsrv_process_function_setup_array has a NULL pointer in it.");
		}

		status = ( *process_function_setup ) ( record_list );

		if ( status.code != MXE_SUCCESS )
			return status;
	}
#endif

	/*********** Try to create the server socket. ************/

	status = mx_socket_open_as_server( &server_socket, port_number, 0 );

	if ( status.code != MXE_SUCCESS )
		return status;

#ifndef OS_WIN32
	if ( server_socket < 0 || server_socket >= max_sockets ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Server socket value %d returned by mx_socket_open_as_server() is outside "
"the allowed range of 0 to %d.", server_socket, max_sockets - 1 );
	}
#endif /* OS_WIN32 */

	/* Add the socket to the list of sockets that we handle. */

	for ( i = 0; i < handler_array_size; i++ ) {

		/* Look for an empty slot in the array. */

		if ( socket_handler_list->array[i] == NULL ) {
			break;                     /* Exit the for() loop. */
		}
	}

	if ( i >= handler_array_size ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Attempted to use more than the maximum number (%d) "
			"of socket handlers allowed.", handler_array_size );
	}

	socket_handler_list->array[i] = (MXSRV_SOCKET_HANDLER *)
				malloc( sizeof(MXSRV_SOCKET_HANDLER) );

	if ( socket_handler_list->array[i] == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
"Ran out of memory trying to allocate MXSRV_SOCKET_HANDLER for socket %d.",
			server_socket );
	}

	socket_handler_list->array[i]->synchronous_socket = server_socket;
	socket_handler_list->array[i]->callback_socket = MX_INVALID_SOCKET;
	socket_handler_list->array[i]->handler_array_index = i;
	socket_handler_list->array[i]->event_handler = event_handler;

	socket_handler_list->num_sockets_in_use++;

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_server_socket_process_event( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER *socket_handler,
                                MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	const char fname[] = "mxsrv_mx_server_socket_process_event()";

	MX_LIST_HEAD *list_head;
	MXSRV_MX_SERVER_SOCKET *server_socket_struct;
	MXSRV_SOCKET_HANDLER *new_socket_handler;
	MX_SOCKET client_socket;
	int i, max_sockets, handler_array_size;
	int saved_errno, connection_allowed;
	char *error_string;
	struct sockaddr_in client_address;
	int client_address_size = sizeof( struct sockaddr_in );
	char client_address_string[MXU_ADDRESS_STRING_LENGTH + 1];
	char *client_address_string_ptr;
	mx_status_type status;

#ifdef OS_DJGPP
	unsigned char tempbuffer[255];
#endif

	MX_DEBUG(1,("%s invoked.", fname));

	if ( socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MXSRV_SOCKET_HANDLER pointer passed was NULL." );
	}
	if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_SOCKET_HANDLER_LIST pointer passed was NULL." );
	}
	if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MXSRV_EVENT_HANDLER pointer passed was NULL." );
	}

	max_sockets = socket_handler_list->max_sockets;
	handler_array_size = socket_handler_list->handler_array_size;

	if ( socket_handler_list->num_sockets_in_use >= (max_sockets - 1) ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Maximum number (%d) of sockets allowed reached.  "
			"Cannot handle client request for a new connection.",
			max_sockets );
	}

	server_socket_struct = (MXSRV_MX_SERVER_SOCKET *)
					(event_handler->type_struct);

	if ( server_socket_struct == (MXSRV_MX_SERVER_SOCKET *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"MXSRV_MX_SERVER_SOCKET pointer for event handler type '%s' is NULL.",
			event_handler->name );
	}

	if ( server_socket_struct->client_event_handler == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
	"client_event_handler pointer for event handler type '%s' is NULL.",
			event_handler->name );
	}

	/* The server socket should be ready for an accept() operation,
	 * so let's try it.
	 */

	client_socket = accept( socket_handler->synchronous_socket,
				( struct sockaddr *) &client_address,
				&client_address_size );

        saved_errno = mx_socket_check_error_status(
                        &client_socket, MXF_SOCKCHK_INVALID, &error_string );

        if ( saved_errno != 0 ) {
                return mx_error( MXE_NETWORK_IO_ERROR, fname,
               "Attempt to perform accept() on MX server port failed.  "
                        "Errno = %d.  Error string = '%s'.",
                        saved_errno, error_string );
        }

	/* Convert the host address of the socket to standard numbers-and-dots
	 * notation.
	 */

#ifdef OS_DJGPP
	client_address_string_ptr
		    = inet_ntoa( tempbuffer, client_address.sin_addr.s_addr );
#else
	client_address_string_ptr = inet_ntoa( client_address.sin_addr );
#endif

	mx_strncpy( client_address_string, client_address_string_ptr,
						MXU_ADDRESS_STRING_LENGTH );

	/* Check the connection access control list to see if the remote
	 * client is permitted to connect to us.
	 */

	list_head = mx_get_record_list_head_struct( record_list );

	if ( list_head == NULL ) {
		mx_info( "Exiting due to corrupt list head record." );
		exit( MXE_CORRUPT_DATA_STRUCTURE );
	}

	if ( list_head->connection_acl == NULL ) {
		connection_allowed = TRUE;
	} else {
		status = mxsrv_check_socket_connection_acl_permissions(
				record_list, client_address_string,
				&connection_allowed );

		if ( status.code != MXE_SUCCESS )
			return status;
	}

	if ( connection_allowed == FALSE ) {
		(void) mx_socket_close( client_socket );

		return mx_error( MXE_CLIENT_REQUEST_DENIED, fname,
		"Network client '%s' is not allowed to connect to us.",
			client_address_string );
	}

	/* Now that we seem to have a working client socket, let's put
	 * it into the socket handler list.
	 */

	new_socket_handler = (MXSRV_SOCKET_HANDLER *)
				malloc( sizeof(MXSRV_SOCKET_HANDLER) );

	if ( new_socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
"Attempt to allocate a new socket handler for new client socket %d failed.",
			client_socket );
	}

	new_socket_handler->synchronous_socket = client_socket;
	new_socket_handler->callback_socket = MX_INVALID_SOCKET;
	new_socket_handler->event_handler
			= server_socket_struct->client_event_handler;

	for ( i = 0; i < handler_array_size; i++ ) {

		/* Look for an empty slot in the array. */

		if ( socket_handler_list->array[i] == NULL ) {
			break;                     /* Exit the for() loop. */
		}
	}

	if ( i >= handler_array_size ) {
		free(new_socket_handler);

		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
			"Attempted to use more than the maximum number (%d) "
			"of socket handlers allowed.", handler_array_size );
	}

	new_socket_handler->handler_array_index = i;

	socket_handler_list->array[i] = new_socket_handler;

	socket_handler_list->num_sockets_in_use++;

	mx_info("Client %d (socket %d) connected from host '%s'.",
			i, client_socket, client_address_string );

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_client_socket_init( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	/* Nothing to do here.  Client sockets are actually initialized
	 * in mxsrv_mx_server_socket_process_event().
	 */

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_mx_client_socket_process_event( MX_RECORD *record_list,
				MXSRV_SOCKET_HANDLER *socket_handler,
                                MXSRV_SOCKET_HANDLER_LIST *socket_handler_list,
				MXSRV_EVENT_HANDLER *event_handler )
{
	const char fname[] = "mxsrv_mx_client_socket_process_event()";

	union {
		mx_uint32_type header[MX_NETWORK_HEADER_LENGTH_VALUE];
		char buffer[MX_NETWORK_HEADER_LENGTH_VALUE
					* sizeof(mx_uint32_type)];
	} aligned_buffer;

	mx_uint32_type *header;
	char *receive_buffer;

	int queue_a_message, record_field_name_at_message_start;
	char *record_name, *field_name;
	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_SOCKET client_socket;
	void *event_data;

	char *ptr, *message;
	int saved_errno;
	long bytes_left, bytes_received, initial_recv_length;
	long magic_value, header_length, message_length;
	long message_type, returned_message_type;
	mx_status_type status;

	char separators[] = MX_RECORD_FIELD_SEPARATORS;

        if ( socket_handler == (MXSRV_SOCKET_HANDLER *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                        "MXSRV_SOCKET_HANDLER pointer passed was NULL." );
        }

	record = NULL;

	client_socket = socket_handler->synchronous_socket;

	MX_DEBUG(1,("%s invoked for socket handler %d.", fname,
				socket_handler->handler_array_index));

#if 0
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));
#endif

        if ( socket_handler_list == (MXSRV_SOCKET_HANDLER_LIST *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                "MXSRV_SOCKET_HANDLER_LIST pointer passed was NULL." );
        }
        if ( event_handler == (MXSRV_EVENT_HANDLER *) NULL ) {
                return mx_error( MXE_NULL_ARGUMENT, fname,
                        "MXSRV_EVENT_HANDLER pointer passed was NULL." );
        }

	/* Try to read the beginning of the header of the incoming message. */

	header = &(aligned_buffer.header[0]);
	ptr    = &(aligned_buffer.buffer[0]);

	initial_recv_length = 4 * sizeof( mx_uint32_type );

	bytes_left = initial_recv_length;

	while( bytes_left > 0 ) {
		bytes_received = recv( client_socket, ptr, bytes_left, 0 );

		switch( bytes_received ) {
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			switch( saved_errno ) {
#ifdef OS_WIN32
			case WSAECONNRESET:
#else
			case ECONNRESET:
#endif
				break;
			default:
				return mx_error( MXE_NETWORK_IO_ERROR, fname,
"Error receiving header from remote host.  Errno = %d, error text = '%s'",
			saved_errno, mx_socket_strerror( saved_errno ) );
				break;
			}
			status = mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

			return status;
			break;
		case 0:
			/* The remote socket has closed so remove the
			 * socket handler from the socket handler list.
			 */

			status = mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

			return status;
			break;
		default:
			bytes_left -= bytes_received;
			ptr += bytes_received;
			break;
		}
	}

#if 0
	{
		long i;
		unsigned char c;

		for ( i = 0; i < initial_recv_length; i++ ) {
			c = aligned_buffer.buffer[i];

			if ( isprint(c) ) {
				MX_DEBUG(-2,("buffer[%ld] = %02x (%c)",
						i, c, c));
			} else {
				MX_DEBUG(-2,("buffer[%ld] = %02x", i, c));
			}
		}
	}
#endif

	magic_value    = ntohl( header[ MX_NETWORK_MAGIC ] );
        header_length  = ntohl( header[ MX_NETWORK_HEADER_LENGTH ] );
        message_length = ntohl( header[ MX_NETWORK_MESSAGE_LENGTH ] );
	message_type   = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );

#if 0
        MX_DEBUG(-2,("%s: magic_value    = %lx", fname, magic_value));
        MX_DEBUG(-2,("%s: header_length  = %ld", fname, header_length));
        MX_DEBUG(-2,("%s: message_length = %ld", fname, message_length));
#endif

        MX_DEBUG(1,("%s: message_type   = %ld", fname, message_type));

        if ( magic_value != MX_NETWORK_MAGIC_VALUE ) {
                return mx_error( MXE_NETWORK_IO_ERROR, fname,
                "Wrong magic number %x in received message.", magic_value );
        }

	/* Allocate a buffer to write the message to.
	 *
	 * The one extra byte is so that a '\0' byte can be placed at
	 * the end of the receive buffer for the sake of network commands
	 * that pass their message bodies as strings.
	 */

	receive_buffer = (char *) malloc( header_length + message_length + 1 );

	if ( receive_buffer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating receive buffer for MX message.");
	}

	memcpy( receive_buffer,
			&(aligned_buffer.buffer[0]), initial_recv_length );

        /* Receive the rest of the data. */

	ptr = receive_buffer + initial_recv_length;

        bytes_left = header_length + message_length - initial_recv_length;

        while( bytes_left > 0 ) {
                bytes_received = recv( client_socket, ptr, bytes_left, 0 );

                switch( bytes_received ) {
		case MX_SOCKET_ERROR:
			saved_errno = mx_socket_get_last_error();

			free(receive_buffer);

			(void) mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

                        return mx_error( MXE_NETWORK_IO_ERROR, fname,
                        "Error receiving message body from remote host.  "
                        "Errno = %d, error text = '%s'",
			saved_errno, mx_socket_strerror( saved_errno ) );

                        break;
                case 0:
			free(receive_buffer);

			(void) mxsrv_free_client_socket_handler(
					socket_handler, socket_handler_list );

                        return mx_error( MXE_NETWORK_IO_ERROR, fname,
                                        "Socket closed unexpectedly." );
                        break;
                default:
                        bytes_left -= bytes_received;
                        ptr += bytes_received;
                        break;
                }
        }

	message = receive_buffer + header_length;

	/* Ensure that a '\0' byte is placed after the end of the
	 * message field.
	 */

	message[ message_length ] = '\0';

#if 0
	MX_DEBUG(-2,("%s: message_length = %ld", fname, message_length));
	{
		long i;

		for ( i = 0; i < message_length; i++ ) {
			MX_DEBUG(-2,("message[%ld] = %x '%c'",
				i, message[i], message[i]));
		}
	}
#endif
#if 0
	MX_DEBUG(-2,("%s: message = '%*s'", fname, message_length, message));
#endif

	/* Set some flags depending on the message type. */

	switch( message_type ) {
	case MX_NETMSG_GET_ARRAY_ASCII_COMMAND:
	case MX_NETMSG_PUT_ARRAY_ASCII_COMMAND:
		queue_a_message = TRUE;
		record_field_name_at_message_start = TRUE;
		break;
	case MX_NETMSG_GET_FIELD_TYPE_COMMAND:
		queue_a_message = FALSE;
		record_field_name_at_message_start = TRUE;
		break;
	default:
		queue_a_message = FALSE;
		record_field_name_at_message_start = FALSE;
		break;
	}

	/* If this message type requires it, we parse the beginning of the
	 * message to look for a record name and a field name.  If found,
	 * we then get the corresponding MX_RECORD and MX_RECORD_FIELD
	 * pointers.
	 */

	if ( record_field_name_at_message_start == FALSE ) {
		record = NULL;
		record_field = NULL;
	} else {
		/* The do...while(0) loop below is just a trick to make
		 * it easy to jump to the end of this block of code.
		 */

		do {
			/* Which record and field does this correspond to? */

			record_name = message;

			ptr = strchr( message, '.' );

			if ( ptr == NULL ) {
				status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
				"No period found in record field name '%s'.",
					message );

				break;	/* Exit the do...while(0) loop. */
			}

			*ptr = '\0';

			field_name = ++ptr;

		        ptr = field_name + strcspn( field_name, separators );

		        *ptr = '\0';

			record = mx_get_record( record_list, record_name );

			if ( record == (MX_RECORD *) NULL ) {
				status = mx_error( MXE_UNDEFINED_RECORD, fname,
				"This server has no record named '%s'.",
					record_name );

				break;	/* Exit the do...while(0) loop. */
			}

			status = mx_find_record_field( record, field_name,
							&record_field );
		} while (0);
	}
	if ( status.code != MXE_SUCCESS ) {
		free(receive_buffer);

		returned_message_type = mxsrv_get_returning_message_type(
						message_type );

		/* Send back the error message. */

		(void) mx_socket_send_error_message( client_socket,
					MX_NETMSG_UNEXPECTED_ERROR, status );

		return status;
	}

	/* Do we finish handling this event now or do we queue an event
	 * to be processed later?
	 */

	if ( queue_a_message == TRUE ) {

		event_data = (void *) receive_buffer;

		/* Queue the event. */

		status = mxsrv_add_queued_event( socket_handler,
					record, record_field, event_data );

		if ( status.code != MXE_SUCCESS ) {
			free(receive_buffer);

			returned_message_type
			    = mxsrv_get_returning_message_type( message_type );

			/* Send back the error message. */

			(void) mx_socket_send_error_message( client_socket,
					MX_NETMSG_UNEXPECTED_ERROR, status );
		}

		return MX_SUCCESSFUL_RESULT;
	}

	/* Here we handle message types that are to be dealt
	 * with immediately.
	 */

	switch( message_type ) {
	case MX_NETMSG_GET_FIELD_TYPE_COMMAND:
		status = mxsrv_handle_get_field_type( client_socket,
					record_field, message, message_length);
		break;
	default:
		status = mx_error( MXE_NOT_YET_IMPLEMENTED, fname,
			"MX operation type %ld is not yet implemented.",
			message_type );

		(void) mx_socket_send_error_message( client_socket,
			MX_NETMSG_UNEXPECTED_ERROR, status );
		break;
	}
	free(receive_buffer);

#if 0
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));

	MX_DEBUG(-2,("%s exiting.", fname));
#endif

	return status;
}

mx_status_type
mxsrv_mx_client_socket_proc_queued_event( MXSRV_QUEUED_EVENT *queued_event )
{
	const char fname[] = "mxsrv_mx_client_socket_proc_queued_event()";

	MX_RECORD *record_list;
	MXSRV_SOCKET_HANDLER *socket_handler;
	MX_SOCKET client_socket;
	char *message_buffer;
	mx_uint32_type *header;
	long message_type;
	mx_status_type status;

	MX_DEBUG(1,("%s invoked.", fname));

	if ( queued_event == (MXSRV_QUEUED_EVENT *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MXSRV_QUEUED_EVENT structure pointer passed is NULL." );
	}

	record_list = queued_event->record->list_head;
	socket_handler = queued_event->socket_handler;

#if 0
	MX_DEBUG(-2,("socket_handler = %p", socket_handler));
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));
#endif

	message_buffer = (char *) queued_event->event_data;

	header = (mx_uint32_type *) message_buffer;

	message_type = ntohl( header[ MX_NETWORK_MESSAGE_TYPE ] );

	/* Handle the request. */

	client_socket = socket_handler->synchronous_socket;

	status = MX_SUCCESSFUL_RESULT;

	switch ( message_type ) {
	case MX_NETMSG_GET_ARRAY_ASCII_COMMAND:
		status = mxsrv_handle_get_array_ascii( client_socket,
				queued_event->record,
				queued_event->record_field,
				message_buffer );

		break;
	case MX_NETMSG_PUT_ARRAY_ASCII_COMMAND:
		status = mxsrv_handle_put_array_ascii( client_socket,
				queued_event->record,
				queued_event->record_field,
				message_buffer );

		break;
	default:
		free(message_buffer);

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
			"Illegal client message type %ld", message_type );
		break;
	}

	free(message_buffer);

#if 0
	MX_DEBUG(-2,("socket_handler->synchronous_socket = %d",
				socket_handler->synchronous_socket));
	MX_DEBUG(-2,("socket_handler->callback_socket = %d",
				socket_handler->callback_socket));
	MX_DEBUG(-2,("socket_handler->handler_array_index = %d",
				socket_handler->handler_array_index));
	MX_DEBUG(-2,("socket_handler->event_handler = %p",
				socket_handler->event_handler));

	MX_DEBUG(-2,("%s exiting.", fname));
#endif

	return status;
}

mx_status_type
mxsrv_handle_get_array_ascii( MX_SOCKET mx_socket,
			MX_RECORD *record, MX_RECORD_FIELD *record_field,
			char *received_buffer )
{
	const char fname[] = "mxsrv_handle_get_array_ascii()";

	char *send_buffer, *send_buffer_message;
	mx_uint32_type *send_buffer_header;
	long send_buffer_header_length, send_buffer_message_length;
	mx_status_type (*process_fn) (void *, void *, int);
	void *pointer_to_value;
	mx_status_type ( *token_constructor )
		(void *, char *, size_t, MX_RECORD *, MX_RECORD_FIELD *);
	mx_status_type status;

	status = MX_SUCCESSFUL_RESULT;

	MX_DEBUG(1,("***** %s invoked for socket %d *****", fname, mx_socket));

	/* The do...while(0) loop below is just a trick to make it easy
	 * to jump to the end of this block of code.
	 */

	send_buffer = NULL;
	pointer_to_value = NULL;

	do {
		if ( record == (MX_RECORD *) NULL ) {
			status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

		if ( record_field == (MX_RECORD_FIELD *) NULL ) {
			status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

#if 0
	        MX_DEBUG(-2,("%s: record_name = '%s'", fname, record->name));
		MX_DEBUG(-2,("%s: field_name = '%s'", fname,
						record_field->name));
#endif

	        if ( record_field->flags & MXFF_VARARGS ) {
			pointer_to_value
				= mx_read_void_pointer_from_memory_location(
					record_field->data_pointer );
	        } else {
	                pointer_to_value = record_field->data_pointer;
		}

		/* If there is one, call the process function to read data
		 * from the hardware.
		 */

		process_fn = record_field->process_function;

		if ( process_fn != NULL ) {

			status = ( *process_fn )
				( record, record_field, MX_PROCESS_GET );

			MX_DEBUG(1,(
				"%s: process_fn returned status.code = %ld",
				fname, status.code ));
		}
	} while (0);

	/* Allocate a buffer to construct the message to be sent back to
	 * the client.
	 *
	 * The length of 'send_buffer_message_length' currently is hard coded
	 * which is unfortunate, but hopefully can be fixed later.
	 */

	send_buffer_message_length = 1024;
	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer = (char *) malloc( send_buffer_header_length
						+ send_buffer_message_length );

	if ( send_buffer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating send buffer for MX response message.");
	}

	send_buffer_header = (mx_uint32_type *) send_buffer;

	send_buffer_message = send_buffer + send_buffer_header_length;

	if ( status.code == MXE_SUCCESS ) {

	        /* Construct the string description of this field. */

	        status = mx_get_token_constructor(
                        record_field->datatype, &token_constructor );

		if ( status.code == MXE_SUCCESS ) {
			strcpy( send_buffer_message, "" );

		        if ( (record_field->num_dimensions == 0)
		          || ((record_field->datatype == MXFT_STRING)
		             && (record_field->num_dimensions == 1)) ) {

		                status = (*token_constructor)(
					pointer_to_value,
					send_buffer_message,
					send_buffer_message_length,
					record, record_field );
			} else {
				status = mx_create_array_description(
					pointer_to_value,
					record_field->num_dimensions - 1,
					send_buffer_message,
					send_buffer_message_length,
					record, record_field,
					token_constructor );
			}
	        }
	}

	/* Fill in the header of the message to be sent back to the client. */

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl(status.code);

        if ( status.code != MXE_SUCCESS ) {
		*send_buffer_message = '\0';

		strncat( send_buffer_message,
			status.message, send_buffer_message_length );
	}

	send_buffer_header[ MX_NETWORK_MAGIC ] = htonl(MX_NETWORK_MAGIC_VALUE);
	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
				= htonl(MX_NETMSG_GET_ARRAY_ASCII_RESPONSE);
	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl(send_buffer_header_length);
	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl(strlen( send_buffer_message ) + 1);

	/* Send the string description back to the client. */

#if 0
	MX_DEBUG(-2,("%s: sending response = '%s'",
			fname, send_buffer_message ));
#endif

	status = mx_socket_send_message( mx_socket, send_buffer );

	MX_DEBUG(1,("%s: mx_socket_send_message() returned status.code = %ld",
				fname, status.code ));
	MX_DEBUG(1,("***** %s exiting *****", fname));

	free( send_buffer );

	return status;
}

mx_status_type
mxsrv_handle_put_array_ascii( MX_SOCKET mx_socket,
			MX_RECORD *record, MX_RECORD_FIELD *record_field,
			char *receive_buffer )
{
	const char fname[] = "mxsrv_handle_put_array_ascii()";

	char *receive_buffer_message;
	mx_uint32_type *receive_buffer_header;
	long receive_buffer_header_length;

	char *send_buffer, *send_buffer_message;
	mx_uint32_type *send_buffer_header;
	long send_buffer_header_length, send_buffer_message_length;

	MX_RECORD_FIELD_PARSE_STATUS parse_status;
	mx_status_type (*process_fn) (void *, void *, int);
	char token_buffer[500];
	char *value_string;
	void *pointer_to_value;
	mx_status_type ( *token_parser )
	    ( void *, char *, MX_RECORD *, MX_RECORD_FIELD_PARSE_STATUS * );
	mx_status_type status;

	char separators[] = MX_RECORD_FIELD_SEPARATORS;

	status = MX_SUCCESSFUL_RESULT;

	MX_DEBUG(1,("***** %s invoked for socket %d *****", fname, mx_socket));

	/* The do...while(0) loop below is just a trick to make it easy
	 * to jump to the end of this block of code.
	 */

	do {
		if ( record == (MX_RECORD *) NULL ) {
			status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

		if ( record_field == (MX_RECORD_FIELD *) NULL ) {
			status =  mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );

			break;		/* Exit the do...while(0) loop. */
		}

#if 0
	        MX_DEBUG(-2,("%s: record_name = '%s'", fname, record->name));
		MX_DEBUG(-2,("%s: field_name = '%s'", fname,
						record_field->name));
#endif

	        if ( record_field->flags & MXFF_VARARGS ) {
			pointer_to_value
				= mx_read_void_pointer_from_memory_location(
					record_field->data_pointer );
	        } else {
	                pointer_to_value = record_field->data_pointer;
		}

	        status = mx_get_token_parser(
                        record_field->datatype, &token_parser );

		if ( status.code != MXE_SUCCESS )
			break;		/* Exit the do...while(0) loop. */

		/* Get a pointer to the start of the value string. */

		receive_buffer_header = (mx_uint32_type *) receive_buffer;

		receive_buffer_header_length
		  = ntohl( receive_buffer_header[ MX_NETWORK_HEADER_LENGTH ] );

		receive_buffer_message
			= receive_buffer + receive_buffer_header_length;

		value_string = receive_buffer_message
					+ MXU_RECORD_FIELD_NAME_LENGTH;

		/* Parse the string description of this field and
		 * write it into the corresponding database record.
		 */

#if 0
		MX_DEBUG(-2,("%s: value_string = '%s'", fname, value_string));
#endif

                parse_status.description = value_string;
                parse_status.separators = separators;
                parse_status.num_separators = strlen(separators);
                parse_status.start_of_trailing_whitespace = NULL;

                if ( (record_field->num_dimensions == 0)
                  || ((record_field->datatype == MXFT_STRING)
                     && (record_field->num_dimensions == 1)) ) {

                        status = mx_get_next_record_token( &parse_status,
                                        token_buffer, sizeof(token_buffer) );

                        if ( status.code != MXE_SUCCESS )
                                break;	/* Exit the do...while(0) loop. */

                        status = ( *token_parser ) ( pointer_to_value,
                                        token_buffer, record, &parse_status );
                } else {
                        status = mx_parse_array_description(
                                        pointer_to_value,
                                        (record_field->num_dimensions - 1),
                                        record, record_field,
                                        &parse_status, token_parser );
                }

		if ( status.code != MXE_SUCCESS )
			break;		/* Exit the do...while(0) loop. */

		/* If there is one, call the process function to write data
		 * to the hardware.
		 */

		process_fn = record_field->process_function;

		if ( process_fn != NULL ) {

			status = ( *process_fn )
				( record, record_field, MX_PROCESS_PUT );

			MX_DEBUG(1,(
				"%s: process_fn returned status.code = %ld",
				fname, status.code ));
		}
	} while (0);

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

        if ( status.code == MXE_SUCCESS ) {
		send_buffer_message_length = 1;
	} else {
		send_buffer_message_length = 1 + strlen(status.message);
	}

	send_buffer = (char *) malloc( send_buffer_header_length
						+ send_buffer_message_length );

	if ( send_buffer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating send buffer for MX response message.");
	}

	send_buffer_header = (mx_uint32_type *) send_buffer;

	send_buffer_message = send_buffer + send_buffer_header_length;

	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl(status.code);

        if ( status.code == MXE_SUCCESS ) {
		strcpy( send_buffer_message, "" );
	} else {
		strcpy( send_buffer_message, status.message );
	}

	send_buffer_header[ MX_NETWORK_MAGIC ] = htonl(MX_NETWORK_MAGIC_VALUE);
	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
				= htonl(MX_NETMSG_PUT_ARRAY_ASCII_RESPONSE);
	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );
	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_length );

	/* Send the string description back to the client. */

#if 0
	MX_DEBUG(-2,("%s: sending response = '%s'",
			fname, send_buffer_message ));
#endif

	status = mx_socket_send_message( mx_socket, send_buffer );

	MX_DEBUG(1,("%s: mx_socket_send_message() returned status.code = %ld",
				fname, status.code ));
	MX_DEBUG(1,("***** %s exiting *****", fname));

	free( send_buffer );

	return status;
}

mx_status_type
mxsrv_handle_get_field_type( MX_SOCKET mx_socket, MX_RECORD_FIELD *field,
				char *message, long message_length )
{
	const char fname[] = "mxsrv_handle_get_field_type()";

	char *send_buffer;
	mx_uint32_type *send_buffer_header, *send_buffer_message;
	long i, send_buffer_header_length, send_buffer_message_length;
	mx_status_type status;

	MX_DEBUG(1,("%s for '%s', message_length = %ld",
			fname, message, message_length));
	MX_DEBUG(1,("%s: field->datatype = %ld, field->num_dimensions = %ld",
			fname, field->datatype, field->num_dimensions));
	MX_DEBUG(1,("%s: field->dimension[0] = %ld",
			fname, field->dimension[0]));

	/* Allocate a buffer to construct the message to be sent
	 * back to the client.
	 */

	send_buffer_header_length = MX_NETWORK_HEADER_LENGTH_VALUE;

	send_buffer_message_length = sizeof(mx_uint32_type)
		* ( 2 + field->num_dimensions );

	/* Ensure that we can always assign a value to the first element
	 * of the dimension array.
	 */

	if ( send_buffer_message_length <= (2 * sizeof(mx_uint32_type)) ) {
		send_buffer_message_length = 3 * sizeof(mx_uint32_type);
	}

	send_buffer = (char *) malloc( send_buffer_header_length
					+ send_buffer_message_length );

	if ( send_buffer == NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating send buffer for MX response message.");
	}

	send_buffer_header = (mx_uint32_type *) send_buffer;

	send_buffer_message = (mx_uint32_type *)
				(send_buffer + send_buffer_header_length);

	/* Construct the header of the message. */

	send_buffer_header[ MX_NETWORK_MAGIC ] = htonl(MX_NETWORK_MAGIC_VALUE);
	send_buffer_header[ MX_NETWORK_MESSAGE_TYPE ]
				= htonl(MX_NETMSG_GET_FIELD_TYPE_RESPONSE);
	send_buffer_header[ MX_NETWORK_HEADER_LENGTH ]
				= htonl( send_buffer_header_length );
	send_buffer_header[ MX_NETWORK_MESSAGE_LENGTH ]
				= htonl( send_buffer_message_length );
	send_buffer_header[ MX_NETWORK_STATUS_CODE ] = htonl( MXE_SUCCESS );

	/* Construct the body of the message. */

	send_buffer_message[0] = htonl( field->datatype );
	send_buffer_message[1] = htonl( field->num_dimensions );

	if ( field->num_dimensions <= 0 ) {
		send_buffer_message[2] = htonl( 0L );
	} else {
		for ( i = 0; i < field->num_dimensions; i++ ) {
			send_buffer_message[i+2] = htonl(field->dimension[i]);
		}
	}

#if 0
	mx_network_display_message_buffer( send_buffer );
#endif

	/* Send the field type information back to the client. */

	status = mx_socket_send_message( mx_socket, send_buffer );

	MX_DEBUG(1,("%s: mx_socket_send_message() returned status.code = %ld",
				fname, status.code ));

	free( send_buffer );

	return status;
}

