/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    ms_prdou.c
 *
 * Purpose: Functions used to process MX digital output record events.
 */

#include <stdio.h>

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_drivr.h"
#include "mx_rec.h"
#include "mx_dout.h"
#include "mx_sock.h"

#include "ms_serv.h"
#include "ms_prdou.h"

mx_status_type
mxsrv_setup_digital_output_process_functions( MX_RECORD *record_list_head )
{
	const char fname[] = "mxsrv_setup_digital_output_process_functions()";

	MX_RECORD *current_record;
	MX_RECORD_FIELD *record_field;
	MX_RECORD_FIELD *record_field_array;
	long i;

	MX_DEBUG(2,("%s invoked.", fname));

	/* The following is done just in case we were called with
	 * a record that isn't really the list head.
	 */

	record_list_head = record_list_head->list_head;

	/* Go through the record list looking for digital_output records. */

	current_record = record_list_head->next_record;

	while ( current_record != record_list_head ) {

		if ( (current_record->superclass == MXR_DEVICE)
		  && (current_record->class == MXC_DIGITAL_OUTPUT) ) {

			record_field_array
				= current_record->record_field_array;

			for ( i = 0;
				i < current_record->num_record_fields;
				i++ )
			{
				record_field = &record_field_array[i];

				switch( record_field->label_value ) {
				case MXLV_DOU_VALUE:
				    record_field->process_function
					= mxsrv_digital_output_process_function;
				    break;
				default:
				    break;
				}
			}
		}
		current_record = current_record->next_record;
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_digital_output_process_function( void *record_ptr,
			void *record_field_ptr, int operation )
{
	const char fname[] = "mxsrv_digital_output_process_function()";

	MX_RECORD *record;
	MX_RECORD_FIELD *record_field;
	MX_DIGITAL_OUTPUT *digital_output;
	char operation_string[30];
	long label_value;
	unsigned long value;
	mx_status_type status;

	record = (MX_RECORD *) record_ptr;
	record_field = (MX_RECORD_FIELD *) record_field_ptr;

	switch ( operation ) {
	case MX_PROCESS_GET:
		strcpy( operation_string, "(MX_PROCESS_GET)" );
		break;
	case MX_PROCESS_PUT:
		strcpy( operation_string, "(MX_PROCESS_PUT)" );
		break;
	default:
		sprintf( operation_string, "(Operation %d)", operation );
		break;
	}

	label_value = record_field->label_value;

	MX_DEBUG(1,("%s: '%s.%s' %s (%ld)",
		fname, record->name, record_field->name,
		operation_string, label_value));

	/*
	 * A label value < 0 tells us that we do not need to do any
	 * further MX record processing. 
	 */

	if ( label_value < 0 ) {
		return MX_SUCCESSFUL_RESULT;
	}

	digital_output = (MX_DIGITAL_OUTPUT *) (record->record_class_struct);

	status = MX_SUCCESSFUL_RESULT;

	switch( operation ) {
	case MX_PROCESS_GET:
		switch( label_value ) {
		case MXLV_DOU_VALUE:
			status = mx_digital_output_read(record, &value);
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_GET label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	case MX_PROCESS_PUT:
		switch( label_value ) {
		case MXLV_DOU_VALUE:
			status = mx_digital_output_write(record,
							digital_output->value);
			break;
		default:
			MX_DEBUG( 1,(
			    "%s: *** Unknown MX_PROCESS_PUT label value = %ld",
				fname, label_value));
			break;
		}

		/* Now invoke the value changed callback. */

		if ( status.code == MXE_SUCCESS ) {
			status = mxsrv_invoke_callback_list(
					MXSF_VALUE_CHANGED_CALLBACK,
					record_field );
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Unknown operation code = %ld", operation );
		break;
	}

	return status;
}

