/*
 * Copyright 1999 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

/*
 * Name:    ms_serv.c
 *
 * Purpose: MX server functions.
 */

#include <stdio.h>
#include <stdlib.h>

#ifndef OS_WIN32
#include <sys/times.h>
#endif

#include "mxconfig.h"
#include "mx_util.h"
#include "mx_rec.h"
#include "mx_sock.h"
#include "mx_clock.h"

#include "ms_serv.h"

mx_status_type
mxsrv_add_queued_event( MXSRV_SOCKET_HANDLER *socket_handler,
		MX_RECORD *record,
		MX_RECORD_FIELD *record_field,
		void *event_data )
{
	const char fname[] = "mxsrv_add_queued_event()";

	MXSRV_QUEUED_EVENT *event, *current_event;
	MX_CLOCK_TICK current_clock_tick;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD pointer passed was NULL." );
	}

	/* Create the event structure. */

	event = (MXSRV_QUEUED_EVENT *) malloc( sizeof(MXSRV_QUEUED_EVENT) );

	if ( event == (MXSRV_QUEUED_EVENT *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Ran out of memory allocating an MXSRV_QUEUED_EVENT for record '%s'",
			record->name );
	}

	event->socket_handler = socket_handler;
	event->record = record;
	event->record_field = record_field;
	event->event_data = event_data;
	event->next_event = NULL;

	/* Add the event to the queue. */

	current_event = record->event_queue;

	if ( current_event == NULL ) {
		record->event_queue = event;
	} else {
		while ( current_event->next_event != NULL ) {
			current_event = current_event->next_event;
		}
		current_event->next_event = event;
	}

	if ( (record->event_interval.low_order != 0)
	  || (record->event_interval.high_order != 0) ) {

		current_clock_tick = mx_current_clock_tick();

		record->next_event_time = mx_add_clock_ticks(
			current_clock_tick, record->event_interval );
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_process_queued_event( MX_RECORD *record,
			MX_CLOCK_TICK current_clock_tick )
{
	const char fname[] = "mxsrv_process_queued_event()";

	MXSRV_QUEUED_EVENT *queued_event;
	MXSRV_EVENT_HANDLER *event_handler;
	mx_status_type ( *process_queued_event_fn ) (MXSRV_QUEUED_EVENT *);
	int process_event, comparison;

	if ( record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"MX_RECORD pointer passed was NULL." );
	}

	MX_DEBUG(1,(
	"%s invoked for record '%s', current_clock_time = (%lu,%ld)",
		fname, record->name,
		(long) current_clock_tick.high_order,
		(long) current_clock_tick.low_order ));

	queued_event = (MXSRV_QUEUED_EVENT *) record->event_queue;

	if ( queued_event == NULL ) {

		/* There are no events to dequeue. */

		return MX_SUCCESSFUL_RESULT;
	}

	/* Is it time to process the next event? */

	if ( (record->event_interval.low_order == 0)
	  && (record->event_interval.high_order == 0) ) {

		process_event = TRUE;
	} else {
		comparison = mx_compare_clock_ticks( current_clock_tick,
						record->next_event_time );
		if ( comparison >= 0 ) {
			process_event = TRUE;
		} else {
			process_event = FALSE;
		}
	}

	if ( process_event == TRUE ) {
		/* Remove this event from the queue. */

		record->event_queue = queued_event->next_event;

		/* Get a pointer to the event type specific
		 * process_queued_event function.
		 */

		event_handler = queued_event->socket_handler->event_handler;

		process_queued_event_fn = event_handler->process_queued_event;

		if ( process_queued_event_fn == NULL ) {
			mx_free( queued_event );

			return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Event handler type %ld does not have a process_queued_event function.",
				event_handler->type );
		}

		/* Invoke the event type specific processing function. */

		(void) ( *process_queued_event_fn ) ( queued_event );

		/* Dispose of the event now that we are done with it. */

		mx_free( queued_event );
	}

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_add_callback( long callback_type, MX_RECORD_FIELD *record_field,
	    mx_status_type( *callback_function ) (MX_RECORD_FIELD *, void *),
	    void *callback_argument )
{
	const char fname[] = "mxsrv_add_callback()";

	MXSRV_CALLBACK_LIST *server_callback_list;
	MXSRV_CALLBACK **list_ptr, **new_callback_ptr;
	MXSRV_CALLBACK *current_callback;

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );
	}

	if ( record_field->server_callback_list != NULL ) {
		server_callback_list = record_field->server_callback_list;
	} else {
		server_callback_list = (MXSRV_CALLBACK_LIST *)
				malloc( sizeof(MXSRV_CALLBACK_LIST) );

		if ( server_callback_list == (MXSRV_CALLBACK_LIST *) NULL ) {
			return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Attempt to allocate server_callback_list for field '%s' failed.",
				record_field->name );
		}

		server_callback_list->value_changed_callback_list = NULL;

		record_field->server_callback_list = server_callback_list;
	}

	/* Get the address of the pointer to the particular callback
	 * list we will be using.
	 */

	switch( callback_type ) {
	case MXSF_VALUE_CHANGED_CALLBACK:
		list_ptr = &(server_callback_list->value_changed_callback_list);
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The server callback type %ld is unrecognized.",
			callback_type );
		break;
	}

	/* We will be malloc'ing a new MXSRV_CALLBACK structure in a moment,
	 * but first we must determine the location at which the pointer
	 * returned by malloc() is to be stored.
	 */

	if ( *list_ptr == (MXSRV_CALLBACK *)NULL ) {
		new_callback_ptr = list_ptr;
	} else {
		/* Search until we find the end of the list. */

		current_callback = *list_ptr;

		while ( current_callback->next_callback != NULL ) {
			current_callback = current_callback->next_callback;
		}

		new_callback_ptr = &(current_callback->next_callback);
	}

	/* Now allocate the memory and store a pointer to it. */

	*new_callback_ptr = (MXSRV_CALLBACK *) malloc(sizeof(MXSRV_CALLBACK));

	if ( *new_callback_ptr == (MXSRV_CALLBACK *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
	"Attempt to allocate MXSRV_CALLBACK structure for field '%s' failed.",
			record_field->name );
	}

	current_callback = *new_callback_ptr;

	/* Finally, we initialize the new callback structure. */

	current_callback->callback_function = callback_function;

	current_callback->callback_argument = callback_argument;

	current_callback->callback_enabled = TRUE;

	current_callback->next_callback = NULL;

	return MX_SUCCESSFUL_RESULT;
}

mx_status_type
mxsrv_delete_callback( long callback_type, MX_RECORD_FIELD *record_field,
	    mx_status_type( *callback_function ) (MX_RECORD_FIELD *, void *),
	    void *callback_argument )
{
	const char fname[] = "mxsrv_delete_callback()";

	MXSRV_CALLBACK_LIST *server_callback_list;
	MXSRV_CALLBACK **list_ptr;
	MXSRV_CALLBACK *current_callback, *next_callback;
	MXSRV_CALLBACK *next_plus_one_callback;

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );
	}

	server_callback_list = (MXSRV_CALLBACK_LIST *)
				(record_field->server_callback_list);

	if ( server_callback_list == (MXSRV_CALLBACK_LIST *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"No callbacks are currently defined for field '%s'",
			record_field->name );
	}
	switch( callback_type ) {
	case MXSF_VALUE_CHANGED_CALLBACK:
		list_ptr = &(server_callback_list->value_changed_callback_list);
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The server callback type %ld is unrecognized.",
			callback_type );
		break;
	}

	current_callback = NULL;
	next_callback = *list_ptr;

	while ( next_callback != NULL ) {

		/* Check the next callback in the list to see if it is
		 * the one we want to delete.
		 */

		if (next_callback->callback_function == callback_function){
			if ( (callback_argument == NULL)
			  || (callback_argument
					== next_callback->callback_argument) )
			{
				/* We've found the callback to delete,
				 * so eliminate it from the list.
				 *
				 * If next_plus_one_callback turns out to
				 * be NULL, that's OK, so we don't need
				 * to test for it.
				 */

				next_plus_one_callback
					= next_callback->next_callback;

				if ( current_callback == NULL ) {

					/* This is the first callback in
					 * the list.
					 */
					
					*list_ptr = next_plus_one_callback;
				} else {
					current_callback->next_callback
						= next_plus_one_callback;
				}

				/* next_callback has been removed from
				 * the list, so we can now free it.
				 */

				next_callback->callback_function = NULL;
				next_callback->callback_argument = NULL;
				next_callback->callback_enabled = FALSE;
				next_callback->next_callback = NULL;

				mx_free(next_callback);

				/* We're done, so return to our caller. */

				return MX_SUCCESSFUL_RESULT;
			}
		}
		current_callback = next_callback;
		next_callback = current_callback->next_callback;
	}

	/* If we get here, we didn't find the callback. */

	return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
"Attempted to delete a non-existant callback function %p for field '%s'",
		callback_function, record_field->name );
}

mx_status_type
mxsrv_find_callback( long callback_type, MX_RECORD_FIELD *record_field,
	    mx_status_type( *callback_function ) (MX_RECORD_FIELD *, void *),
	    void *callback_argument, MXSRV_CALLBACK **callback_found )
{
	const char fname[] = "mxsrv_find_callback()";

	MXSRV_CALLBACK_LIST *server_callback_list;
	MXSRV_CALLBACK *current_callback;

	if ( callback_found == (MXSRV_CALLBACK **) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"Pointer to MXSRV_CALLBACK pointer passed was NULL." );
	}

	*callback_found = NULL;

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );
	}

	MX_DEBUG( 2,("%s invoked for record field '%s'",
			fname, record_field->name));

	server_callback_list = (MXSRV_CALLBACK_LIST *)
				(record_field->server_callback_list);

	if ( server_callback_list == (MXSRV_CALLBACK_LIST *) NULL ) {
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"No callbacks are currently defined for field '%s'",
			record_field->name );
	}
	switch( callback_type ) {
	case MXSF_VALUE_CHANGED_CALLBACK:
		current_callback
			= server_callback_list->value_changed_callback_list;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The server callback type %ld is unrecognized.",
			callback_type );
		break;
	}

	while ( current_callback != NULL ) {

		/* Check the callback to see if it is the one we want. */

		if (current_callback->callback_function == callback_function){
			if ( (callback_argument == NULL)
			  || (callback_argument
				== current_callback->callback_argument) )
			{
				/* We've found the callback so return it. */

				*callback_found = current_callback;

				return MX_SUCCESSFUL_RESULT;
			}
		}
		current_callback = current_callback->next_callback;
	}

	/* If we get here, we didn't find the callback. */

	return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
		"Did not find callback function %p for field '%s'",
		callback_function, record_field->name );
}

mx_status_type
mxsrv_invoke_callback_list( long callback_type, MX_RECORD_FIELD *record_field )
{
	const char fname[] = "mxsrv_invoke_callback_list()";

	MXSRV_CALLBACK_LIST *server_callback_list;
	MXSRV_CALLBACK *current_callback;
	int callback_enabled;
	mx_status_type ( *callback_fn ) ( MX_RECORD_FIELD *, void * );

	if ( record_field == (MX_RECORD_FIELD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
			"MX_RECORD_FIELD pointer passed was NULL." );
	}

	MX_DEBUG(1,("%s: $$$ invoked.", fname));

	server_callback_list = (MXSRV_CALLBACK_LIST *)
				(record_field->server_callback_list);

	if ( server_callback_list == (MXSRV_CALLBACK_LIST *) NULL ) {

		/* No callbacks are defined, so just return success. */

		return MX_SUCCESSFUL_RESULT;
	}

	switch( callback_type ) {
	case MXSF_VALUE_CHANGED_CALLBACK:
		current_callback
			= server_callback_list->value_changed_callback_list;
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"The server callback type %ld is unrecognized.",
			callback_type );
		break;
	}

	/* Invoke all the callbacks.  Do not abort if some of the
	 * callbacks fail.
	 */

	while( current_callback != NULL ) {

		callback_fn = current_callback->callback_function;

		callback_enabled = current_callback->callback_enabled;

		if ( callback_enabled == FALSE ) {

			MX_DEBUG(1,(
			"%s: Callback %p for field '%s' is disabled.",
				fname, current_callback, record_field->name));

		} else if ( callback_fn != NULL ) {

			MX_DEBUG(1,(
			"%s: Invoking callback %p for field '%s'.",
				fname, current_callback, record_field->name));

			(void) ( *callback_fn ) ( record_field,
					current_callback->callback_argument );
		}
		current_callback = current_callback->next_callback;
	}

	MX_DEBUG(1,("%s: $$$ succeeded.", fname));

	return MX_SUCCESSFUL_RESULT;
}

